/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AlbumInformationManager.h"
#include "AlbumInformationManager_p.h"

#include "CppUtil.h"
#include "XmlError.h"

#include <QNetworkRequest>
#include <QRegExp>
#include <QUrl>

namespace Private
{

    static const QLatin1String url( "https://ws.audioscrobbler.com/2.0/" );
    static const QLatin1String api( "0cd48d7dbf544440b00b1ca5d956f554" );

    //_______________________________________________________________________________
    AlbumInformationLocationJob::AlbumInformationLocationJob( const TrackInformation& trackInformation, QNetworkReply* networkReply ):
        NetworkJob( networkReply ),
        trackInformation_( trackInformation )
    { Debug::Throw( "Private::AlbumInformationLocationJob::AlbumInformationLocationJob.\n" ); }

    //_______________________________________________________________________________
    void AlbumInformationLocationJob::_finished()
    {
        Debug::Throw()
            << "Private::AlbumInformationLocationJob::_finished -"
            << " artist: " << trackInformation_.artist()
            << " album: " << trackInformation_.album()
            << endl;

        SignalEmiter( this );

        // check reply and status
        if( !( _hasNetworkReply() && _networkReply().error() == QNetworkReply::NoError ) )
        {
            emit releaseDateFound( trackInformation_, 0 );
            return;
        }

        QString extraLargeUrl;
        QString largeUrl;

        // read data
        const auto data( _networkReply().readAll() );

        // create dom document
        QDomDocument document;
        XmlError error;
        if( !document.setContent( data, error ) )
        {
            emit releaseDateFound( trackInformation_, 0 );
            return;
        }

        // parse document
        // try read release date directly
        auto nodeList( document.elementsByTagName( "releasedate" ) );
        for( int i = 0; i < nodeList.count(); ++i )
        {

            // get element
            const auto element( nodeList.at( i ).toElement() );

            // check parent
            if( element.parentNode().nodeName() != QLatin1String("album") ) continue;

            auto date( element.text() );
            static QRegExp regExp( "\\s+(\\d{4,4})(?:\\s|,)" );
            if( date.indexOf( regExp ) >= 0 )
            {
                const auto releaseDate = regExp.cap(1).toInt();
                Debug::Throw() << "AlbumInformationLocationJob::_finished - " << trackInformation_ << " release: " << releaseDate << endl;
                emit releaseDateFound( trackInformation_, releaseDate );
                return;
            }
        }

        // if release date was not found try get url instead
        nodeList = document.elementsByTagName( "url" );
        for( int i = 0; i < nodeList.count(); ++i )
        {

            // get element
            const auto element( nodeList.at( i ).toElement() );

            // check parent
            if( element.parentNode().nodeName() != QLatin1String("album") ) continue;

            emit informationFound( trackInformation_, QUrl( element.text() ) );
            return;
        }

        // invalid date
        emit releaseDateFound( trackInformation_, 0 );
        return;

    }


    //_______________________________________________________________________________
    AlbumInformationJob::AlbumInformationJob( const TrackInformation& trackInformation, QNetworkReply* networkReply ):
        NetworkJob( networkReply ),
        trackInformation_( trackInformation )
    { Debug::Throw( "Private::AlbumInformationJob::AlbumInformationJob.\n" ); }

    //_______________________________________________________________________________
    void AlbumInformationJob::_finished()
    {
        Debug::Throw( "Private::AlbumInformationJob::_finished.\n" );

        SignalEmiter( this );

        // check reply and status
        if( !( _hasNetworkReply() && _networkReply().error() == QNetworkReply::NoError ) )
        {
            emit releaseDateFound( trackInformation_, 0 );
            return;
        }

        // parse output
        const QString content( _networkReply().readAll() );
        static QRegExp regExp( "Release date</h2>\\s+<p.*(\\d{4,4})</p>" );
        regExp.setMinimal( true );

        if( content.indexOf( regExp ) >= 0 )
        {
            const auto releaseDate = regExp.cap(1).toInt();
            Debug::Throw() << "AlbumInformationJob::_finished - " << trackInformation_ << " release: " << releaseDate << endl;
            emit releaseDateFound( trackInformation_, releaseDate );
        }

        emit releaseDateFound( trackInformation_, 0 );
        return;

    }

}


//_______________________________________________________________________________
AlbumInformationManager::AlbumInformationManager( QObject* parent ):
    QObject( parent ),
    Counter( "AlbumInformationManager" ),
    networkAccessManager_( new QNetworkAccessManager( this ) )
{ Debug::Throw( "AlbumInformationManager::AlbumInformationManager.\n" ); }

//_______________________________________________________________________________
void AlbumInformationManager::get( const TrackInformation& trackInformation )
{

    Debug::Throw( "AlbumInformationManager::get.\n" );

    // check track information
    if( trackInformation.artist().isEmpty() ) return;
    if( trackInformation.album().isEmpty() || trackInformation.artist().isEmpty() ) return;

    // setup query items
    auto queryItems = Base::makeT<Base::QueryItemList>(
    {
        { QLatin1String( "method" ), QLatin1String( "album.getInfo" ) },
        { QLatin1String( "api_key" ), Private::api },
        { QLatin1String( "autocorrect" ), QLatin1String( "1" ) },
        { QLatin1String( "artist" ), trackInformation.artist() },
        { QLatin1String( "album" ), trackInformation.album() }
    });

    // create url
    QUrl url( Private::url );
    QUrlUtil::addQueryItems( url, queryItems );

    // network reply
    auto albumInformationLocationJob = new Private::AlbumInformationLocationJob( trackInformation, networkAccessManager_->get( QNetworkRequest(url) ) );
    connect( albumInformationLocationJob, SIGNAL(releaseDateFound(TrackInformation,int)), SLOT(_save(TrackInformation,int)) );
    connect( albumInformationLocationJob, SIGNAL(informationFound(TrackInformation,QUrl)), SLOT(_download(TrackInformation,QUrl)) );
}


//_______________________________________________________________________________
void AlbumInformationManager::timerEvent( QTimerEvent* event )
{

    if( event->timerId() == timer_.timerId() )
    {
        timer_.stop();
        if( !pendingItems_.empty() )
        {
            emit releaseDatesFound( pendingItems_ );
            pendingItems_.clear();
        }

    } else return QObject::timerEvent( event );

}

//_______________________________________________________________________________
void AlbumInformationManager::_download( const TrackInformation& trackInformation, const QUrl& source )
{
    Debug::Throw( "AlbumInformationManager::_download.\n" );
    auto albumInformationJob = new Private::AlbumInformationJob( trackInformation, networkAccessManager_->get( QNetworkRequest( source ) ) );
    connect( albumInformationJob, SIGNAL(releaseDateFound(TrackInformation,int)), SLOT(_save(TrackInformation,int)) );
    connect( albumInformationJob, SIGNAL(finished()), albumInformationJob, SLOT(deleteLater()) );
}

//_______________________________________________________________________________
void AlbumInformationManager::_save( const TrackInformation& trackInformation, int releaseDate )
{
    Debug::Throw( "AlbumInformationManager::_save.\n" );
    _delayedReleaseDateFound( trackInformation, releaseDate );
}

//_______________________________________________________________________________
void AlbumInformationManager::_delayedReleaseDateFound( const TrackInformation& trackInformation, int releaseDate )
{
    pendingItems_.append( { trackInformation, releaseDate } );
    if( !timer_.isActive() ) timer_.start( 100, this );
}
