/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Application.h"

#include "AlbumInformationManager.h"
#include "AudioStreamInterface.h"
#include "AudioStreamWidget.h"
#include "CollectionItem.h"
#include "CollectionWatcher.h"
#include "CollectionWidget.h"
#include "ConfigurationDialog.h"
#include "CoverManager.h"
#include "CppUtil.h"
#include "CurrentTrackInformationWidget.h"
#include "IconNames.h"
#include "LogDialog.h"
#include "LyricsInformation.h"
#include "LyricsManager.h"
#include "LyricsProviderWidget.h"
#include "LyricsWidget.h"
#include "MainWindow.h"
#include "MpdCommand.h"
#include "MpdController.h"
#include "MpdInterface.h"
#include "MpdIdleInterface.h"
#include "MpdThread.h"
#include "PlaybackControlToolBar.h"
#include "PlaylistWidget.h"
#include "ReconnectionWidget.h"
#include "StoredPlaylistWidget.h"
#include "SystemTrayIcon.h"
#include "SystemNotifications.h"
#include "ToolTipWidget.h"

#if WITH_CDIO_PARANOIA
#include "AudioDeviceInterface.h"
#include "AudioDeviceWidget.h"
#endif

#if WITH_TAGLIB
#include "TagEditionDialog.h"
#endif

#include <QApplication>
#include <QMenu>
#include <QTimer>

#include <algorithm>

// use a long timeout for starting the server
static const int serverTimeOut = 5000;

//____________________________________________
Application::Application( CommandLineArguments arguments ):
    BaseApplication( 0, arguments ),
    Counter( "Application" )
{
    Debug::Throw( "Application::Application.\n" );

    qRegisterMetaType<Mpd::State>( "Mpd::State" );
    qRegisterMetaType<Mpd::Command>( "Mpd::Command" );
    qRegisterMetaType<Mpd::Command::List>( "Mpd::Command::List" );

    qRegisterMetaType<LyricsInformation>( "LyricsInformation" );

    qRegisterMetaType<TrackInformation>( "TrackInformation" );
    qRegisterMetaType<TrackInformation::List>( "TrackInformation::List" );

    qRegisterMetaType<CollectionItem>( "CollectionItem" );
    qRegisterMetaType<CoverManager::CoverList>( "CoverManager::CoverList" );

    qRegisterMetaType<AudioStreamItem>( "AudioStreamItem" );
    qRegisterMetaType<AudioStreamItem::List>( "AudioStreamItem::List" );

    qRegisterMetaType<AlbumInformationManager::ReleaseDateList>( "AlbumInformationManager::ReleaseDateList" );
}

//____________________________________________
Application::~Application()
{
    Debug::Throw( "Application::~Application.\n" );

    // terminate mpd
    if( XmlOptions::get().get<bool>( "MPD_CLOSE_DEAMON_ON_EXIT" ) && controller_ && controller_->serverIsRunning() )
    { _stopServer(); }

    // terminate all threads
    for( auto thread:findChildren<Mpd::Thread*>() )
    {
        thread->quit();
        thread->wait();
    };

}

//____________________________________________
bool Application::initApplicationManager()
{
    Debug::Throw( "Application::initApplicationManager.\n" );

    // retrieve files from arguments and expand if needed
    auto parser = commandLineParser( _arguments() );
    for( auto& orphan:parser.orphans() )
    { if( !orphan.isEmpty() ) File::expand( orphan ); }

    // replace arguments
    _setArguments( parser.arguments() );

    // base class initialization
    return BaseApplication::initApplicationManager();

}

//____________________________________________
bool Application::realizeWidget()
{
    Debug::Throw( "Application::realizeWidget.\n" );

    // check if the method has already been called.
    if( !BaseApplication::realizeWidget() ) return false;

    // disable exit on last window
    qApp->setQuitOnLastWindowClosed( false );

    // actions
    _installActions();

    // cover cache
    coverCache_.reset( new CoverCache );

    // main window
    mainWindow_ = new MainWindow;

    // initialize server configuration
    const auto serverConfiguration( Mpd::ServerConfiguration::fromOptions() );

    // initialize controller
    controller_ = new Mpd::Controller( this );
    controller_->setServerConfiguration( serverConfiguration );
    controller_->checkServerConfiguration();

    // collection watcher
    collectionWatcherThread_ = new Mpd::Thread( this, new CollectionWatcher );
    collectionWatcherThread_->start();

    // cover manager
    coverManagerThread_ = new Mpd::Thread( this, new CoverManager );
    coverManagerThread_->start();

    // album information manager
    albumInformationManagerThread_ = new Mpd::Thread( this, new AlbumInformationManager );
    albumInformationManagerThread_->start();

    // lyrics provider
    lyricsManagerThread_ = new Mpd::Thread( this, new LyricsManager );
    lyricsManagerThread_->start();

    // audio stream
    audioStreamThread_ = new Mpd::Thread( this, new AudioStreamInterface );
    audioStreamThread_->start();

    // audio device
    #if WITH_CDIO_PARANOIA
    auto audioDeviceInterface = new AudioDeviceInterface;
    audioDeviceThread_ = new Mpd::Thread( this, audioDeviceInterface );
    audioDeviceThread_->start();
    QMetaObject::invokeMethod( audioDeviceThread_->get(), "initializeServer", Qt::QueuedConnection );
    #endif

    // mpd interface
    auto interface = new Mpd::Interface;
    interface->setServerConfiguration( serverConfiguration );
    interfaceThread_ = new Mpd::Thread( this, interface );
    interfaceThread_->start();

    // mpd idle interface
    auto idleInterface = new Mpd::IdleInterface;
    idleInterface->setServerConfiguration( serverConfiguration );
    idleThread_ = new Mpd::Thread( this, idleInterface );
    idleThread_->start();

    // system notifications
    systemNotifications_ = new SystemNotifications( this );
    systemNotifications_->initialize();
    systemNotifications_->setApplicationName( applicationName() );
    systemNotifications_->setApplicationIcon( applicationIcon() );

    // connections
    connect( aboutMpdAction_, SIGNAL(triggered()), mainWindow_, SLOT(mpdInformation()) );

    connect( mainWindow_, SIGNAL(commandRequested(Mpd::Command)), interfaceThread_->get(), SLOT(addCommand(Mpd::Command)) );
    connect( mainWindow_, SIGNAL(commandsRequested(Mpd::Command::List)), interfaceThread_->get(), SLOT(addCommands(Mpd::Command::List)) );
    connect( mainWindow_, SIGNAL(newNotification(TrackInformation)), this, SLOT(_notify(TrackInformation)) );
    connect( &mainWindow_->reconnectionWidget().reconnectionButton(), SIGNAL(clicked()), this, SLOT(_reconnectServer()) );

    connect( collectionWatcherThread_->get(), SIGNAL(collectionChanged()), mainWindow_, SLOT(updateCollection()) );
    connect( collectionWatcherThread_->get(), SIGNAL(collectionChanged(QString)), mainWindow_, SLOT(updateCollection(QString)) );

    connect( mainWindow_, SIGNAL(requestCover(CollectionItem)), coverManagerThread_->get(), SLOT(get(CollectionItem)) );
    connect( coverManagerThread_->get(), SIGNAL(coversFound(CoverManager::CoverList)), coverCache_.get(), SLOT(saveCovers(CoverManager::CoverList)) );
    connect( coverManagerThread_->get(), SIGNAL(coversFound(CoverManager::CoverList)), &mainWindow_->currentTrackInformationWidget(), SLOT(updateCover(CoverManager::CoverList)) );
    connect( coverManagerThread_->get(), SIGNAL(coversFound(CoverManager::CoverList)), &mainWindow_->playlistWidget().toolTipWidget(), SLOT(updateCover(CoverManager::CoverList)) );
    connect( coverManagerThread_->get(), SIGNAL(coversFound(CoverManager::CoverList)), &mainWindow_->collectionWidget().toolTipWidget(), SLOT(updateCover(CoverManager::CoverList)) );

    #if WITH_TAGLIB
    connect( &mainWindow_->tagEditionDialog(), SIGNAL(requestAlbumInformation(TrackInformation)), albumInformationManagerThread_->get(), SLOT(get(TrackInformation)) );
    connect( albumInformationManagerThread_->get(), SIGNAL(releaseDatesFound(AlbumInformationManager::ReleaseDateList)), &mainWindow_->tagEditionDialog(), SLOT(updateYear(AlbumInformationManager::ReleaseDateList)) );
    #endif

    mainWindow_->playlistWidget().model().setCoverCache( coverCache_ );
    mainWindow_->collectionWidget().model().setCoverCache( coverCache_ );
    mainWindow_->storedPlaylistWidget().model().setCoverCache( coverCache_ );

    connect( &mainWindow_->lyricsWidget(), SIGNAL(requestLyrics(TrackInformation)), lyricsManagerThread_->get(), SLOT(get(TrackInformation)) );
    connect( lyricsManagerThread_->get(), SIGNAL(lyricsFound(TrackInformation,LyricsInformation)), &mainWindow_->lyricsWidget(), SLOT(lyricsFound(TrackInformation,LyricsInformation)) );
    connect( lyricsManagerThread_->get(), SIGNAL(lyricsNotFound(TrackInformation)), &mainWindow_->lyricsWidget(), SLOT(lyricsNotFound(TrackInformation)) );

    connect( &mainWindow_->audioStreamWidget(), SIGNAL(requestChildren(AudioStreamItem)), audioStreamThread_->get(), SLOT(findChildren(AudioStreamItem)) );
    connect( audioStreamThread_->get(), SIGNAL(childrenFound(AudioStreamItem,AudioStreamItem::List)), &mainWindow_->audioStreamWidget(), SLOT(addChildren(AudioStreamItem,AudioStreamItem::List)) );

    #if WITH_CDIO_PARANOIA
    mainWindow_->audioDeviceWidget().model().setCoverCache( coverCache_ );

    connect( mainWindow_, SIGNAL(elapsedTimeChanged(quint32)), audioDeviceThread_->get(), SLOT(saveElapsedTime(quint32)) );
    connect( &mainWindow_->audioDeviceWidget(), SIGNAL(requestReadDevice(QString)), audioDeviceThread_->get(), SLOT(readDevice(QString)) );
    connect( &mainWindow_->audioDeviceWidget(), SIGNAL(requestTryReadDevice(QString)), audioDeviceThread_->get(), SLOT(tryReadDevice(QString)) );
    connect( audioDeviceThread_->get(), SIGNAL(tracksFound(TrackInformation::List)), &mainWindow_->audioDeviceWidget(), SLOT(addTracks(TrackInformation::List)) );
    connect( audioDeviceThread_->get(), SIGNAL(invalidDevice(QString)), &mainWindow_->audioDeviceWidget(), SLOT(showNotificationWidget()) );

    connect( audioDeviceThread_->get(), SIGNAL(trackUpdated(TrackInformation,TrackInformation)), mainWindow_, SLOT(updateCover(TrackInformation,TrackInformation)) );
    connect( audioDeviceThread_->get(), SIGNAL(trackUpdated(TrackInformation,TrackInformation)), &mainWindow_->audioDeviceWidget(), SLOT(updateTrack(TrackInformation,TrackInformation)) );
    connect( audioDeviceThread_->get(), SIGNAL(trackUpdated(TrackInformation,TrackInformation)), &mainWindow_->playlistWidget(), SLOT(updateTrack(TrackInformation,TrackInformation)) );
    #endif

    connect( interfaceThread_->get(), SIGNAL(versionAvailable(QString)), mainWindow_, SLOT(saveMpdVersion(QString)) );
    connect( interfaceThread_->get(), SIGNAL(commandCompleted(Mpd::Command)), mainWindow_, SLOT(processCommandReply(Mpd::Command)) );
    connect( interfaceThread_->get(), SIGNAL(commandCompleted(Mpd::Command)), &mainWindow_->logDialog(), SLOT(addCommand(Mpd::Command)) );
    connect( interfaceThread_->get(), SIGNAL(socketError()), this, SLOT(_processSocketError()) );

    connect( idleThread_->get(), SIGNAL(requestUpdateStatus()), mainWindow_, SLOT(updateStatus()) );
    connect( idleThread_->get(), SIGNAL(requestUpdatePlaylist()), mainWindow_, SLOT(updatePlaylist()) );
    connect( idleThread_->get(), SIGNAL(requestLoadCollection()), mainWindow_, SLOT(loadCollection()) );
    connect( idleThread_->get(), SIGNAL(requestLoadStoredPlaylists()), mainWindow_, SLOT(loadStoredPlaylists()) );
    connect( idleThread_->get(), SIGNAL(socketError()), this, SLOT(_processSocketError()) );

    // start server
    if( !controller_->serverIsRunning() && controller_->startServer() )
    { controller_->waitForStarted( serverTimeOut ); }

    // reconnect idle thread
    QMetaObject::invokeMethod( idleThread_->get(), "connectSocket", Qt::QueuedConnection );

    // initialize playlist
    _populatePlaylist();

    // process command line arguments
    _processArguments( _arguments() );

    if( XmlOptions::get().get<bool>( "USE_SYSTEMTRAY" ) && QSystemTrayIcon::isSystemTrayAvailable() )
    {
        auto visibility = static_cast<ConfigurationDialog::WindowVisibilityMode>( XmlOptions::get().get<int>( "MAIN_WINDOW_VISIBILITY" ) );
        switch( visibility )
        {
            case ConfigurationDialog::WindowVisibilityMode::Show: showMainWindow(); break;

            case ConfigurationDialog::WindowVisibilityMode::Restore:
            if( !XmlOptions::get().contains( "MAIN_WINDOW_VISIBLE" ) ||  XmlOptions::get().get<bool>( "MAIN_WINDOW_VISIBLE" ) )
            { showMainWindow(); }
            break;

            default: break;
        }

    } else showMainWindow();

    // configuration
    connect( this, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    emit configurationChanged();

    return true;

}

//____________________________________________
CommandLineParser Application::commandLineParser( CommandLineArguments arguments, bool ignoreWarnings ) const
{
    Debug::Throw( "Application::commandLineParser.\n" );
    CommandLineParser out( BaseApplication::commandLineParser() );

    out.setGroup( CommandLineParser::applicationGroupName );
    out.registerFlag( CommandLineParser::Tag( "--play" ), tr( "start playback") );
    out.registerFlag( CommandLineParser::Tag( "--pause" ), tr( "pause playback") );
    out.registerFlag( CommandLineParser::Tag( "--stop" ), tr( "stop playback") );
    out.registerFlag( CommandLineParser::Tag( "--clear" ), tr( "clear playlist") );
    out.registerFlag( CommandLineParser::Tag( "--append" ), tr( "append tracks to playlist rather than replace playlist") );
    out.registerFlag( CommandLineParser::Tag( "--next" ), tr( "move to next track in playlist" ) );
    out.registerFlag( CommandLineParser::Tag( "--previous" ), tr( "move to previous track in playlist" ) );

    if( !arguments.isEmpty() ) out.parse( arguments, ignoreWarnings );
    return out;
}

//____________________________________________
void Application::usage() const
{
    _usage( "musicbox", tr("[options] [files]") );
    commandLineParser().usage();
}

//________________________________________________
void Application::toggleMainWindowVisibility()
{
    Debug::Throw( "Application::toggleMainWindowVisibility.\n" );
    if( mainWindow_->isHidden() || mainWindow_->isMinimized() ) showMainWindow();
    else hideMainWindow();
}

//________________________________________________
void Application::showMainWindow()
{
    Debug::Throw( "Application::showMainWindow.\n" );
    showMainWindowAction_->setText( tr( "Minimize" ) );
    XmlOptions::get().set<bool>( "MAIN_WINDOW_VISIBLE", true );
    mainWindow_->uniconify();
}

//________________________________________________
void Application::hideMainWindow()
{
    Debug::Throw( "Application::hideMainWindow.\n" );
    showMainWindowAction_->setText( tr( "Restore" ) );
    XmlOptions::get().set<bool>( "MAIN_WINDOW_VISIBLE", false );
    mainWindow_->hide();
}

//___________________________________________________________
void Application::requestExit()
{
    Debug::Throw( "Application::requestExit.\n" );
    if( controller_ && controller_->serverIsRunning() && mainWindow_ )
    {

        /*
        need to abort fade in all cases, in order to prevent an incorrect
        sound level to be saved
        */
        mainWindow_->playbackControlToolBar().abortFade();

        if( XmlOptions::get().get<bool>( "RESUME_PLAYBACK_ON_START" ) )
        {

            if( mainWindow_->state() == Mpd::State::Playing ) mainWindow_->pausePlayback();

        } else if( mainWindow_->state() != Mpd::State::Stopped ) {

            mainWindow_->stopPlayback();

        }
    }

    // exit
    qApp->exit();
}

//___________________________________________________________
void Application::_configuration()
{

    Debug::Throw( "Application::_configuration.\n" );
    emit saveConfiguration();

    // create configuration dialog
    ConfigurationDialog dialog( nullptr );
    dialog.setTextNotificationsSupported( systemNotifications_->isSupported() );
    dialog.realizeWidget();

    // get list of lyrics provider names and assign
    auto providerList = static_cast<LyricsManager*>( lyricsManagerThread_->get() )->providers();
    QStringList providerNames;
    std::transform( providerList.begin(), providerList.end(), std::back_inserter( providerNames ),
        []( BaseLyricsProvider* provider ) { return provider->name(); } );
    dialog.lyricsProviderWidget().model().set( providerNames );

    connect( &dialog, SIGNAL(configurationChanged()), SIGNAL(configurationChanged()) );
    connect( &dialog, SIGNAL(requestCollectionUpdate()), this, SLOT(_rescanCollection()) );

    dialog.centerOnWidget( qApp->activeWindow() );
    dialog.exec();
}

//________________________________________________
bool Application::_processCommand( Server::ServerCommand command )
{
    Debug::Throw( "Application::_processCommand.\n" );
    if( BaseApplication::_processCommand( command ) ) return true;
    if( command.command() == Server::ServerCommand::Raise )
    {
        _processArguments( command.arguments() );
        showMainWindow();
        return true;
    }

    return false;

}

//_______________________________________________
void Application::_updateConfiguration()
{

    // system tray icon
    if( QSystemTrayIcon::isSystemTrayAvailable() && XmlOptions::get().get<bool>( "USE_SYSTEMTRAY" ) )
    {
        if( !systemTrayIcon_ )
        {

            // create system tray icon
            systemTrayIcon_.reset( new SystemTrayIcon );
            connect( systemTrayIcon_.get(), SIGNAL( activated(QSystemTrayIcon::ActivationReason) ), SLOT( _systemTrayActivated(QSystemTrayIcon::ActivationReason)) );

            // create system tray menu
            auto menu = new QMenu;
            menu->addAction( &mainWindow_->playbackControlToolBar().previousTrackAction() );
            menu->addAction( &mainWindow_->playbackControlToolBar().startPlaybackAction() );
            menu->addAction( &mainWindow_->playbackControlToolBar().stopPlaybackAction() );
            menu->addAction( &mainWindow_->playbackControlToolBar().nextTrackAction() );
            menu->addSeparator();
            menu->addAction( showMainWindowAction_ );
            menu->addSeparator();
            menu->addAction( &closeAction() );

            systemTrayIcon_->setContextMenu( menu );

        }

        systemTrayIcon_->show();

    } else {

        mainWindow_->show();
        systemTrayIcon_.reset();

    }

    // update collection watcher
    if( collectionWatcherThread_ )
    {
        QMetaObject::invokeMethod( collectionWatcherThread_->get(), "updateEnableState", Qt::QueuedConnection );
        QMetaObject::invokeMethod( collectionWatcherThread_->get(), "updateMusicDirectory", Qt::QueuedConnection );
    }

    // enable cover manager
    static_cast<CoverManager*>( coverManagerThread_->get() )->setEnabled( XmlOptions::get().get<bool>( "ENABLE_COVER_MANAGER" ) );

    // enable lyrics manager
    static_cast<LyricsManager*>( lyricsManagerThread_->get() )->setEnabled( XmlOptions::get().get<bool>( "ENABLE_LYRICS_MANAGER" ) );

    // enable cddb lookup
    #if WITH_CDIO_PARANOIA
    static_cast<AudioDeviceInterface*>( audioDeviceThread_->get() )->setCddbEnabled( XmlOptions::get().get<bool>( "ENABLE_CDDB" ) );
    #endif

    // get server configuration from controller
    const auto serverConfiguration( Mpd::ServerConfiguration::fromOptions() );
    const auto oldServerConfiguration( controller_->serverConfiguration() );
    if( serverConfiguration != oldServerConfiguration )
    {

        // gracefully disconnect threads
        QMetaObject::invokeMethod( idleThread_->get(), "disconnectSocket", Qt::QueuedConnection );
        QMetaObject::invokeMethod( interfaceThread_->get(), "disconnectSocket", Qt::QueuedConnection );

        // shutdown controller
        if( controller_->serverIsRunning() )
        {
            _lock();
            _stopServer();
            controller_->waitForStopped();
            _unlock();
        }

        // assign new configuration to controller
        controller_->setServerConfiguration( serverConfiguration );

        // assign new configuration to interfaces
        static_cast<Mpd::Interface*>( interfaceThread_->get() )->setServerConfiguration( serverConfiguration );

        // assign new configuration to interfaces
        static_cast<Mpd::IdleInterface*>( idleThread_->get() )->setServerConfiguration( serverConfiguration );

        // restart controler
        controller_->checkServerConfiguration();
        if( controller_->startServer() )
        { controller_->waitForStarted(serverTimeOut); }

        // reconnect idle thread
        QMetaObject::invokeMethod( idleThread_->get(), "connectSockete", Qt::QueuedConnection );

        // update
        _populatePlaylist();

    }

}

//___________________________________________________________
void Application::_processSocketError()
{
    if( !( locked_ || controller_->serverIsRunning() ) && mainWindow_ )
    {
        mainWindow_->clear();
        mainWindow_->reconnectionWidget().show();
    }
}

//_______________________________________________
void Application::_rescanCollection()
{
    Debug::Throw( "Application::_rescanCollection.\n" );

    // clear playlist
    mainWindow_->clearPlaylist();

    // gracefully disconnect threads
    QMetaObject::invokeMethod( idleThread_->get(), "disconnectSocket", Qt::QueuedConnection );
    QMetaObject::invokeMethod( interfaceThread_->get(), "disconnectSocket", Qt::QueuedConnection );

    // shutdown controller
    if( controller_->serverIsRunning() )
    {
        _lock();
        _stopServer();
        controller_->waitForStopped();
        _unlock();
    }

    // remove db file
    const auto dbFile( static_cast<Mpd::Interface*>( interfaceThread_->get() )->serverConfiguration().dbFile() );
    if( dbFile.exists() ) dbFile.remove();

    // clear collection model
    mainWindow_->collectionWidget().model().clear();

    // restart server
    if( controller_->startServer() )
    { controller_->waitForStarted(serverTimeOut); }

    // reconnect idle thread
    QMetaObject::invokeMethod( idleThread_->get(), "connectSocket", Qt::QueuedConnection );

    // update status
    /* there is no need to update the collection, because this is triggered automatically by mpd
    when starting with the db file removed */
    mainWindow_->updateStatus();

}

//_______________________________________________
void Application::_systemTrayActivated( QSystemTrayIcon::ActivationReason reason )
{
    Debug::Throw( "Application::_systemTrayActivated.\n" );
    switch( reason )
    {

        default: break;

        case QSystemTrayIcon::Trigger:
        toggleMainWindowVisibility();
        break;

    }
}

//_______________________________________________
void Application::_reconnectServer()
{
    Debug::Throw( "Application::_reconnectServer.\n" );

    // restart server
    if( !controller_->serverIsRunning() && controller_->startServer() )
    { controller_->waitForStarted( serverTimeOut ); }

    // reconnect idle thread
    QMetaObject::invokeMethod( idleThread_->get(), "connectSocket", Qt::QueuedConnection );

    // ask for playist population
    _populatePlaylist();

}

//_______________________________________________
void Application::_restartServer()
{
    Debug::Throw( "Application::_restartServer.\n" );

    // gracefully disconnect threads
    QMetaObject::invokeMethod( idleThread_->get(), "disconnectSocket", Qt::QueuedConnection );
    QMetaObject::invokeMethod( interfaceThread_->get(), "disconnectSocket", Qt::QueuedConnection );

    // shutdown controller
    if( controller_->serverIsRunning() )
    {
        _lock();
        _stopServer();
        controller_->waitForStopped();
        _unlock();
    }

    // restart controler
    controller_->checkServerConfiguration();
    if( controller_->startServer() )
    { controller_->waitForStarted(serverTimeOut); }

    // reconnect idle thread
    QMetaObject::invokeMethod( idleThread_->get(), "connectSocket", Qt::QueuedConnection );

    // ask for playist population
    _populatePlaylist();

}

//_______________________________________________
void Application::_notify( const TrackInformation& trackInformation )
{

    // build message
    #if defined(Q_OS_WIN)
    auto message = tr( "%1 by %2 on %3" ).arg(
        trackInformation.title(),
        trackInformation.artist(),
        trackInformation.albumOrChannel() );
    #else
    auto message = tr( "<b>%1</b> by <b>%2</b> on <b>%3</b>" ).arg(
        trackInformation.title(),
        trackInformation.artist(),
        trackInformation.albumOrChannel() );
    #endif

    // add duration
    if( trackInformation.duration() > 0 )
    { message += QString( " (%1)" ).arg( trackInformation.durationString() ); }

    Debug::Throw( "ClientApplication::_notify (text).\n" );
    if( !XmlOptions::get().get<bool>( "USE_NOTIFICATIONS" ) ) return;
    else if( systemNotifications_->isSupported() )
    {

        // try system notifications
        Notification notification( "Now playing", message );

        // try load icon from cache
        QIcon icon;
        if( trackInformation.isAudioStream() ) icon = coverCache_->icon( CollectionItem( trackInformation ) );
        else icon = coverCache_->icon( CollectionItem( CollectionItem::Type::Album, trackInformation.artist(), trackInformation.album() ) );

        if( !icon.isNull() ) notification.setIcon( icon );

        systemNotifications_->sendNotification( notification );

    } else if( QSystemTrayIcon::supportsMessages() && QSystemTrayIcon::isSystemTrayAvailable() && systemTrayIcon_ ) {

        // try system tray messages
        systemTrayIcon_->showMessage( "Now playing", message );

    }

}

//_______________________________________________
void Application::_installActions()
{
    Debug::Throw( "Application::_installActions.\n" );

    // need to modify closeAction signal for proper exit
    closeAction().disconnect();
    connect( &closeAction(), SIGNAL(triggered()), SLOT(requestExit()) );

    // raise main window
    showMainWindowAction_ = new QAction( tr( "Restore" ), this );
    connect( showMainWindowAction_, SIGNAL(triggered()), SLOT(toggleMainWindowVisibility()) );

    // restart mpd
    restartServerAction_ = new QAction( IconEngine::get( IconNames::Reload), tr( "Restart Server" ), this );
    connect( restartServerAction_, SIGNAL(triggered()), SLOT(_restartServer()) );

    // about mpd
    aboutMpdAction_ = new QAction( applicationIcon(), tr( "Server Information" ), this );

}

//___________________________________________________________
void Application::_populatePlaylist()
{
    Debug::Throw( "Application::_populatePlaylist.\n" );

    if( mainWindow_ )
    {
        if( XmlOptions::get().get<bool>( "UPDATE_COLLECTION_AT_STARTUP" ) )
        { mainWindow_->updateCollection(); }

        mainWindow_->loadCollection();
        mainWindow_->loadStoredPlaylists();
        mainWindow_->updatePlaylist();
        mainWindow_->updateStatus();
        mainWindow_->loadHandlers();

    }
}

//___________________________________________________________
void Application::_processArguments( const CommandLineArguments& arguments )
{

    Debug::Throw( "Application::_processArguments.\n" );

    if( !mainWindow_ ) return;

    // command line parser
    const auto parser( commandLineParser( arguments ) );

    // clear playlist
    if( parser.hasFlag( "--clear" ) )
    { mainWindow_->clearPlaylist(); }

    // read files from orphans
    File::List files;
    for( const auto& orphan:parser.orphans() )
    {
        const File file( orphan );
        if( !file.exists() ) continue;

        // in case of directory list content, recursively
        if( file.isDirectory() ) files.append( file.listFiles( static_cast<File::ListFlags>(File::Recursive|File::FollowLinks) ) );
        else files.append( file );

    }

    // add files to playlist
    if( !files.isEmpty() )
    {
        if( parser.hasFlag( "--append" ) ) mainWindow_->addFromFileSystem( files );
        else mainWindow_->replaceFromFileSystem( files );
    }

    // navigate playlist
    if( parser.hasFlag( "--next" ) ) mainWindow_->playbackControlToolBar().emitRequestNextTrack();
    else if( parser.hasFlag( "--previous" ) ) mainWindow_->playbackControlToolBar().emitRequestPreviousTrack();

    // playback options
    if( parser.hasFlag( "--play" ) )
    {
        switch( mainWindow_->state() )
        {
            default:
            case Mpd::State::Playing:
            break;

            case Mpd::State::Paused:
            mainWindow_->playbackControlToolBar().emitRequestResumePlayback();
            break;

            case Mpd::State::Stopped:
            mainWindow_->playbackControlToolBar().emitRequestStartPlayback();
            break;
        }

    } else if( parser.hasFlag( "--pause" ) && mainWindow_->state() == Mpd::State::Playing ) {

        mainWindow_->playbackControlToolBar().emitRequestPausePlayback();

    } else if( parser.hasFlag( "--stop" ) ) {

        switch( mainWindow_->state() )
        {
            case Mpd::State::Playing:
            case Mpd::State::Paused:
            mainWindow_->playbackControlToolBar().emitRequestStopPlayback();
            break;

            default:
            break;
        }

    }

}

//_______________________________________________
void Application::_stopServer()
{

    Debug::Throw( "Application::_stopServer.\n" );

    if( interfaceThread_ )
    {

        QMetaObject::invokeMethod(
            interfaceThread_->get(), "addCommand",
            Qt::QueuedConnection,
            Q_ARG( Mpd::Command, Mpd::Command( Mpd::Command::Type::Kill, "kill" ) ) );

    } else if( controller_ ) {

        controller_->stopServer();

    }

}
