#ifndef Application_h
#define Application_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "BaseApplication.h"
#include "CommandLineArguments.h"
#include "CommandLineParser.h"
#include "Config.h"
#include "CoverCache.h"
#include "IconEngine.h"
#include "MpdCommand.h"
#include "TrackInformation.h"

#include <QAction>
#include <QHash>
#include <QIcon>
#include <QLocalSocket>
#include <QSystemTrayIcon>

#include <memory>

class MainWindow;
class SystemNotifications;
class SystemTrayIcon;

namespace Mpd
{
    class Controller;
    class Thread;
}

//* Main Window singleton object
class Application: public BaseApplication, private Base::Counter<Application>
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    explicit Application( CommandLineArguments );

    //* destructor
    ~Application() override;

    //* initialize application manager
    bool initApplicationManager() override;

    //* create all widgets
    bool realizeWidget() override;

    //*@name accessors
    //@{

    //* command line parser
    CommandLineParser commandLineParser( CommandLineArguments = CommandLineArguments(), bool ignoreWarnings = true ) const override;

    //* command line help
    void usage() const override;

    //* application name
    QString applicationName() const override
    { return "Musicbox"; }

    //* application icon
    QIcon applicationIcon() const override
    { return IconEngine::get( ":/musicbox.png" ); }

    //* application version
    QString applicationVersion() const override
    { return VERSION; }

    //* main window
    MainWindow& mainWindow() const
    { return *mainWindow_; }

    //* show main window
    QAction& showMainWindowAction() const
    { return *showMainWindowAction_; }

    //* restart mpd
    QAction& restartServerAction() const
    { return *restartServerAction_; }

    //* about mpd
    QAction& aboutMpdAction() const
    { return *aboutMpdAction_; }

    //@}

    public Q_SLOTS:

    //* toggle main window visibility
    void toggleMainWindowVisibility();

    //* show main window
    void showMainWindow();

    //* hide main window
    void hideMainWindow();

    //* request exit
    void requestExit();

    protected Q_SLOTS:

    //* configuration
    void _configuration() override;

    //* process request from application manager
    bool _processCommand( Server::ServerCommand ) override;

    private Q_SLOTS:

    //* configuration
    void _updateConfiguration();

    //* rescan collection entirely
    void _rescanCollection();

    //* system tray icon activated
    void _systemTrayActivated( QSystemTrayIcon::ActivationReason );

    //* process socket error
    void _processSocketError();

    //* notify server lost
    void _reconnectServer();

    //* restart server
    void _restartServer();

    //* text notification
    void _notify( const TrackInformation& );

    private:

    //* install actions
    void _installActions();

    //* play a song
    void _populatePlaylist();

    //* process command line arguments
    void _processArguments( const CommandLineArguments& );

    //* stop server
    void _stopServer();

    //* lock
    void _lock()
    { locked_ = true; }

    //* unlock
    void _unlock()
    { locked_ = false; }

    //* mpd controller
    Mpd::Controller* controller_ = nullptr;

    //* collection watcher
    Mpd::Thread* collectionWatcherThread_ = nullptr;

    //* cover manager
    Mpd::Thread* coverManagerThread_ = nullptr;

    //* album information
    Mpd::Thread* albumInformationManagerThread_ = nullptr;

    //* lyrics provider
    Mpd::Thread* lyricsManagerThread_ = nullptr;

    //* audio stream interface
    Mpd::Thread* audioStreamThread_ = nullptr;

    #if WITH_CDIO_PARANOIA
    //* audio device thread
    Mpd::Thread* audioDeviceThread_ = nullptr;
    #endif

    //* mpd interface
    Mpd::Thread* interfaceThread_ = nullptr;

    //* mpd interface
    Mpd::Thread* idleThread_ = nullptr;

    //* main window
    MainWindow* mainWindow_ = nullptr;

    //* system notifications
    SystemNotifications* systemNotifications_ = nullptr;

    //* icon cache (for notifications)
    CoverCache::Pointer coverCache_;

    //* system tray icon
    std::unique_ptr<SystemTrayIcon> systemTrayIcon_;

    //* locked
    bool locked_ = false;

    //*@name actions
    //@{

    //* show main window
    QAction* showMainWindowAction_ = nullptr;

    //* restart mpd server
    QAction* restartServerAction_ = nullptr;

    //* mpd server information
    QAction* aboutMpdAction_ = nullptr;

    //@}
};

#endif
