/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AudioStreamInterface.h"
#include "AudioStreamInterface_p.h"

#include "CppUtil.h"
#include "QUrlUtil.h"
#include "XmlError.h"

#if WITH_ZLIB
#include <zlib.h>
#endif

#include <QSet>

//* hard coded provider strings
const QString AudioStreamInterface::iceCastUrl = QLatin1String( "http://dir.xiph.org/yp.xml" );
const QString AudioStreamInterface::shoutCastHost = QLatin1String( "api.shoutcast.com" );
const QString AudioStreamInterface::shoutCastApiKey = QLatin1String( "fa1669MuiRPorUBw" );

namespace
{

    namespace Local
    {

        //* move all genres that have only one channel to "various"
        void _cleanChannelList( AudioStreamItem::List& channels )
        {
            // sort
            bool modified( false );
            std::sort( channels.begin(), channels.end() );
            for( auto iter = channels.begin(); iter != channels.end(); )
            {

                const auto genre( iter->genre() );
                auto next = iter+1;
                if( next == channels.end() || next->genre() != iter->genre() )
                {
                    iter->setGenre( QObject::tr( "Various" ) );
                    modified = true;
                }

                iter = std::find_if_not( iter, channels.end(), AudioStreamItem::SameGenreFTor( genre ) );

            }

            // sort again
            if( modified )
            { std::sort( channels.begin(), channels.end() ); }
        }

        //* get genre list from channels
        AudioStreamItem::List _genres( const AudioStreamItem::List& channels )
        {

            AudioStreamItem::List genres;
            for( auto iter = channels.begin(); iter != channels.end(); iter = std::find_if_not( iter, channels.end(), AudioStreamItem::SameGenreFTor( *iter ) ) )
            {
                auto genreItem = *iter;
                genreItem.setType( AudioStreamItem::Type::Genre );
                genreItem.setChannel( QString() );
                genreItem.setUrl( QString() );
                genres.append( genreItem );
            }

            return genres;
        }

    }

}


namespace Private
{

    //_______________________________________________________________________________
    AudioStreamJob::AudioStreamJob( const AudioStreamItem& item, QNetworkReply* networkReply ):
        NetworkJob( networkReply ),
        parentItem_( item )
    { Debug::Throw( "Private::AudioStreamJob::AudioStreamJob.\n" ); }

    //_______________________________________________________________________________
    void AudioStreamJob::_finished()
    {
        Debug::Throw( "Private::AudioStreamJob::_finished.\n" );
        SignalEmiter( this );

        // check reply and status
        if( !( _hasNetworkReply() && _networkReply().error() == QNetworkReply::NoError ) )
        { return; }

        switch( parentItem_.providerType() )
        {
            default: break;

            case AudioStreamItem::ProviderType::IceCast:
            if( parentItem_.type() == AudioStreamItem::Type::Provider )
            { _processIceCastReply(); }
            break;

            case AudioStreamItem::ProviderType::ShoutCast:
            if( parentItem_.type() == AudioStreamItem::Type::Channel ) _processShoutCastChannelReply();
            else _processShoutCastReply();
            break;

        }

        return;

    }

    //_______________________________________________________________________________
    void AudioStreamJob::_processIceCastReply()
    {
        Debug::Throw( "Private::AudioStreamJob::_processIceCastReply.\n" );

        // read data
        #if WITH_ZLIB
        const auto data( _tryUncompress( _networkReply() ) );
        #else
        const auto data( _networkReply().readAll() );
        #endif

        // create dom document
        QDomDocument document;
        XmlError error;
        if( !document.setContent( data, error ) ) return;

        // channels
        AudioStreamItem::List channels;

        // genres
        QSet<QString> genreNames;

        // channels
        QSet<QString> channelNames;

        // parse document
        const auto nodeList( document.elementsByTagName( "entry" ) );
        for( int i = 0; i < nodeList.count(); ++i )
        {

            // get element
            const auto element( nodeList.at( i ).toElement() );

            // create item
            AudioStreamItem item( AudioStreamItem::ProviderType::IceCast, AudioStreamItem::Type::Channel );
            item.setProvider( parentItem_.provider() );

            // genres
            QSet<QString> genreList;

            // parse element children
            for( auto&& childNode = element.firstChild(); !childNode.isNull(); childNode = childNode.nextSibling() )
            {
                const auto childElement = childNode.toElement();
                if( childElement.isNull() ) continue;

                const auto tagName( childElement.tagName() );
                if( tagName == QLatin1String( "server_name" ) ) item.setChannel( childElement.text() );
                else if( tagName == QLatin1String( "listen_url" ) ) item.setUrl( childElement.text() );
                else if( tagName == QLatin1String( "genre" ) )
                { genreList = _parseGenres( childElement.text() ); }
            }

            if( channelNames.contains( item.channel() ) ) continue;
            else channelNames.insert( item.channel() );

            for( const auto& genre:genreList )
            {
                genreNames.insert( genre );
                item.setGenre( genre );
                channels.append( item );
            }

        }

        Debug::Throw() << "Private::AudioStreamJob::_processIceCastReply - genres: " << genreNames.size() << endl;
        Debug::Throw() << "Private::AudioStreamJob::_processIceCastReply - channels: " << channels.size() << endl;

        Local::_cleanChannelList( channels );

        // generate genres
        if( !channels.empty() ) {

            // add genres
            AudioStreamItem::List genres( Local::_genres( channels ) );
            if( !genres.empty() ) emit childrenFound( parentItem_, genres );

            // add channels
            emit childrenFound( parentItem_, channels );

        }

    }

    //_______________________________________________________________________________
    void AudioStreamJob::_processShoutCastReply()
    {
        Debug::Throw( "Private::AudioStreamJob::_processShoutCastReply.\n" );

        const auto data( _networkReply().readAll() );

        // create dom document
        QDomDocument document;
        XmlError error;
        if( !document.setContent( data, error ) ) return;

        // channels
        AudioStreamItem::List items;

        // parse genres from document
        {
            const auto nodeList( document.elementsByTagName( "genre" ) );
            for( int i = 0; i < nodeList.count(); ++i )
            {

                // create item
                AudioStreamItem item( AudioStreamItem::ProviderType::ShoutCast, AudioStreamItem::Type::Genre );
                item.setProvider( parentItem_.provider() );

                if( parentItem_.type() == AudioStreamItem::Type::Genre ) item.setParentGenre( parentItem_.genre() );

                // genre id
                // it is needed for generating the proper url
                int genreId = -1;

                // parse attributes
                const auto element( nodeList.at( i ).toElement() );
                const auto attributes( element.attributes() );
                for( int i=0; i<attributes.count(); i++ )
                {
                    const auto attribute( attributes.item( i ).toAttr() );
                    if( attribute.name() == QLatin1String( "name" ) ) item.setGenre( attribute.value() );
                    else if( attribute.name() == QLatin1String( "id" ) ) genreId = attribute.value().toInt();
                }

                if( genreId < 0 ) continue;

                // generate url
                item.setUrl( QLatin1String("http://")+AudioStreamInterface::shoutCastHost+QString("/genre/secondary?parentid=%1").arg( genreId )+QLatin1String("&f=xml&k=")+AudioStreamInterface::shoutCastApiKey );

                items.append( item );

            }
        }

        // also parse stations
        {
            // parse tuneIn base
            QString base;
            const auto tuneInNodeList( document.elementsByTagName( "tunein" ) );
            for( int i = 0; i < tuneInNodeList.count() && base.isEmpty(); ++i )
            {

                const auto element( tuneInNodeList.at( i ).toElement() );
                const auto attributes( element.attributes() );
                for( int i=0; i<attributes.count(); i++ )
                {
                    const auto attribute( attributes.item( i ).toAttr() );
                    if( attribute.name() == QLatin1String( "base" ) )
                    {
                        base = attribute.value();
                        break;
                    }
                }
            }

            // use default if not found
            if( base.isEmpty() ) base = QLatin1String( "/sbin/tunein-station.pls" );

            // parse stations
            const auto nodeList( document.elementsByTagName( "station" ) );
            for( int i = 0; i < nodeList.count(); ++i )
            {

                // create item
                AudioStreamItem item( AudioStreamItem::ProviderType::ShoutCast, AudioStreamItem::Type::Channel );
                item.setProvider( parentItem_.provider() );
                if( parentItem_.type() == AudioStreamItem::Type::Genre ) item.setGenre( parentItem_.genre() );

                // it is needed for generating the proper url
                int stationId = -1;

                // parse attributes
                const auto element( nodeList.at( i ).toElement() );
                const auto attributes( element.attributes() );
                for( int i=0; i<attributes.count(); i++ )
                {
                    const auto attribute( attributes.item( i ).toAttr() );
                    if( attribute.name() == QLatin1String( "name" ) ) item.setChannel( attribute.value() );
                    else if( attribute.name() == QLatin1String( "id" ) ) stationId = attribute.value().toInt();
                }

                if( stationId < 0 ) continue;

                // build url
                item.setUrl( QString( "http://yp.shoutcast.com%1?id=%2" ).arg( base ).arg( stationId ) );
                items.append( item );

            }
        }

        // emit signals
        if( !items.empty() )
        {
            std::sort( items.begin(), items.end() );
            emit childrenFound( parentItem_, items );
        }

    }

    //_______________________________________________________________________________
    void AudioStreamJob::_processShoutCastChannelReply()
    {

        Debug::Throw( "Private::AudioStreamJob::_processShoutCastChannelReply.\n" );
        const QString data( _networkReply().readAll() );

        // split by line
        const auto lines( data.split( "\n" ) );
        for( const auto& line:lines )
        {

            // parse key/value separated by '='
            const auto position( line.indexOf( '=' ) );
            if( position <= 0 ) continue;

            const auto string = line.left( position );
            const auto value = line.mid( position+1 );
            if( string == QLatin1String( "File1" ) )
            { emit audioStreamUpdated( parentItem_, value ); }
        }

    }

    //________________________________________________________
    QSet<QString> AudioStreamJob::_parseGenres( const QString& genreString )
    {
        QSet<QString> genreList;
        for( auto genre:genreString.split( ' ', QString::SkipEmptyParts ) )
        {
            if( genre.size() < 2 ) continue;
            if( genre.contains( QRegExp("^#x[0-9a-f][0-9a-f]") ) ) continue;

            // capitalize first letter
            genre[0] = genre[0].toUpper();

            // Convert 80 -> 80s.
            if( genre.contains( QRegExp("^[0-9]0$") ) ) genre += 's';

            genreList.insert( genre );

        }

        if( genreList.isEmpty() ) genreList.insert( tr( "Other" ) );
        return genreList;
    }

    //________________________________________________
    QByteArray AudioStreamJob::_tryUncompress( QIODevice& in )
    {

        QByteArray out;
        if( !( in.isOpen() && in.isReadable() ) ) return out;

        #if WITH_ZLIB

        // initialize zlib stream
        z_stream stream;
        stream.zalloc = nullptr;
        stream.zfree = nullptr;
        stream.opaque = nullptr;
        stream.avail_in = 0;
        stream.next_in = nullptr;

        if( inflateInit2(&stream, MAX_WBITS + 16) != Z_OK) return out;

        // read chunks of max size chunksize from input source
        const int chunkSize = 16384;
        forever
        {

            // read chunk from input and store in stream
            auto source = in.read( chunkSize );
            if( source.isEmpty() )
            {
                out.chop(stream.avail_out);
                break;
            }

            stream.avail_in = source.size();
            stream.next_in = reinterpret_cast<Bytef*>(source.data());

            // uncompress chunk and append to output stream
            int result = Z_OK;
            do
            {

                // prepare destination buffer
                auto size = out.size();
                out.resize( size+chunkSize );
                stream.avail_out = chunkSize;
                stream.next_out = reinterpret_cast<Bytef*>(out.data() + size);

                result = inflate(&stream, Z_NO_FLUSH);
                switch( result )
                {
                    case Z_NEED_DICT:
                    case Z_DATA_ERROR:
                    case Z_STREAM_ERROR:
                    case Z_MEM_ERROR:
                    {
                        inflateEnd( &stream );
                        out.chop(stream.avail_out);
                        return out;
                    }
                }


            } while( stream.avail_out == 0 );

            // Chop off trailing space in the buffer
            out.chop(stream.avail_out);

            if( result == Z_STREAM_END )
            {
                // Make sure there are no more members to process before exiting
                if( !(stream.avail_in && inflateReset(&stream) == Z_OK ) ) break;
            }

        }

        inflateEnd(&stream);
        #endif
        return out;

    }

}

//________________________________________________________
AudioStreamConverter::AudioStreamConverter( QObject* parent, const AudioStreamItem::List& items, int row ):
    QObject( parent ),
    Counter( "AudioStreamConverter" ),
    networkAccessManager_( new QNetworkAccessManager( this ) ),
    items_( items ),
    row_( row )
{ Debug::Throw( "AudioStreamConverter::AudioStreamConverter.\n" ); }

//________________________________________________________
void AudioStreamConverter::run()
{
    Debug::Throw( "AudioStreamConverter::run.\n" );

    bool completed( true );
    for( auto&& item:items_ )
    {

        if( item.type() == AudioStreamItem::Type::Channel && item.providerType() == AudioStreamItem::ProviderType::ShoutCast )
        {

            completed = false;

            // build request to fetch actual channel
            QNetworkRequest networkRequest( item.url() );

            auto audioStreamJob = new Private::AudioStreamJob( item, networkAccessManager_->get( networkRequest ) );
            connect( audioStreamJob, SIGNAL(audioStreamUpdated(AudioStreamItem,QString)), this, SLOT(_updateAudioStream(AudioStreamItem,QString)) );
            connect( audioStreamJob, SIGNAL(finished()), audioStreamJob, SLOT(deleteLater()) );

        } else {

            item.setConverted( true );

        }

    }

    if( completed )
    {
        emit finished( items_, row_ );
        emit finished();
    }

}

//________________________________________________________
void AudioStreamConverter::_updateAudioStream( const AudioStreamItem& item, const QString& url )
{
    Debug::Throw( "AudioStreamConverter::_updateAudioStream.\n" );

    auto iter = std::find_if(
        items_.begin(), items_.end(),
        AudioStreamItem::SameUrlFTor( item ) );

    if( iter != items_.end() )
    {
        iter->setUrl( url );
        iter->setConverted( true );
    }

    if( std::all_of( items_.begin(), items_.end(), AudioStreamItem::IsConvertedFTor() ) )
    {
        emit finished( items_, row_ );
        emit finished();
    }

}

//________________________________________________________
AudioStreamInterface::AudioStreamInterface( QObject* parent ):
    QObject( parent ),
    Counter( "AudioStreamInterface" ),
    networkAccessManager_( new QNetworkAccessManager( this ) )
{ Debug::Throw( "AudioStreamInterface::AudioStreamInterface.\n" ); }

//________________________________________________________
void AudioStreamInterface::_findIceCastChildren( const AudioStreamItem& item )
{
    Debug::Throw( "AudioStreamInterface::_findIceCastChildren.\n" );

    // build request
    QNetworkRequest networkRequest( item.url() );

    #if WITH_ZLIB
    networkRequest.setRawHeader( "Accept-Encoding", "gzip" );
    #endif

    // create job
    auto audioStreamJob = new Private::AudioStreamJob( item, networkAccessManager_->get( networkRequest ) );
    connect( audioStreamJob, SIGNAL(finished()), audioStreamJob, SLOT(deleteLater()) );
    connect( audioStreamJob, SIGNAL(childrenFound(AudioStreamItem,AudioStreamItem::List)), this, SIGNAL(childrenFound(AudioStreamItem,AudioStreamItem::List)) );
}

//________________________________________________________
void AudioStreamInterface::findChildren( const AudioStreamItem& item )
{

    Debug::Throw( "AudioStreamInterface::findChildren.\n" );

    // load from internet
    switch( item.providerType() )
    {

        default: break;

        case AudioStreamItem::ProviderType::IceCast:
        if( item.type() == AudioStreamItem::Type::Provider ) _findIceCastChildren( item );
        break;

        case AudioStreamItem::ProviderType::ShoutCast:
        if( item.type() == AudioStreamItem::Type::Provider  || item.type() == AudioStreamItem::Type::Genre ) _findShoutCastChildren( item );
        break;

    }

}

//________________________________________________________
void AudioStreamInterface::_findShoutCastChildren( const AudioStreamItem& item )
{
    Debug::Throw( "AudioStreamInterface::_findShoutCastChildren.\n" );

    {
        // build request to fetch genres
        QNetworkRequest networkRequest( item.url() );
        networkRequest.setHeader(QNetworkRequest::ContentTypeHeader, "application/x-www-form-urlencoded");

        // create job
        auto audioStreamJob = new Private::AudioStreamJob( item, networkAccessManager_->get( networkRequest ) );
        connect( audioStreamJob, SIGNAL(finished()), audioStreamJob, SLOT(deleteLater()) );
        connect( audioStreamJob, SIGNAL(childrenFound(AudioStreamItem,AudioStreamItem::List)), this, SIGNAL(childrenFound(AudioStreamItem,AudioStreamItem::List)) );

    }

    // for genre items, also asks for list of station
    if( item.type() == AudioStreamItem::Type::Genre )
    {

        QUrl url(QLatin1String("http://")+AudioStreamInterface::shoutCastHost+QLatin1String("/legacy/genresearch"));
        QUrlUtil::addQueryItems( url, Base::makeT<Base::QueryItemList>(
        {
            { QLatin1String( "k" ), AudioStreamInterface::shoutCastApiKey },
            { QLatin1String( "genre" ), item.genre() }
        } ) );

        QNetworkRequest networkRequest( url );
        networkRequest.setHeader(QNetworkRequest::ContentTypeHeader, "application/x-www-form-urlencoded");

        // create job
        auto audioStreamJob = new Private::AudioStreamJob( item, networkAccessManager_->get( networkRequest ) );
        connect( audioStreamJob, SIGNAL(finished()), audioStreamJob, SLOT(deleteLater()) );
        connect( audioStreamJob, SIGNAL(childrenFound(AudioStreamItem,AudioStreamItem::List)), this, SIGNAL(childrenFound(AudioStreamItem,AudioStreamItem::List)) );

    }

}
