/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AudioStreamItem.h"

#include "CppUtil.h"
#include "Debug.h"
#include "XmlDef.h"

//__________________________________________________________________
const QString AudioStreamItem::MimeType = "musicbox/audioStream-item-list";

//_______________________________________________________________________
AudioStreamItem::AudioStreamItem( const QDomElement& element ):
    Counter( "AudioStreamItem" ),
    type_( Type::Channel )
{

    Debug::Throw( "AudioStreamItem::AudioStreamItem (dom).\n" );

    // parse attributes
    const auto attributes( element.attributes() );
    for( int i=0; i<attributes.count(); i++ )
    {
        const auto attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;
        else if( attribute.name() == Xml::AudioStreamProviderType ) providerType_ = static_cast<ProviderType>( attribute.value().toInt() );
        else if( attribute.name() == Xml::AudioStreamItemType ) type_ = static_cast<Type>( attribute.value().toInt() );
        else if( attribute.name() == Xml::AudioStreamProvider ) provider_ = attribute.value();
        else if( attribute.name() == Xml::AudioStreamParentGenre ) parentGenre_ = attribute.value();
        else if( attribute.name() == Xml::AudioStreamGenre ) genre_ = attribute.value();
        else if( attribute.name() == Xml::AudioStreamChannel ) channel_ = attribute.value();
        else if( attribute.name() == Xml::AudioStreamAlias ) alias_ = attribute.value();
        else if( attribute.name() == Xml::AudioStreamUrl ) url_ = attribute.value();
    }

}

//__________________________________________________________________
bool AudioStreamItem::isChild( const AudioStreamItem& other ) const
{

    switch( type_ )
    {
        default:
        case Type::Invalid:
        case Type::Provider: return false;
        case Type::Genre:
        {
            if( parentGenre_.isEmpty() ) return other.type_ == Type::Provider && other.provider_ == provider_;
            else return other.type_ == Type::Genre && other.provider_ == provider_ && other.genre_ == parentGenre_;
        }

        case Type::Channel:
        {
            if( providerType_ == ProviderType::Favorites )
            {

                // for favorites, genre items are ignored
                return other.type_ == Type::Provider && other.providerType_ == providerType_;

            } else {

                return other.type_ == Type::Genre && other.provider_ == provider_ && other.genre_ == genre_;

            }
        }

    }

}

//__________________________________________________________________
QDomElement AudioStreamItem::domElement( QDomDocument& document ) const
{
    Debug::Throw( "AudioStreamItem::domElement.\n" );

    auto out( document.createElement( Xml::AudioStreamItem ) );
    out.setAttribute( Xml::AudioStreamProviderType, QString::number( Base::toIntegralType( providerType_ ) ) );
    out.setAttribute( Xml::AudioStreamItemType, QString::number( Base::toIntegralType( type_ ) ) );
    out.setAttribute( Xml::AudioStreamProvider, provider_ );
    if( !parentGenre_.isEmpty() ) out.setAttribute( Xml::AudioStreamParentGenre, parentGenre_ );
    if( !genre_.isEmpty() ) out.setAttribute( Xml::AudioStreamGenre, genre_ );
    if( !channel_.isEmpty() ) out.setAttribute( Xml::AudioStreamChannel, channel_ );
    if( !alias_.isEmpty() ) out.setAttribute( Xml::AudioStreamAlias, alias_ );

    if( !url_.isEmpty() ) out.setAttribute( Xml::AudioStreamUrl, url_ );
    return out;

}

//__________________________________________________________________
bool operator == (const AudioStreamItem& lhs, const AudioStreamItem& rhs )
{
    if( lhs.type_ != rhs.type_ ) return false;
    else if( lhs.providerType_ != rhs.providerType_ ) return false;
    else switch( lhs.type_ )
    {
        default: return false;
        case AudioStreamItem::Type::Provider: return lhs.provider_ == rhs.provider_;
        case AudioStreamItem::Type::Genre: return lhs.provider_ == rhs.provider_ && lhs.parentGenre_ == rhs.parentGenre_ && lhs.genre_ == rhs.genre_;
        case AudioStreamItem::Type::Channel: return lhs.provider_ == rhs.provider_ && lhs.genre_ == rhs.genre_ && lhs.channel_ == rhs.channel_ && lhs.alias_ == rhs.alias_;
    }
}

//__________________________________________________________________
bool operator < (const AudioStreamItem& lhs, const AudioStreamItem& rhs )
{
    if( lhs.type_ != rhs.type_ )
    {

        if( lhs.type_ == AudioStreamItem::Type::Provider ) return true;
        else if( lhs.type_ == AudioStreamItem::Type::Genre ) return rhs.type_ == AudioStreamItem::Type::Channel;
        else return false;

    } else switch( lhs.type_ ) {

        default: return false;
        case AudioStreamItem::Type::Provider: return lhs.provider_ < rhs.provider_;
        case AudioStreamItem::Type::Genre:
        {
            if( lhs.provider_ != rhs.provider_ ) return lhs.provider_ < rhs.provider_;
            else if( lhs.parentGenre_ != rhs.parentGenre_ ) return lhs.parentGenre_ < rhs.parentGenre_;
            else return lhs.genre_ < rhs.genre_;
        }

        case AudioStreamItem::Type::Channel:
        {
            if( lhs.providerType_ == rhs.providerType_ && lhs.providerType_ == AudioStreamItem::ProviderType::Favorites ) return lhs.aliasOrChannel() < rhs.aliasOrChannel();
            else if( lhs.provider_ != rhs.provider_ ) return lhs.provider_ < rhs.provider_;
            else if( lhs.genre_ != rhs.genre_ ) return lhs.genre_ < rhs.genre_;
            else return lhs.channel_ < rhs.channel_;
        }

    }

}

//_______________________________________________________________________
QTextStream& operator << ( QTextStream& out, const AudioStreamItem& item )
{
    out << "Type : " << Base::toIntegralType( item.type_ );
    out << " provider: " << item.provider_;
    switch( item.type_ )
    {
        default:
        break;

        case AudioStreamItem::Type::Genre:
        {
            if( !item.parentGenre_.isEmpty() ) out << " parent genre: " << item.parentGenre_;
            out << " genre: " << item.genre_;
            break;
        }

        case AudioStreamItem::Type::Channel:
        {
            out << " genre: " << item.genre_;
            out << " channel: " << item.channel_;
            break;
        }
    }

    return out;

};

//_______________________________________________________________________
AudioStreamItem::List AudioStreamItem::Helper::list( const QDomElement& parent )
{
    Debug::Throw( "AudioStreamItem::Helper::list.\n" );
    AudioStreamItem::List out;
    for( auto&& node = parent.firstChild(); !node.isNull(); node = node.nextSibling() )
    {
        auto element = node.toElement();
        if( element.isNull() ) continue;
        if( element.tagName() == Xml::AudioStreamItem )
        { out.append( AudioStreamItem( element ) ); }
    }

    return out;
}


//_______________________________________________________________________
AudioStreamItem::Set AudioStreamItem::Helper::set( const QDomElement& parent )
{
    Debug::Throw( "AudioStreamItem::Helper::set.\n" );
    AudioStreamItem::Set out;
    for( auto&& node = parent.firstChild(); !node.isNull(); node = node.nextSibling() )
    {
        auto element = node.toElement();
        if( element.isNull() ) continue;
        if( element.tagName() == Xml::AudioStreamItem )
        { out.insert( AudioStreamItem( element ) ); }
    }

    return out;
}
