#ifndef AudioStreamItem_h
#define AudioStreamItem_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "Debug.h"
#include "Functors.h"
#include "QOrderedSet.h"
#include "XmlDef.h"

#include <QDomElement>
#include <QDomDocument>
#include <QList>

class AudioStreamItem: private Base::Counter<AudioStreamItem>
{

    public:

    using List = QList<AudioStreamItem>;
    using Set = QOrderedSet<AudioStreamItem>;

    //* used when LogEntry keyword is not defined
    static const QString MimeType;

    //* default constructor
    explicit AudioStreamItem():
        Counter( "AudioStreamItem" )
        {}

    //* item type
    enum class Type
    {
        Invalid,
        Provider,
        Genre,
        Channel
    };

    //* provider type
    enum class ProviderType
    {
        Unknown,
        Favorites,
        Dirble,
        IceCast,
        ShoutCast,
        TuneIn
    };

    //* filled constructor
    explicit AudioStreamItem( ProviderType providerType, Type type, const QString& url = QString() ):
        Counter( "AudioStreamItem" ),
        providerType_( providerType ),
        type_( type ),
        url_( url )
    {}


    //* constructor from dom
    explicit AudioStreamItem( const QDomElement& );

    //*@name accessors
    //@{

    //* provider type
    ProviderType providerType() const
    { return providerType_; }

    //* item type
    Type type() const
    { return type_; }

    //* true if is children of other item
    bool isChild( const AudioStreamItem& ) const;

    //* provider name
    const QString& provider() const
    { return provider_; }

    //* parent genre name
    const QString& parentGenre() const
    { return parentGenre_; }

    //* genre name
    const QString& genre() const
    { return genre_; }

    //* channel name
    const QString& channel() const
    { return channel_; }

    //* channel alias name
    const QString& alias() const
    { return alias_; }

    //* channel alias or name
    const QString& aliasOrChannel() const
    { return alias_.isEmpty() ? channel_:alias_; }

    //* url
    const QString& url() const
    { return url_; }

    //* converted
    bool isConverted() const
    { return converted_; }

    //* dom element
    QDomElement domElement( QDomDocument& ) const;

    //@}

    //*@name modifiers
    //@{

    //* provider type
    void setProviderType( ProviderType value )
    { providerType_ = value; }

    //* item type
    void setType( Type value )
    { type_ = value; }

    //* provider
    void setProvider( const QString& value )
    { provider_ = value; }

    //* parent genre
    void setParentGenre( const QString& value )
    { parentGenre_ = value; }

    //* genre
    void setGenre( const QString& value )
    { genre_ = value; }

    //* channel
    void setChannel( const QString& value )
    { channel_ = value; }

    //* alias
    void setAlias( const QString& value )
    { alias_ = value; }

    //* url
    void setUrl( const QString& value )
    { url_ = value; }

    //* converted
    void setConverted( bool value )
    { converted_ = value; }

    //@}

    using SameUrlFTor = Base::Functor::UnaryEqual<AudioStreamItem, const QString&, &AudioStreamItem::url>;
    using SameGenreFTor = Base::Functor::UnaryEqual<AudioStreamItem, const QString&, &AudioStreamItem::genre>;
    using SameProviderTypeFTor = Base::Functor::UnaryEqual<AudioStreamItem, ProviderType, &AudioStreamItem::providerType>;
    using SameTypeFTor = Base::Functor::UnaryEqual<AudioStreamItem, Type, &AudioStreamItem::type>;
    using DifferentTypeFTor = Base::Functor::UnaryDiffer<AudioStreamItem, Type, &AudioStreamItem::type>;
    using IsConvertedFTor = Base::Functor::UnaryTrue<AudioStreamItem, &AudioStreamItem::isConverted>;

    //* helper class to parse List from string
    class Helper
    {
        public:

        //* get list from dom element
        static AudioStreamItem::List list( const QDomElement& );

        //* get list from dom element
        static AudioStreamItem::Set set( const QDomElement& );

        //* get dom element from list
        template< class T>
        static QDomElement domElement( const T&, QDomDocument& );

    };

    private:

    //* provider type
    ProviderType providerType_ = ProviderType::Unknown;

    //* type
    Type type_ = Type::Invalid;

    //* provider
    QString provider_;

    //* parent genre
    QString parentGenre_;

    //* genre
    QString genre_;

    //* channel
    QString channel_;

    //* channel alias
    QString alias_;

    //* url
    QString url_;

    //* true when url is converted to valid stream
    bool converted_ = false;

    //* comparison operators
    friend bool operator == (const AudioStreamItem& lhs, const AudioStreamItem& rhs );
    friend bool operator < (const AudioStreamItem& lhs, const AudioStreamItem& rhs );

    //* streamer
    friend QTextStream& operator << (QTextStream&, const AudioStreamItem& );

};

//_______________________________________________________________________
template< class T >
QDomElement AudioStreamItem::Helper::domElement( const T& AudioStreamItemList, QDomDocument& parent )
{
    Debug::Throw( "AudioStreamItem::Helper::domElement.\n" );

    // create main element
    auto out( parent.createElement( Xml::AudioStreamItemList ) );
    for( const auto& AudioStreamItem:AudioStreamItemList )
    { out.appendChild( AudioStreamItem.domElement( parent ) ); }

    return out;
}

#endif
