/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AudioStreamItemToolTipWidget.h"

#include "Debug.h"
#include "GridLayout.h"
#include "GridLayoutItem.h"
#include "QtUtil.h"
#include "Singleton.h"
#include "TimeStamp.h"
#include "XmlOptions.h"

#include <QLayout>

//_______________________________________________________
AudioStreamItemToolTipWidget::AudioStreamItemToolTipWidget( QWidget* parent ):
    BaseToolTipWidget( parent )
{

    Debug::Throw( "AudioStreamItemToolTipWidget::AudioStreamItemToolTipWidget.\n" );

    setFollowMouse( true );

    // layout
    auto hLayout = new QHBoxLayout;
    hLayout->setMargin( 10 );
    hLayout->setSpacing( 10 );
    setLayout( hLayout );

    hLayout->addWidget( iconLabel_ = new QLabel( this ) );
    iconLabel_->setAlignment( Qt::AlignHCenter|Qt::AlignTop );

    auto vLayout = new QVBoxLayout;
    vLayout->setMargin( 0 );
    vLayout->setSpacing( 5 );
    hLayout->addLayout( vLayout );

    // grid layout
    vLayout->addStretch( 1 );
    auto gridLayout = new GridLayout;
    gridLayout->setMaxCount( 2 );
    gridLayout->setColumnAlignment( 0, Qt::AlignVCenter|Qt::AlignRight );
    gridLayout->setColumnAlignment( 1, Qt::AlignVCenter|Qt::AlignLeft );
    gridLayout->setMargin( 0 );
    gridLayout->setSpacing( 5 );
    vLayout->addLayout( gridLayout );

    // items
    ( providerItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Provider:" ) );
    ( parentGenreItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Parent genre:" ) );
    ( genreItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Genre:" ) );
    ( channelItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Channel:" ) );
    ( aliasItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Alias:" ) );
    ( urlItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Url:" ) );
    ( entriesItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Entries:" ) );

    // add stretch
    vLayout->addStretch( 1 );

    // configuration
    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );

}

//_____________________________________________
void AudioStreamItemToolTipWidget::_updateConfiguration()
{
    Debug::Throw( "AudioStreamItemToolTipWidget::_updateConfiguration.\n" );
    _setPixmapSize( 5*(fontMetrics().height() + 5) );
}

//_____________________________________________
void AudioStreamItemToolTipWidget::setAudioStreamItem( const AudioStreamItem& item, const QIcon& icon, int entries )
{
    Debug::Throw( "AudioStreamItemToolTipWidget::setAudioStream.\n" );
    audioStreamItem_ = item;
    icon_ = icon;
    entries_ = entries;

    // update icon
    if( !icon.isNull() )
    {

        // get pixmap
        iconLabel_->setPixmap( icon.pixmap( QSize( pixmapSize_, pixmapSize_ ) ) );
        iconLabel_->show();

    } else iconLabel_->hide();

    switch( item.type() )
    {
        case AudioStreamItem::Type::Provider:
        {
            providerItem_->setText( item.provider() );
            if( !item.url().isEmpty() ) urlItem_->setText( item.url() );
            else urlItem_->hide();
            if( entries > 0 ) entriesItem_->setText( QString::number( entries ) );
            else entriesItem_->hide();
            parentGenreItem_->hide();
            genreItem_->hide();
            channelItem_->hide();
            aliasItem_->hide();
            break;
        }

        case AudioStreamItem::Type::Genre:
        {
            providerItem_->setText( item.provider() );
            if( entries > 0 ) entriesItem_->setText( QString::number( entries ) );
            else entriesItem_->hide();
            if( !item.parentGenre().isEmpty() ) parentGenreItem_->setText( item.parentGenre() );
            else parentGenreItem_->hide();
            genreItem_->setText( item.genre() );
            channelItem_->hide();
            aliasItem_->hide();
            urlItem_->hide();
            break;
        }

        case AudioStreamItem::Type::Channel:
        {
            providerItem_->setText( item.provider() );
            parentGenreItem_->hide();
            genreItem_->setText( item.genre() );
            channelItem_->setText( item.channel() );
            if( !item.alias().isEmpty() ) aliasItem_->setText( item.alias() );
            else aliasItem_->hide();
            urlItem_->setText( item.url() );
            entriesItem_->hide();
            break;
        }

        default:
        {
            providerItem_->hide();
            genreItem_->hide();
            channelItem_->hide();
            aliasItem_->hide();
            urlItem_->hide();
            entriesItem_->hide();
            break;
        }

    }

    adjustSize();

}
