/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AudioStreamModel.h"

#include "CppUtil.h"
#include "IconEngine.h"
#include "IconNames.h"

#include <QMimeData>

//_______________________________________________________________________________________
AudioStreamModel::AudioStreamModel( QObject* parent ):
    TreeModel( parent ),
    Counter( "AudioStreamModel" )
{ Debug::Throw( "AudioStreamModel::AudioStreamModel.\n" ); }

//_______________________________________________________________________________________
Qt::ItemFlags AudioStreamModel::flags( const QModelIndex& index ) const
{
    if( !index.isValid() ) return static_cast<Qt::ItemFlags>(0);

    const auto& item( _find( index.internalId() ).get() );
    switch( item.type() )
    {
        default:
        case AudioStreamItem::Type::Provider:
        if( item.providerType() == AudioStreamItem::ProviderType::Favorites )
        {

            return Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsDropEnabled;

        } else return Qt::ItemIsEnabled | Qt::ItemIsSelectable;

        case AudioStreamItem::Type::Genre:
        return Qt::ItemIsEnabled | Qt::ItemIsSelectable;

        case AudioStreamItem::Type::Channel:
        if( item.providerType() == AudioStreamItem::ProviderType::Favorites )
        {

            return Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsDragEnabled | Qt::ItemIsEditable;

        } else return Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsDragEnabled;

    }
}

//__________________________________________________________________
QStringList AudioStreamModel::mimeTypes() const
{
    return Base::makeT<QStringList>( { AudioStreamItem::MimeType } );
}

//______________________________________________________________________
QMimeData* AudioStreamModel::mimeData(const QModelIndexList &indexes) const
{
    if( indexes.empty() ) return nullptr;
    auto mime = new QMimeData;
    mime->setData( AudioStreamItem::MimeType, nullptr );
    return mime;
}

//__________________________________________________________________
QVariant AudioStreamModel::data( const QModelIndex& index, int role ) const
{

    if( !index.isValid() ) return QVariant();

    switch( role )
    {
        case Qt::DisplayRole:
        {

            const auto& item( _find( index.internalId() ).get() );
            switch( item.type() )
            {
                case AudioStreamItem::Type::Provider: return item.provider();
                case AudioStreamItem::Type::Genre: return item.genre();
                case AudioStreamItem::Type::Channel: return item.aliasOrChannel();
                default: return QVariant();
            }
        }

        case Qt::DecorationRole:
        return icon( _find( index.internalId() ).get() );

        default: return QVariant();
    }

}

//__________________________________________________________________
QVariant AudioStreamModel::headerData(int section, Qt::Orientation orientation, int role) const
{

    switch( role )
    {
        case Qt::DisplayRole:
        {
            switch( section )
            {
                case Name: return tr( "Name" );
                default: return QVariant();
            }
            break;
        }

        default: return QVariant();
    }

    // return empty
    return QVariant();

}

//__________________________________________________________________
bool AudioStreamModel::hasChildren( const QModelIndex& parent ) const
{
    if( !parent.isValid() ) return TreeModel::hasChildren( parent );

    const auto item( get( parent ) );
    return item.type() == AudioStreamItem::Type::Provider || item.type() == AudioStreamItem::Type::Genre;
}

//__________________________________________________________________
QIcon AudioStreamModel::icon( const AudioStreamItem& item ) const
{
    switch( item.type() )
    {
        case AudioStreamItem::Type::Provider: return IconEngine::get( IconNames::AudioStreamProvider );
        case AudioStreamItem::Type::Genre: return IconEngine::get( IconNames::AudioStreamGenre );
        case AudioStreamItem::Type::Channel: return IconEngine::get( IconNames::AudioStreamChannel );
        default: return QIcon();
    }

}

//__________________________________________________________________
bool AudioStreamModel::setData(const QModelIndex &index, const QVariant& value, int role )
{
    Debug::Throw( "AudioStreamModel::setData.\n" );

    // check index
    if( !(index.isValid() && role == Qt::EditRole ) ) return false;

    // get item and check
    auto item( get( index ) );
    if( !( item.type() == AudioStreamItem::Type::Channel && item.providerType() == AudioStreamItem::ProviderType::Favorites ) )
    { return false; }

    const auto alias( value.toString() );
    if( alias != item.alias() )
    {

        const auto copy( item );
        item.setAlias( alias );

        replace( copy, item );

        emit audioStreamRenamed( copy, item );

        return true;

    } else return false;

}

//____________________________________________________________
bool AudioStreamModel::dropMimeData(const QMimeData* data , Qt::DropAction action, int row, int, const QModelIndex& parent )
{

    // drop from current playlist
    if( data->hasFormat( AudioStreamItem::MimeType ) )
    {

        emit requestAddToFavorites();
        return true;

    } else return false;
}
