/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AudioStreamWidget.h"

#include "AudioStreamFindWidget.h"
#include "AudioStreamInterface.h"
#include "AudioStreamItemToolTipWidget.h"
#include "AudioStreamView.h"
#include "ContextMenu.h"
#include "CppUtil.h"
#include "Debug.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "NewAudioStreamDialog.h"
#include "QUrlUtil.h"
#include "QuestionDialog.h"
#include "ScrollBarMonitor.h"
#include "Singleton.h"
#include "XmlError.h"
#include "XmlOptions.h"

#include <QDir>
#include <QDomDocument>
#include <QFile>
#include <QLayout>

#include <algorithm>

#if WITH_ZLIB
#include <zlib.h>
#endif

//________________________________________________________
AudioStreamWidget::AudioStreamWidget( QWidget* parent ):
    QWidget( parent ),
    Counter( "AudioStreamWidget" )
{
    Debug::Throw( "AudioStreamWidget::AudioStreamWidget.\n" );

    _installActions();

    auto vLayout = new QVBoxLayout;
    vLayout->setMargin(0);
    vLayout->setSpacing(2);
    setLayout( vLayout );

    vLayout->addWidget( findWidget_ = new AudioStreamFindWidget( this ) );

    proxyModel_.setSourceModel( &model_ );
    connect( findWidget_, SIGNAL(requestSearchAudioStreams(QString)), this, SLOT(_setFilter(QString)) );

    vLayout->addWidget( list_ = new AudioStreamView( this ) );
    list_->setModel( &proxyModel_ );
    list_->setOptionName( "AudioStream" );

    scrollBarMonitor_ = new ScrollBarMonitor( list_ );

    // tooltip widget
    toolTipWidget_ = new AudioStreamItemToolTipWidget( this );

    connect( list_, SIGNAL(activated(QModelIndex)), SLOT(_itemActivated(QModelIndex)) );
    connect( list_, SIGNAL(expanded(QModelIndex)), SLOT(_updateAudioStreams(QModelIndex)) );
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateActions()) );
    connect( list_, SIGNAL(hovered(QModelIndex)), SLOT(_showToolTip(QModelIndex)) );

    list_->addAction( removeFromFavoritesAction_ );

    connect( &model_, SIGNAL(requestAddToFavorites()), SLOT(_addSelectionToFavorites()) );

    {
        // context menu
        auto menu = new ContextMenu( list_ );
        menu->addAction( updateAudioStreamsAction_ );
        menu->addAction( addToFavoritesAction_ );
        menu->addAction( removeFromFavoritesAction_ );
        menu->addAction( renameAudioStreamAction_ );
        menu->addAction( addToPlaylistAction_ );
        menu->addAction( replacePlaylistAction_ );
        menu->addAction( newAudioStreamAction_ );
        menu->addSeparator();
        menu->addAction( &list_->collapseAllAction() );
        menu->setHideDisabledActions( true );

        connect( menu, SIGNAL(aboutToShow()), SLOT(_updateActions()) );
    }

    // install providers
    _installProviders();

    // read favorites
    _readFavorites();
    loadedItems_.insert( favoritesItem_ );

    // connections
    connect( Base::Singleton::get().application(), SIGNAL(saveConfiguration()), SLOT(_saveConfiguration()) );

}

//________________________________________________________
QModelIndexList AudioStreamWidget::selectedRows() const
{
    auto indexes( list_->selectionModel()->selectedRows() );
    for( auto&& index:indexes ) { index = proxyModel_.mapToSource( index ); }
    return indexes;
}

//________________________________________________________
void AudioStreamWidget::saveExpandedItems()
{
    Debug::Throw( "AudioStreamWidget::saveExpandedItems.\n" );
    expandedItems_ = _expandedItems();
}

//________________________________________________________
void AudioStreamWidget::restoreExpandedItems()
{
    Debug::Throw( "AudioStreamWidget::restoreExpandedItems.\n" );
    for( const auto& item:expandedItems_ )
    {
        const auto index( model_.index( item ) );
        if( index.isValid() ) list_->setExpanded( index, true );
    }

    expandedItems_.clear();
}

//________________________________________________________
void AudioStreamWidget::lock()
{
    locked_ = true;
    list_->setUpdatesEnabled( false );
    scrollBarMonitor_->save();
}

//________________________________________________________
void AudioStreamWidget::unlock()
{
    scrollBarMonitor_->restore();
    list_->setUpdatesEnabled( true );
    locked_ = false;
}

//________________________________________________________
void AudioStreamWidget::addChildren( const AudioStreamItem& parent, AudioStreamItem::List children )
{
    Debug::Throw( "AudioStreamWidget::addChildren.\n" );
    lock();
    saveExpandedItems();

    // clear
    if( children.empty() ) model_.clear( parent );

    // add children
    model_.add( parent, children );

    restoreExpandedItems();
    unlock();
}

//________________________________________________________
void AudioStreamWidget::addStreamsToFavorites( const AudioStreamItem::List& constItems )
{

    Debug::Throw( "AudioStreamWidget::addStreamsToFavorites.\n" );

    // copy
    auto items( constItems );

    // change provider type
    for( auto&& item:items )
    { item.setProviderType( AudioStreamItem::ProviderType::Favorites ); }

    // re-add in model
    lock();
    model_.add( favoritesItem_, items );
    unlock();

}

//________________________________________________________
void AudioStreamWidget::_saveConfiguration()
{
    Debug::Throw( "AudioStreamWidget::_saveConfiguration.\n" );

    for( const auto& item:loadedItems_ )
    { _writeAudioStreams( item ); }
}

//________________________________________________________
void AudioStreamWidget::_setFilter( const QString & value )
{
    Debug::Throw( "AudioStreamWidget::_setFilter.\n" );
    if( value.isEmpty() ) proxyModel_.setFilterRegExp( QString() );
    else proxyModel_.setFilterRegExp( QRegExp( value, Qt::CaseInsensitive, QRegExp::FixedString ) );
}

//________________________________________________________
void AudioStreamWidget::_itemActivated( const QModelIndex& index )
{
    Debug::Throw()
        << "AudioStreamWidget::_itemActivated -"
        << " index: " << index.row() << " " << index.column()
        << endl;
}


//________________________________________________________
void AudioStreamWidget::_updateAudioStreams()
{
    Debug::Throw( "AudioStreamWidget::_updateAudioStreams.\n" );

    const auto selectedIndexes( selectedRows() );
    const auto selection( model_.get( selectedIndexes ) );
    const int providerCount( std::count_if( selection.begin(), selection.end(),
        []( const AudioStreamItem& item )
        { return item.type() == AudioStreamItem::Type::Provider && item.providerType() != AudioStreamItem::ProviderType::Favorites; } ) );
    if( !providerCount ) return;

    QuestionDialog dialog( this, providerCount > 1 ? tr( "Update audioStream lists ?"):tr( "Update audioStream List ?" ) );
    dialog.okButton().setText( tr( "Update" ) );
    dialog.okButton().setIcon( IconEngine::get( IconNames::Reload ) );
    if( !dialog.exec() ) return;

    for( const auto& index:selectedIndexes )
    {
        const auto& item( model_.get( index ) );
        if( item.type() == AudioStreamItem::Type::Provider && item.providerType() != AudioStreamItem::ProviderType::Favorites )
        { _doUpdateAudioStreams( index ); }
    }
}

//________________________________________________________
void AudioStreamWidget::_newAudioStream()
{
    Debug::Throw( "AudioStreamWidget::_newAudioStream.\n" );
    NewAudioStreamDialog dialog( this );
    dialog.setUrlHandlers( urlHandlers_ );
    if( !dialog.exec() ) return;

    // create audio stream item
    AudioStreamItem item( AudioStreamItem::ProviderType::Favorites, AudioStreamItem::Type::Channel, dialog.url() );
    item.setAlias( dialog.name() );

    // add to model
    lock();

    model_.add( favoritesItem_, item );
    unlock();

}

//________________________________________________________
void AudioStreamWidget::_addSelectionToPlaylist()
{
    Debug::Throw( "AudioStreamWidget::_addSelectionToPlaylist.\n" );

    // get selected channels
    auto items( selectedItems() );

    // erase everything that is not a Channel
    items.erase( std::remove_if( items.begin(), items.end(), AudioStreamItem::DifferentTypeFTor( AudioStreamItem::Type::Channel ) ), items.end() );

    // emit signal
    if( !items.empty() ) emit requestAddToPlaylist( items );

}

//________________________________________________________
void AudioStreamWidget::_replacePlaylist()
{
    Debug::Throw( "AudioStreamWidget::_replacePlaylist.\n" );

    // get selected channels
    auto items( selectedItems() );

    // erase everything that is not a Channel
    items.erase( std::remove_if( items.begin(), items.end(), AudioStreamItem::DifferentTypeFTor( AudioStreamItem::Type::Channel ) ), items.end() );

    // emit signal
    if( !items.empty() ) emit requestReplacePlaylist( items );

}

//________________________________________________________
void AudioStreamWidget::_addSelectionToFavorites()
{
    Debug::Throw( "AudioStreamWidget::_addSelectionToFavorites.\n" );

    // get selected channels
    auto items( selectedItems() );

    // erase everything that is not a Channel
    items.erase( std::remove_if( items.begin(), items.end(),
        []( const AudioStreamItem& item )
    { return item.type() != AudioStreamItem::Type::Channel || item.providerType() == AudioStreamItem::ProviderType::Favorites; } ), items.end() );

    if( std::any_of( items.begin(), items.end(), AudioStreamItem::SameProviderTypeFTor( AudioStreamItem::ProviderType::ShoutCast ) ) )
    {

        // create AudioStreamConverter
        auto converter = new AudioStreamConverter( this, items, -1 );
        connect( converter, SIGNAL(finished(AudioStreamItem::List,int)), SLOT(addStreamsToFavorites(AudioStreamItem::List)) );
        connect( converter, SIGNAL(finished()), converter, SLOT(deleteLater()) );

        // run
        converter->run();

    } else {

        addStreamsToFavorites( items );

    }

}


//________________________________________________________
void AudioStreamWidget::_removeSelectionFromFavorites()
{
    Debug::Throw( "AudioStreamWidget::_removeSelectionFromFavorites.\n" );

    // get selected channels
    auto items( selectedItems() );

    // erase everything that is not a Channel
    items.erase( std::remove_if( items.begin(), items.end(),
        []( const AudioStreamItem& item )
    { return !( item.type() == AudioStreamItem::Type::Channel && item.providerType() == AudioStreamItem::ProviderType::Favorites ); } ), items.end() );

    if( items.empty() ) return;

    QuestionDialog dialog( this, items.size() > 1 ? tr( "Remove selected items from favorites ?" ):tr( "Remove selected item from favorites ?" ) );
    dialog.okButton().setText( tr( "Remove" ) );
    if( !dialog.exec() ) return;

    lock();
    model_.remove( favoritesItem_, items );
    unlock();

}

//________________________________________________________
void AudioStreamWidget::_renameAudioStream()
{
    Debug::Throw( "AudioStreamWidget::_renameAudioStream.\n" );

    auto indexes( selectedRows() );

    // erase everything that is not a Channel
    indexes.erase( std::remove_if( indexes.begin(), indexes.end(),
        [this]( const QModelIndex& index )
    {
        const auto& item( model_.get( index ) );
        return !( item.type() == AudioStreamItem::Type::Channel && item.providerType() == AudioStreamItem::ProviderType::Favorites );
    } ), indexes.end() );

    if( indexes.size() != 1 ) return;

    list_->edit( proxyModel_.mapFromSource( indexes.front() ) );

}


//________________________________________________________
void AudioStreamWidget::_updateActions()
{
    Debug::Throw( "AudioStreamWidget::_updateActions.\n" );
    const auto selection( selectedItems() );

    const bool hasFavorites( std::any_of( selection.begin(), selection.end(),
        []( const AudioStreamItem& item )
        { return item.type() == AudioStreamItem::Type::Channel && item.providerType() == AudioStreamItem::ProviderType::Favorites; } ) );
    removeFromFavoritesAction_->setEnabled( hasFavorites );
    renameAudioStreamAction_->setEnabled( hasFavorites && selection.size() == 1 )  ;

    const bool hasProviderChannels( std::any_of( selection.begin(), selection.end(),
        []( const AudioStreamItem& item )
        { return item.type() == AudioStreamItem::Type::Channel && item.providerType() != AudioStreamItem::ProviderType::Favorites; } ) );
    addToFavoritesAction_->setEnabled( hasProviderChannels );

    const bool hasChannels( std::any_of( selection.begin(), selection.end(), AudioStreamItem::SameTypeFTor( AudioStreamItem::Type::Channel ) ) );
    addToPlaylistAction_->setEnabled( hasChannels );
    replacePlaylistAction_->setEnabled( hasChannels );

    const bool hasProvider( std::any_of( selection.begin(), selection.end(),
        []( const AudioStreamItem& item )
        { return item.type() == AudioStreamItem::Type::Provider && item.providerType() != AudioStreamItem::ProviderType::Favorites; } ) );
    updateAudioStreamsAction_->setEnabled( hasProvider );

}

//________________________________________________________
void AudioStreamWidget::_showToolTip( const QModelIndex& index )
{
    Debug::Throw( "AudioStreamWidget::_showToolTip.\n" );

    // check index and window
    if( !index.isValid() ) toolTipWidget_->hide();
    else {

        // get collection item and check
        const auto& item( model_.get( proxyModel_.mapToSource( index ) ) );

        // move tooltip
        toolTipWidget_->setIndexRect(
            list_->visualRect( index ).
            translated( list_->viewport()->mapToGlobal( QPoint( 0, 0 ) ) ) );

        // get icon
        auto icon( model_.icon( item ) );
        if( item.type() == AudioStreamItem::Type::Channel ) toolTipWidget_->setAudioStreamItem( item, icon );
        else toolTipWidget_->setAudioStreamItem( item, icon, proxyModel_.rowCount( index ) );

        // show
        toolTipWidget_->showDelayed();

    }

}

//________________________________________________________
void AudioStreamWidget::_installActions()
{
    Debug::Throw( "AudioStreamWidget::_installActions.\n" );

    addAction( updateAudioStreamsAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Update Stream List" ), this ) );
    updateAudioStreamsAction_->setShortcut( QKeySequence::Refresh );
    updateAudioStreamsAction_->setShortcutContext( Qt::WidgetWithChildrenShortcut );
    connect( updateAudioStreamsAction_, SIGNAL(triggered()), SLOT(_updateAudioStreams()) );
    updateAudioStreamsAction_->setEnabled( false );

    addAction( addToPlaylistAction_ = new QAction( IconEngine::get( IconNames::Add ), tr( "Add to Playlist" ), this ) );
    connect( addToPlaylistAction_, SIGNAL(triggered()), SLOT(_addSelectionToPlaylist()) );
    addToPlaylistAction_->setEnabled( false );

    addAction( replacePlaylistAction_ = new QAction( tr( "Replace Playlist" ), this ) );
    connect( replacePlaylistAction_, SIGNAL(triggered()), SLOT(_replacePlaylist()) );
    replacePlaylistAction_->setEnabled( false );

    addAction( addToFavoritesAction_ = new QAction( IconEngine::get( IconNames::AddFavorite ), tr( "Add to Favorites" ), this ) );
    connect( addToFavoritesAction_, SIGNAL(triggered()), SLOT(_addSelectionToFavorites()) );
    addToFavoritesAction_->setEnabled( false );

    addAction( removeFromFavoritesAction_ = new QAction( IconEngine::get( IconNames::RemoveFavorite ), tr( "Remove from Favorites..." ), this ) );
    removeFromFavoritesAction_->setShortcut( QKeySequence::Delete );
    removeFromFavoritesAction_->setShortcutContext( Qt::WidgetWithChildrenShortcut );
    connect( removeFromFavoritesAction_, SIGNAL(triggered()), SLOT(_removeSelectionFromFavorites()) );
    removeFromFavoritesAction_->setEnabled( false );

    addAction( renameAudioStreamAction_ = new QAction( IconEngine::get( IconNames::Edit ), tr( "Rename Stream..." ), this ) );
    renameAudioStreamAction_->setShortcut( Qt::Key_F2 );
    renameAudioStreamAction_->setShortcutContext( Qt::WidgetWithChildrenShortcut );
    connect( renameAudioStreamAction_, SIGNAL(triggered()), SLOT(_renameAudioStream()) );
    renameAudioStreamAction_->setEnabled( false );

    addAction( newAudioStreamAction_ = new QAction( IconEngine::get( IconNames::Add ), tr( "Add New Stream to Favorites..." ), this ) );
    connect( newAudioStreamAction_, SIGNAL(triggered()), SLOT(_newAudioStream()) );


}

//________________________________________________________
void AudioStreamWidget::_installProviders()
{
    Debug::Throw( "AudioStreamWidget::_installProviders.\n" );

    {
        // favorites
        favoritesItem_ = AudioStreamItem( AudioStreamItem::ProviderType::Favorites, AudioStreamItem::Type::Provider, QString() );
        favoritesItem_.setProvider( tr( "Favorites" ) );
        model_.add( favoritesItem_ );
    }

    {
        // icecast
        AudioStreamItem item( AudioStreamItem::ProviderType::IceCast, AudioStreamItem::Type::Provider, AudioStreamInterface::iceCastUrl );
        item.setProvider( "IceCast" );
        model_.add( item );
    }

    {
        // shoutcast
        static const auto shoutCastUrl = QLatin1String( "http://" ) + AudioStreamInterface::shoutCastHost + QLatin1String( "/genre/primary?f=xml&k=" ) + AudioStreamInterface::shoutCastApiKey;
        AudioStreamItem item( AudioStreamItem::ProviderType::ShoutCast, AudioStreamItem::Type::Provider, shoutCastUrl );
        item.setProvider( "ShoutCast" );
        model_.add( item );
    }

}


//________________________________________________________
void AudioStreamWidget::_readAudioStreams( const AudioStreamItem& parent )
{

    Debug::Throw( "AudioStreamWidget::_readAudioStreams.\n" );

    // check if not already loaded
    // check filename
    const auto fileName( _fileName( parent.providerType() ) );
    if( fileName.get().isEmpty() ) return;

    // qfile
    QFile qfile( fileName );
    if( !qfile.open( QIODevice::ReadOnly ) ) return;

    // dom document
    XmlError error;
    QDomDocument document;
    if( !document.setContent( qfile.readAll(), error ) ) return;

    // read
    auto channels( AudioStreamItem::Helper::list( document.documentElement() ) );

    // sort if favorites
    if( parent.providerType() == AudioStreamItem::ProviderType::Favorites )
    { std::sort( channels.begin(), channels.end() ); }

    // add
    addChildren( parent, channels );

}

//________________________________________________________
void AudioStreamWidget::_writeAudioStreams( const AudioStreamItem& constParent )
{
    Debug::Throw( "AudioStreamWidget::_writeAudioStreams.\n" );

    // convert to provider
    auto parent( constParent );
    parent.setType( AudioStreamItem::Type::Provider );

    // get filename
    const auto fileName( _fileName( parent.providerType() ) );

    if( fileName.get().isEmpty() ) return;
    QFile file( fileName );
    if( file.open( QIODevice::WriteOnly ) )
    {
        // write
        QDomDocument document;
        document.appendChild( AudioStreamItem::Helper::domElement( model_.children( parent ), document ) );
        file.write( document.toByteArray() );
    }

}

//________________________________________________________
void AudioStreamWidget::_doUpdateAudioStreams( const QModelIndex& index )
{
    Debug::Throw()
        << "AudioStreamWidget::_doUpdateAudioStreams -"
        << " index: " << index.row() << " " << index.column()
        << " locked: " << locked_
        << endl;

    // do nothing if locked
    if( locked_ ) return;

    // find item
    const auto& item( model_.get(index) );

    // load from cache
    if( item.type() == AudioStreamItem::Type::Provider && !loadedItems_.contains( item ) )
    {
        _readAudioStreams( item );
        loadedItems_.insert( item );
    }

    // load from internet
    switch( item.providerType() )
    {

        default: break;

        case AudioStreamItem::ProviderType::IceCast:
        case AudioStreamItem::ProviderType::ShoutCast:
        emit requestChildren( item );
        break;

    }

}

//________________________________________________________
AudioStreamItem::List AudioStreamWidget::_expandedItems() const
{

    // store model expanded items
    AudioStreamItem::List out;
    auto indexes = model_.indexes();

    // get items for all indexes that are expanded
    std::transform(
        indexes.begin(),
        std::remove_if( indexes.begin(), indexes.end(),
        [this]( const QModelIndex& index ) { return !list_->isExpanded( index ); } ),
        std::back_inserter( out ),
        [this]( const QModelIndex& index ) { return model_.get( index ); } );

    return out;
}

//________________________________________________________
File AudioStreamWidget::_fileName( const AudioStreamItem::ProviderType& type )
{

    const File path( XmlOptions::get().raw( "AUDIO_STREAM_CACHE_DIR" ) );
    QDir().mkpath( path );
    switch( type )
    {
        case AudioStreamItem::ProviderType::IceCast:
        return File( "icecast.xml" ).addPath( path );

        case AudioStreamItem::ProviderType::ShoutCast:
        return File( "shoutcast.xml" ).addPath( path );

        case AudioStreamItem::ProviderType::Favorites:
        return File( "favorites.xml" ).addPath( path );

        default: return File();
    }

}
