#ifndef AudioStreamWidget_h
#define AudioStreamWidget_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "AudioStreamItem.h"
#include "AudioStreamModel.h"
#include "AudioStreamProxyModel.h"

#include <QWidget>

class ScrollBarMonitor;
class AudioStreamFindWidget;
class AudioStreamView;
class AudioStreamItemToolTipWidget;

class AudioStreamWidget: public QWidget, private Base::Counter<AudioStreamWidget>
{

    Q_OBJECT

    public:

    //* constructor
    explicit AudioStreamWidget( QWidget* = nullptr );

    //*@name accessors
    //@{

    //* list
    AudioStreamView& list() const
    { return *list_; }

    //* model
    const AudioStreamModel& model() const
    { return model_; }

    //* selected indexes (from model_)
    QModelIndexList selectedRows() const;

    //* selected items
    AudioStreamItem::List selectedItems() const
    { return model_.get( selectedRows() ); }

    //@}

    //*@name modifiers
    //@{

    //* model
    AudioStreamModel& model()
    { return model_; }

    //* save expanded items
    void saveExpandedItems();

    //* restore
    void restoreExpandedItems();

    //* lock
    void lock();

    //* unlock
    void unlock();

    //* url handlers
    void setUrlHandlers( const QStringList& value )
    { urlHandlers_ = value; }

    //@}

    Q_SIGNALS:

    //* request children for a given item
    void requestChildren( const AudioStreamItem& );

    //* add items to playlist
    void requestAddToPlaylist( const AudioStreamItem::List& );

    //* replace playlist
    void requestReplacePlaylist( const AudioStreamItem::List& );

    //* emitted when a stream url is updated
    void audioStreamUpdated( const AudioStreamItem&, const QString& );

    public Q_SLOTS:

    //* append children
    void addChildren( const AudioStreamItem&, AudioStreamItem::List );

    //* add items to favorites
    void addStreamsToFavorites( const AudioStreamItem::List& );

    private Q_SLOTS:

    //* save configuration
    void _saveConfiguration();

    //* filter
    void _setFilter( const QString& );

    //* index activated
    void _itemActivated( const QModelIndex& );

    //* update selected audioStream
    void _updateAudioStreams();

    //* index exmapnded
    void _updateAudioStreams( const QModelIndex& index )
    { _doUpdateAudioStreams( proxyModel_.mapToSource( index ) ); }

    //* new audio stream
    void _newAudioStream();

    //* add to playlist
    void _addSelectionToPlaylist();

    //* replace playlist
    void _replacePlaylist();

    //* add selected audioStream to favorites
    void _addSelectionToFavorites();

    //* remove selected audioStreams from favorites
    void _removeSelectionFromFavorites();

    //* rename audioStream
    void _renameAudioStream();

    //* update actions
    void _updateActions();

    //* show tooltip
    void _showToolTip( const QModelIndex& );

    private:

    //* actions
    void _installActions();

    //* default providers
    void _installProviders();

    //* read audioStreams from cache
    void _readAudioStreams( const AudioStreamItem& );

    //* save audioStreams to cache
    void _writeAudioStreams( const AudioStreamItem& );

    //* read favorites
    void _readFavorites()
    { return _readAudioStreams( favoritesItem_ ); }

    //* save favorites
    void _writeFavorites()
    { return _writeAudioStreams( favoritesItem_ ); }

    //* update audioStreams
    void _doUpdateAudioStreams( const QModelIndex& );

    //* list of expanded items
    AudioStreamItem::List _expandedItems() const;

    //* cache filename
    static File _fileName( const AudioStreamItem::ProviderType& );

    //* locked
    /** prevent list roundtrips */
    bool locked_ = false;

    //* collection find widget
    AudioStreamFindWidget* findWidget_ = nullptr;

    //* treeview
    AudioStreamView* list_ = nullptr;

    //* scrollbar monitor
    ScrollBarMonitor* scrollBarMonitor_ = nullptr;

    //* model
    AudioStreamModel model_;

    //* proxy model
    AudioStreamProxyModel proxyModel_;

    //* favorites parent item
    AudioStreamItem favoritesItem_;

    //* expanded items
    AudioStreamItem::List expandedItems_;

    //* list of loaded items (from cache)
    AudioStreamItem::Set loadedItems_;

    //* tooltip widget
    AudioStreamItemToolTipWidget* toolTipWidget_ = nullptr;

    //* list of valid url handlers
    QStringList urlHandlers_;

    //*@name actions
    //@{

    //* update collection
    QAction* updateAudioStreamsAction_ = nullptr;

    //* add to playlist
    QAction* addToPlaylistAction_ = nullptr;

    //* replace playlist
    QAction* replacePlaylistAction_ = nullptr;

    //* add to favorites
    QAction* addToFavoritesAction_ = nullptr;

    //* remove from favorites
    QAction* removeFromFavoritesAction_ = nullptr;

    //* edit audioStream
    QAction* renameAudioStreamAction_ = nullptr;

    //* new audio stream action
    QAction* newAudioStreamAction_ = nullptr;

    //@}

};

#endif
