/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CollectionWatcher.h"

#include "Debug.h"
#include "File.h"
#include "FileSystemWatcher.h"
#include "XmlOptions.h"

//_________________________________________
CollectionWatcher::CollectionWatcher( QObject* parent ):
    QObject( parent ),
    Counter( "CollectionWatcher" ),
    fileSystemWatcher_( new FileSystemWatcher( this ) )
{ Debug::Throw( "CollectionWatcher::CollectionWatcher.\n" ); }

//_________________________________________
void CollectionWatcher::updateEnableState()
{
    Debug::Throw( "CollectionWatcher::updateEnableState.\n" );
    const bool enabled( XmlOptions::get().get<bool>( "WATCH_COLLECTION_CHANGES" ) );
    if( enabled == enabled_ ) return;

    // assign
    enabled_ = enabled;

    // update connections
    if( enabled )
    {

        connect( fileSystemWatcher_, SIGNAL(directoryChangedDelayed(QString)), this, SLOT(_directoryChanged(QString)) );

    } else {

        disconnect( fileSystemWatcher_, SIGNAL(directoryChangedDelayed(QString)), this, SLOT(_directoryChanged(QString)) );

    }

}

//_________________________________________
void CollectionWatcher::updateMusicDirectory()
{
    Debug::Throw( "CollectionWatcher::updateMusicDirectory.\n" );

    // get music directory
    File musicDirectory( XmlOptions::get().raw( "MUSIC_DIR" ) );
    musicDirectory.addTrailingSlash();
    if( musicDirectory == musicDirectory_ ) return;

    // clear old watched folders
    const auto directories( fileSystemWatcher_->directories() );
    if( !directories.isEmpty() ) fileSystemWatcher_->removePaths( directories );

    // assign
    musicDirectory_ = musicDirectory;
    fileSystemWatcher_->addPath( musicDirectory_ );

    for( const auto& path : File( musicDirectory_ ).listFiles( static_cast<File::ListFlags>(File::Recursive|File::FollowLinks|File::FoldersOnly) ) )
    { fileSystemWatcher_->addPath( path ); }

}


//_________________________________________
void CollectionWatcher::_directoryChanged(QString path)
{
    Debug::Throw() << "CollectionWatcher::_directoryChanged - path: " << path << endl;

    // check path
    if( !path.startsWith( musicDirectory_ ) ) return;

    // update watched path
    const File pathFile( path );
    if( pathFile.exists() )
    {

        fileSystemWatcher_->addPath( path );
        for( const auto& subpath : pathFile.listFiles( static_cast<File::ListFlags>(File::Recursive|File::FollowLinks|File::FoldersOnly) ) )
        { fileSystemWatcher_->addPath( subpath ); }

    }

    // emit signal
    if( path == musicDirectory_ ) emit collectionChanged();
    else emit collectionChanged( path.mid( musicDirectory_.get().size() ) );

}
