/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CollectionWidget.h"

#include "Debug.h"
#include "CollectionFindWidget.h"
#include "CollectionItemToolTipWidget.h"
#include "CollectionView.h"
#include "ContextMenu.h"
#include "CoverDisplayDialog.h"
#include "CppUtil.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "ScrollBarMonitor.h"

#include <QLayout>

#include <algorithm>
#include <numeric>

//________________________________________________________
CollectionWidget::CollectionWidget( QWidget* parent ):
    QWidget( parent ),
    Counter( "CollectionWidget" )
{
    Debug::Throw( "CollectionWidget::CollectionWidget.\n" );

    _installActions();

    auto vLayout = new QVBoxLayout;
    vLayout->setMargin(0);
    vLayout->setSpacing(2);
    setLayout( vLayout );

    vLayout->addWidget( findWidget_ = new CollectionFindWidget( this ) );

    vLayout->addWidget( list_ = new CollectionView( this ) );
    list_->setModel( &model_ );
    list_->setOptionName( "COLLECTION" );

    scrollBarMonitor_ = new ScrollBarMonitor( list_ );

    // tooltip widget
    toolTipWidget_ = new CollectionItemToolTipWidget( this );

    connect( list_, SIGNAL(activated(QModelIndex)), SLOT(_itemActivated(QModelIndex)) );
    connect( list_, SIGNAL(expanded(QModelIndex)), SLOT(_itemExpanded(QModelIndex)) );
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateActions()) );
    connect( list_, SIGNAL(hovered(QModelIndex)), SLOT(_showToolTip(QModelIndex)) );

    {
        // context menu
        auto menu = new ContextMenu( list_ );
        menu->addAction( addToPlaylistAction_ );
        menu->addAction( replacePlaylistAction_ );
        menu->addAction( deleteTracksAction_ );
        menu->addSeparator();
        menu->addAction( displayCoverAction_ );
        #if WITH_TAGLIB
        menu->addAction( editTagsAction_ );
        #endif
        menu->addSeparator();
        menu->addAction( &list_->collapseAllAction() );
        menu->setHideDisabledActions( true );

        connect( menu, SIGNAL(aboutToShow()), SLOT(_updateActions()) );
    }

    // busy widget
    busyWidget_ = new BusyWidget( this, BusyWidget::Location::BottomRight );
    busyWidget_->hide();
}

//________________________________________________________
CollectionItem::List CollectionWidget::expandedItems() const
{
    // store model expanded items
    CollectionItem::List out;
    auto indexes = model_.indexes();

    // get items for all indexes that are expanded
    std::transform(
        indexes.begin(),
        std::remove_if( indexes.begin(), indexes.end(),
        [this]( const QModelIndex& index ) { return !list_->isExpanded( index ); } ),
        std::back_inserter( out ),
        [this]( const QModelIndex& index ) { return model_.get( index ); } );

    return out;
}


//_________________________________________________________
void CollectionWidget::updateArtists( const CollectionItem::List& items )
{

    Debug::Throw( "CollectionWidget::updateArtists.\n" );

    // split items by artist, then albums, then assign to models
    for( auto iter = items.begin(); iter != items.end(); )
    {

        CollectionItem artist( *iter );
        artist.setType( CollectionItem::Type::Artist );
        CollectionItem::List artistCopy;
        CollectionItem::List::const_iterator next;

        std::copy(
            iter,
            next = std::find_if_not( iter, items.end(), TrackInformation::SameArtistFTor(artist) ),
            std::back_inserter( artistCopy ) );

        // update artist album count and duration
        artist.setDuration( std::accumulate( artistCopy.begin(), artistCopy.end(), qint32(0),
            [](qint32 duration, const CollectionItem& item )
            { return duration + item.duration(); } ) );

        // get artist albums
        CollectionItem::List albums( artistCopy );
        albums.erase(
            std::unique( albums.begin(), albums.end(), TrackInformation::SameAlbumBinaryFTor() ),
            albums.end() );

        for( auto& item:albums )
        {
            item.setType( CollectionItem::Type::Album );
            item.setShowDiscNumber( std::count_if( albums.begin(), albums.end(),
                [&item]( const CollectionItem& other ) { return other.album() == item.album(); } ) > 1 );
        }

        // assign number of albums to artist
        artist.setAlbumCount( albums.size() );

        // replace artist in model
        model_.updateArtist( artist );

        // assign albums
        std::sort( albums.begin(), albums.end() );
        model_.set( artist, albums );

        iter = next;
    }

}

//_________________________________________________________
void CollectionWidget::updateAlbums( const CollectionItem::List& items )
{

    Debug::Throw( "CollectionWidget::updateAlbums.\n" );

    // split items by album and assign to model
    for( auto iter = items.begin(); iter != items.end(); )
    {
        CollectionItem album( *iter );
        album.setType( CollectionItem::Type::Album );
        CollectionItem::List copy;
        CollectionItem::List::const_iterator next;
        std::copy(
            iter,
            next = std::find_if_not( iter, items.end(), TrackInformation::SameAlbumFTor(album) ),
            std::back_inserter( copy ) );

        // calculate total duration and assign to album
        album.setDuration( std::accumulate( copy.begin(), copy.end(), qint32(0),
            [](qint32 duration, const CollectionItem& item )
            { return duration + item.duration(); } ) );

        album.setTrackCount( copy.size() );

        // replace album in model
        model_.updateAlbum( album );

        // assign songs
        model_.set( album, copy );
        iter = next;
    }

}

//________________________________________________________
void CollectionWidget::saveExpandedItems()
{
    Debug::Throw( "CollectionWidget::saveExpandedItems.\n" );
    expandedItems_ = expandedItems();
}

//________________________________________________________
void CollectionWidget::restoreExpandedItems()
{
    Debug::Throw( "CollectionWidget::restoreExpandedItems.\n" );
    for( const auto& item:expandedItems_ )
    {
        const auto index( model_.index( item ) );
        if( index.isValid() ) list_->setExpanded( index, true );
    }

    expandedItems_.clear();
}

//________________________________________________________
void CollectionWidget::lock()
{
    locked_ = true;
    list_->setUpdatesEnabled( false );
    scrollBarMonitor_->save();
}

//________________________________________________________
void CollectionWidget::unlock()
{
    scrollBarMonitor_->restore();
    list_->setUpdatesEnabled( true );
    locked_ = false;
}

//________________________________________________________
void CollectionWidget::_itemActivated( const QModelIndex& index )
{
    Debug::Throw()
        << "CollectionWidget::_itemActivated -"
        << " index: " << index.row() << " " << index.column()
        << endl;

    // find item
    const auto& item( model_.get(index) );

    // emit signal
    switch( item.type() )
    {
        default: break;

        case CollectionItem::Type::Artist:
        case CollectionItem::Type::Album:
        list_->setExpanded( index, !list_->isExpanded( index ) );
        break;

        case CollectionItem::Type::Track:
        emit requestAddToPlaylist( Base::makeT<CollectionItem::List>({item}) );
        break;

    }
}

//________________________________________________________
void CollectionWidget::_itemExpanded( const QModelIndex& index )
{
    Debug::Throw()
        << "CollectionWidget::_itemExpanded -"
        << " index: " << index.row() << " " << index.column()
        << " locked: " << locked_
        << endl;

    // do nothing if locked
    if( locked_ ) return;

    // find item
    const auto& item( model_.get(index) );

    // emit signal
    switch( item.type() )
    {
        default: break;
        case CollectionItem::Type::Artist: emit requestAlbums( item ); break;
        case CollectionItem::Type::Album: emit requestTracks( item ); break;
    }
}

//_______________________________________________________
void CollectionWidget::_updateActions()
{
    const auto selectedIndexes( list_->selectionModel()->selectedRows() );
    const bool hasSelection( !selectedIndexes.empty() );
    addToPlaylistAction_->setEnabled( hasSelection );
    replacePlaylistAction_->setEnabled( hasSelection );
    deleteTracksAction_->setEnabled( hasSelection );

    bool displayCoverActionEnabled( false );
    if( selectedIndexes.size() == 1 )
    {
        auto item( model_.get( selectedIndexes.front() ) );
        if( item.type() == CollectionItem::Type::Track ) item.setType( CollectionItem::Type::Album );
        if( ( item.type() == CollectionItem::Type::Artist || item.type() == CollectionItem::Type::Album ) && !model_.icon( item ).isNull() )
        { displayCoverActionEnabled = true; }
    }
    displayCoverAction_->setEnabled( displayCoverActionEnabled );

    #if WITH_TAGLIB
    editTagsAction_->setEnabled( hasSelection );
    #endif

}

//________________________________________________________
void CollectionWidget::_addSelectionToPlaylist()
{
    Debug::Throw( "CollectionWidget::_addSelectionToPlaylist.\n" );
    auto items( model_.get( list_->selectionModel()->selectedRows() ) );
    if( !items.empty() ) emit requestAddToPlaylist( items );
}

//________________________________________________________
void CollectionWidget::_replacePlaylist()
{
    Debug::Throw( "CollectionWidget::_replacePlaylist.\n" );
    auto items( model_.get( list_->selectionModel()->selectedRows() ) );
    if( !items.empty() ) emit requestReplacePlaylist( items );
}

//________________________________________________________
void CollectionWidget::_deleteTracks()
{
    Debug::Throw( "CollectionWidget::_deleteTracks.\n" );
    auto items( model_.get( list_->selectionModel()->selectedRows() ) );
    if( !items.empty() ) emit requestDeleteTracks( items );
}

//________________________________________________________
void CollectionWidget::_showToolTip( const QModelIndex& index )
{
    Debug::Throw( "CollectionWidget::_showToolTip.\n" );

    // check index and window
    if( !index.isValid() ) toolTipWidget_->hide();
    else {

        // get collection item and check
        const auto& item( model_.get( index ) );
        if( item.type() == CollectionItem::Type::Artist && item.artist().isEmpty() )
        {
            toolTipWidget_->hide();
            return;
        }

        // move tooltip
        toolTipWidget_->setIndexRect(
            list_->visualRect( index ).
            translated( list_->viewport()->mapToGlobal( QPoint( 0, 0 ) ) ) );

        // get icon
        auto icon( model_.icon( item ) );
        if( icon.isNull() ) icon = IconEngine::get( IconNames::Track );

        // assign item
        static_cast<CollectionItemToolTipWidget*>(toolTipWidget_)->setCollectionItem( item, icon );

        // show
        toolTipWidget_->showDelayed();

    }

}

//________________________________________________________
void CollectionWidget::_displayCover()
{
    Debug::Throw( "CollectionWidget::_displayCover.\n" );

    const auto selectedIndexes( list_->selectionModel()->selectedRows() );
    if( selectedIndexes.size() != 1 ) return;

    auto item( model_.get( selectedIndexes.front() ) );
    if( item.type() == CollectionItem::Type::Track ) item.setType( CollectionItem::Type::Album );
    if( !( item.type() == CollectionItem::Type::Artist || item.type() == CollectionItem::Type::Album ) ) return;

    const auto icon( model_.icon( item ) );
    if( icon.isNull() ) return;

    CoverDisplayDialog( this, item, icon ).exec();

}

//________________________________________________________
#if WITH_TAGLIB
void CollectionWidget::_editTags()
{ emit requestEditTags( model_.get( list_->selectionModel()->selectedRows() ) ); }
#endif

//________________________________________________________
void CollectionWidget::_installActions()
{
    Debug::Throw( "CollectionWidget::_installActions.\n" );

    addAction( updateCollectionAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Update Collection" ), this ) );
    updateCollectionAction_->setShortcut( QKeySequence::Refresh );
    updateCollectionAction_->setShortcutContext( Qt::WidgetWithChildrenShortcut );
    connect( updateCollectionAction_, SIGNAL(triggered()), SIGNAL(requestCollectionUpdate()) );

    addAction( addToPlaylistAction_ = new QAction( IconEngine::get( IconNames::Add ), tr( "Add to Playlist" ), this ) );
    connect( addToPlaylistAction_, SIGNAL(triggered()), SLOT(_addSelectionToPlaylist()) );
    addToPlaylistAction_->setEnabled( false );

    addAction( replacePlaylistAction_ = new QAction( tr( "Replace Playlist" ), this ) );
    connect( replacePlaylistAction_, SIGNAL(triggered()), SLOT(_replacePlaylist()) );
    replacePlaylistAction_->setEnabled( false );

    addAction( deleteTracksAction_ = new QAction( IconEngine::get( IconNames::Delete ), tr( "Delete tracks" ), this ) );
    connect( deleteTracksAction_, SIGNAL(triggered()), SLOT(_deleteTracks()) );
    deleteTracksAction_->setShortcut( QKeySequence::Delete );
    deleteTracksAction_->setShortcutContext( Qt::WidgetWithChildrenShortcut );
    deleteTracksAction_->setEnabled( false );

    addAction( displayCoverAction_ = new QAction( IconEngine::get( IconNames::DisplayCover ), tr( "Display cover" ), this ) );
    connect( displayCoverAction_, SIGNAL(triggered()), SLOT(_displayCover()) );
    displayCoverAction_->setEnabled( false );

    #if WITH_TAGLIB
    addAction( editTagsAction_ = new QAction( IconEngine::get( IconNames::Edit ), tr( "Edit Track Details..." ), this ) );
    editTagsAction_->setShortcut( Qt::Key_F2 );
    editTagsAction_->setShortcutContext( Qt::WidgetWithChildrenShortcut );
    connect( editTagsAction_, SIGNAL(triggered()), SLOT(_editTags()) );
    editTagsAction_->setEnabled( false );
    #endif
}
