/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CurrentTrackInformationWidget.h"

#include "Color.h"
#include "CollectionItem.h"
#include "CoverDisplayDialog.h"
#include "Debug.h"
#include "ElidedLabel.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "QtUtil.h"
#include "Singleton.h"
#include "XmlOptions.h"

#include <QLayout>
#include <QMainWindow>
#include <QPainter>

#include <cmath>

namespace
{

    namespace Local
    {

        static const int tileSize = 32;

        //_______________________________________________________________
        template <class T>
        inline void memrotate90(const T *src, int w, int h, int sstride, T *dest, int dstride)
        {
            const int numTilesX = (w + tileSize - 1) / tileSize;
            const int numTilesY = (h + tileSize - 1) / tileSize;

            for (int tx = 0; tx < numTilesX; ++tx)
            {
                const int startx = w - tx * tileSize - 1;
                const int stopx = qMax(startx - tileSize, 0);

                for (int ty = 0; ty < numTilesY; ++ty)
                {
                    const int starty = ty * tileSize;
                    const int stopy = qMin(starty + tileSize, h);

                    for (int x = startx; x >= stopx; --x)
                    {
                        T *d = (T *)((char*)dest + (w - x - 1) * dstride) + starty;
                        const char *s = (const char*)(src + x) + starty * sstride;
                        for (int y = starty; y < stopy; ++y)
                        {
                            *d++ = *(const T *)(s);
                            s += sstride;
                        }
                    }
                }
            }
        }

        //_______________________________________________________________
        template <class T>
        inline void memrotate270(const T *src, int w, int h, int sstride, T *dest, int dstride)
        {
            const int numTilesX = (w + tileSize - 1) / tileSize;
            const int numTilesY = (h + tileSize - 1) / tileSize;

            for (int tx = 0; tx < numTilesX; ++tx)
            {
                const int startx = tx * tileSize;
                const int stopx = qMin(startx + tileSize, w);

                for (int ty = 0; ty < numTilesY; ++ty)
                {
                    const int starty = h - 1 - ty * tileSize;
                    const int stopy = qMax(starty - tileSize, 0);

                    for (int x = startx; x < stopx; ++x)
                    {
                        T *d = (T*)((char*)dest + x * dstride) + h - 1 - starty;
                        const char *s = (const char*)(src + x) + starty * sstride;
                        for (int y = starty; y >= stopy; --y)
                        {
                            *d++ = *(const T*)s;
                            s -= sstride;
                        }
                    }
                }
            }
        }

        //_______________________________________________________________
        template <int shift>
        inline int staticShift(int value)
        {
            if (shift == 0) return value;
            else if (shift > 0) return value << (uint(shift) & 0x1f);
            else return value >> (uint(-shift) & 0x1f);
        }

        //_______________________________________________________________
        template<int aprec, int zprec>
        inline void blurInner(uchar *bptr, int &zR, int &zG, int &zB, int &zA, int alpha)
        {
            QRgb *pixel = (QRgb *)bptr;

            #define Z_MASK (0xff << zprec)
            const int A_zprec = staticShift<zprec - 24>(*pixel) & Z_MASK;
            const int R_zprec = staticShift<zprec - 16>(*pixel) & Z_MASK;
            const int G_zprec = staticShift<zprec - 8>(*pixel)  & Z_MASK;
            const int B_zprec = staticShift<zprec>(*pixel)      & Z_MASK;
            #undef Z_MASK

            const int zR_zprec = zR >> aprec;
            const int zG_zprec = zG >> aprec;
            const int zB_zprec = zB >> aprec;
            const int zA_zprec = zA >> aprec;

            zR += alpha * (R_zprec - zR_zprec);
            zG += alpha * (G_zprec - zG_zprec);
            zB += alpha * (B_zprec - zB_zprec);
            zA += alpha * (A_zprec - zA_zprec);

            #define ZA_MASK (0xff << (zprec + aprec))
            *pixel =
                staticShift<24 - zprec - aprec>(zA & ZA_MASK)
                | staticShift<16 - zprec - aprec>(zR & ZA_MASK)
                | staticShift<8 - zprec - aprec>(zG & ZA_MASK)
                | staticShift<-zprec - aprec>(zB & ZA_MASK);
            #undef ZA_MASK
        }

        //_______________________________________________________________
        template<int aprec, int zprec>
        inline void blurRow(QImage & im, int line, int alpha)
        {
            auto bptr = im.scanLine(line);
            int zR = 0;
            int zG = 0;
            int zB = 0;
            int zA = 0;

            const int stride = im.depth() >> 3;
            const int im_width = im.width();
            for (int index = 0; index < im_width; ++index)
            {
                blurInner<aprec, zprec>(bptr, zR, zG, zB, zA, alpha);
                bptr += stride;
            }

            bptr -= stride;

            for (int index = im_width - 2; index >= 0; --index)
            {

                bptr -= stride;
                blurInner<aprec, zprec>(bptr, zR, zG, zB, zA, alpha);
            }
        }

        //_______________________________________________________________
        template <int aprec, int zprec>
        void exponentialBlur(QImage &img, qreal radius )
        {

            // halve the radius if we're using two passes
            radius /= 2;

            // choose the alpha such that pixels at radius distance from a fully
            // saturated pixel will have an alpha component of no greater than
            // the cutOffIntensity
            const qreal cutOffIntensity = 2;
            int alpha = radius <= qreal(1e-5) ?
                ((1 << aprec)-1):
                qRound((1<<aprec)*(1 - std::pow(cutOffIntensity * (1 / qreal(255)), 1 / radius)));

            int img_height = img.height();
            for (int row = 0; row < img_height; ++row)
            {
                for (int i = 0; i <= 1; ++i)
                { blurRow<aprec, zprec>(img, row, alpha); }
            }

            QImage temp(img.height(), img.width(), img.format());

            #if QT_VERSION >= 0x050300
            temp.setDevicePixelRatio(img.devicePixelRatioF());
            #endif

            memrotate90(reinterpret_cast<const quint32*>(img.bits()),
                img.width(), img.height(), img.bytesPerLine(),
                reinterpret_cast<quint32*>(temp.bits()),
                temp.bytesPerLine());

            img_height = temp.height();
            for (int row = 0; row < img_height; ++row)
            {
                for (int i = 0; i <= 1; ++i)
                { blurRow<aprec, zprec>(temp, row, alpha); }
            }

            memrotate270(reinterpret_cast<const quint32*>(temp.bits()),
                temp.width(), temp.height(), temp.bytesPerLine(),
                reinterpret_cast<quint32*>(img.bits()),
                img.bytesPerLine());

        }

        //____________________________________
        class Label: public ElidedLabel
        {

            public:

            //* constructor
            Label( QWidget* parent = nullptr ):
                ElidedLabel( parent )
                { setElideMode( Qt::ElideRight ); }

            protected:

            //* paint event
            void paintEvent( QPaintEvent* event ) override
            {
                QPainter painter( this );
                painter.setRenderHint( QPainter::TextAntialiasing, true );
                painter.setClipRegion( event->region() );
                painter.translate( QPoint( 1, 1 ) );

                const auto text( QLabel::text() );

                auto color( palette().color( backgroundRole() ) );
                color.setAlphaF( 0.1 );
                painter.setPen( color );
                painter.drawText( rect(), Qt::AlignVCenter, text );

                painter.translate( QPoint( -1, -1 ) );
                painter.setPen( palette().color( foregroundRole() ) );
                painter.drawText( rect(), Qt::AlignVCenter, text );
            }

        };

    }

}

// blur radius
static const int blurRadius = 256;

//_______________________________________________________
CurrentTrackInformationWidget::CurrentTrackInformationWidget( QWidget* parent ):
    Counter( "CurrentTrackInformationWidget" )
{

    Debug::Throw( "CurrentTrackInformationWidget::CurrentTrackInformationWidget.\n" );

    _installActions();

    setBackgroundRole( QPalette::ToolTipBase );
    setForegroundRole( QPalette::ToolTipText );
    setAutoFillBackground( true );

    // layout
    auto layout = new QGridLayout;
    setLayout( layout );
    layout->setMargin(0);
    layout->setSpacing(5);

    layout->addWidget( iconLabel_ = new QLabel( this ), 0, 0, 5, 1 );
    iconLabel_->setAlignment( Qt::AlignHCenter|Qt::AlignTop );
    iconLabel_->setMargin( 10 );

    layout->addItem( new QSpacerItem( 0, 10 ), 0, 1, 1, 1 );
    layout->addWidget( trackLabel_ = new Local::Label( this ), 1, 1, 1, 1 );
    layout->addWidget( artistLabel_ = new Local::Label( this ), 2, 1, 1, 1 );
    layout->addWidget( albumLabel_ = new Local::Label( this ), 3, 1, 1, 1 );
    layout->setColumnStretch( 0, 0 );
    layout->setColumnStretch( 1, 1 );
    layout->setRowStretch( 4, 1 );

    trackLabel_->setFont( QtUtil::titleFont( trackLabel_->font() ) );

    // configuration
    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();
    _updateDefaultPixmap();
    _updateIcon();

}

//_____________________________________________
void CurrentTrackInformationWidget::clear()
{

    Debug::Throw( "CurrentTrackInformationWidget::clear.\n" );

    for( auto&& label:{ artistLabel_, albumLabel_ } )
    { label->clear(); }

    trackLabel_->setText( tr( "No song playing" ) );

    trackInformation_ = TrackInformation();
    icon_ = QIcon();
    _updateIcon();

}

//_____________________________________________
void CurrentTrackInformationWidget::setTrackInformation( const TrackInformation& trackInformation, const QIcon& icon )
{
    // assign
    if( &trackInformation_ != &trackInformation ) trackInformation_ = trackInformation;
    if( &icon_ != &icon ) icon_ = icon;

    // icon
    _updateIcon();

    artistLabel_->setText( trackInformation_.artist() );
    albumLabel_->setText( trackInformation_.album() );
    trackLabel_->setText( trackInformation_.titleOrFile() );
}

//_____________________________________________
bool CurrentTrackInformationWidget::updateCover( const CoverManager::CoverList& coverList )
{
    Debug::Throw( "CurrentTrackInformationWidget::updateCover.\n" );
    for( const auto& item:coverList )
    {
        if( trackInformation_.artist() == item.first.artist() && trackInformation_.album() == item.first.album() && !item.second.isNull() )
        {
            icon_ = QIcon( QPixmap::fromImage( item.second ) );
            _updateIcon();
            return true;
        }
    }

    return false;

}

//_____________________________________________
void CurrentTrackInformationWidget::paintEvent( QPaintEvent* event )
{

    if( backgroundImage_.isNull() ) return;

    QPainter painter( this );
    painter.setClipRegion( event->region() );
    painter.drawImage( rect(), backgroundImage_, QRect( QPoint( blurRadius/4, blurRadius/4 ), size() ) );

}

//_____________________________________________
void CurrentTrackInformationWidget::contextMenuEvent( QContextMenuEvent* event )
{

    auto mainwindow( qobject_cast<QMainWindow*>( window() ) );
    if( !mainwindow ) return;
    std::unique_ptr<QMenu> menu( mainwindow->createPopupMenu() );
    if( !menu ) return;

    // add display cover action
    if( !icon_.isNull() )
    {
        // find first separator
        QAction* before( nullptr );
        for( auto action:menu->actions() )
        {
            if( !before ) before = action;
            if( action->isSeparator() )
            {
                before = action;
                break;
            }

        }

        // insert display cover action
        menu->insertAction( before, displayCoverAction_ );

    }

    // move and show menu
    menu->adjustSize();
    menu->exec( event->globalPos() );

}

//_____________________________________________
void CurrentTrackInformationWidget::_updateConfiguration()
{
    Debug::Throw( "CurrentTrackInformationWidget::_updateConfiguration.\n" );

    // pixmap size from font
    pixmapSize_ = 3*fontMetrics().height() + QFontMetrics( QtUtil::titleFont( font() ) ).height() + 4*5;

    // colors
    if( XmlOptions::get().get<bool>( "BANNER_USE_SYSTEM_COLORS" ) ) setPalette( QPalette() );
    else {

        QPalette palette;
        palette.setColor( QPalette::ToolTipText, XmlOptions::get().get<Base::Color>( "BANNER_FOREGROUND_COLOR" ) );
        palette.setColor( QPalette::ToolTipBase, XmlOptions::get().get<Base::Color>( "BANNER_BACKGROUND_COLOR" ) );

        setPalette( palette );
    }

    // update
    _updateDefaultPixmap();
    _updateIcon();

}

//________________________________________________________
void CurrentTrackInformationWidget::_displayCover()
{
    Debug::Throw( "CurrentTrackInformationWidget::_displayCover.\n" );

    if( !icon_.isNull() )
    {
        CollectionItem item( trackInformation_ );
        item.setType( CollectionItem::Type::Album );
        CoverDisplayDialog( this, item, icon_ ).exec();
    }

}

//_____________________________________________
void CurrentTrackInformationWidget::_updateIcon()
{
    Debug::Throw( "CurrentTrackInformationWidget::_updateIcon.\n" );
    if( icon_.isNull() )
    {

        iconLabel_->setPixmap( defaultPixmap_ );
        currentPixmap_ = QPixmap();
        backgroundImage_ = QImage();

    } else {

        currentPixmap_ = icon_.pixmap( QSize( pixmapSize_, pixmapSize_ ) );
        iconLabel_->setPixmap( currentPixmap_ );
        _updateBackgroundImage();

    }

    update();

}

//_____________________________________________
void CurrentTrackInformationWidget::_updateBackgroundImage()
{
    Debug::Throw( "CurrentTrackInformationWidget::_updateBackgroundImage.\n" );

    if( currentPixmap_.isNull() ) return;
    backgroundImage_ = QImage( size() + QSize( blurRadius/2, blurRadius/2 ), QImage::Format_ARGB32_Premultiplied );
    QPainter painter( &backgroundImage_ );
    painter.setRenderHints( QPainter::SmoothPixmapTransform );

    // calculate pixmap rect
    qreal verticalScale = qreal( backgroundImage_.height() ) / currentPixmap_.height();
    qreal horizontalScale = qreal( backgroundImage_.width() ) / currentPixmap_.width();

    QRect sourceRect;
    if( horizontalScale > verticalScale )
    {

        sourceRect = QRect( QPoint(0,0), QSize( currentPixmap_.width(), currentPixmap_.height()/horizontalScale ) );
        sourceRect.moveTop( (currentPixmap_.height() - sourceRect.height()) / 2 );

    } else {

        sourceRect = QRect( QPoint(0,0), QSize( currentPixmap_.width()/verticalScale, currentPixmap_.height() ) );
        sourceRect.moveLeft( (currentPixmap_.width() - sourceRect.width()) / 2 );

    }

    painter.drawPixmap( backgroundImage_.rect(), currentPixmap_, sourceRect );

    // add background color
    auto color( palette().color( backgroundRole() ) );
    color.setAlphaF( 0.3 );
    painter.setBrush( color );
    painter.setPen( Qt::NoPen );
    painter.drawRect( backgroundImage_.rect() );
    painter.end();

    // blur
    Local::exponentialBlur<12, 10>( backgroundImage_, blurRadius );

}

//________________________________________________________
void CurrentTrackInformationWidget::_installActions()
{
    Debug::Throw( "CurrentTrackInformationWidget::_installActions.\n" );

    addAction( displayCoverAction_ = new QAction( IconEngine::get( IconNames::DisplayCover ), tr( "Display cover" ), this ) );
    connect( displayCoverAction_, SIGNAL(triggered()), SLOT(_displayCover()) );
}

//_____________________________________________
void CurrentTrackInformationWidget::_updateDefaultPixmap()
{
    Debug::Throw( "CurrentTrackInformationWidget::_updateDefaultPixmap.\n" );
    defaultPixmap_ = IconEngine::get( IconNames::DefaultCover ).pixmap( QSize( pixmapSize_, pixmapSize_ ) );
}
