
/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CurrentTrackWidget.h"

#include "CppUtil.h"
#include "Debug.h"
#include "TrackInformation.h"

#include <QLayout>
#include <QMouseEvent>
#include <QStyle>
#include <QStyleOptionSlider>
#include <QToolBar>

//* fraction of a second
const int CurrentTrackWidget::precision_ = 10;

//_________________________________________________________________
//* local slider, so that mouse press move slider immediately at the right position
namespace
{
    namespace Local
    {
        class Slider: public QSlider
        {

            public:

            //* constructor
            Slider( QWidget* parent = nullptr ):
                QSlider( parent )
            {}

            //*@name modifiers
            //@{

            //* seek
            void setSeekEnabled( bool value )
            { seekEnabled_ = value; }

            //@}

            protected:

            //* mouse press event
            void mousePressEvent( QMouseEvent* event ) override
            {

                // nothing if no range
                if( maximum() == minimum() ) return;

                if( !seekEnabled_ ) return;

                if( event->button() == Qt::LeftButton )
                {
                    QStyleOptionSlider option;
                    initStyleOption( &option );
                    const QRect sliderRect = style()->subControlRect(QStyle::CC_Slider, &option, QStyle::SC_SliderHandle, this);
                    const QPoint center = sliderRect.center() - sliderRect.topLeft();
                    setSliderPosition( _pixelToRange( option, ( event->pos() - center ) ) );
                    return QSlider::mousePressEvent( event );
                }
            }

            //* convert mouse position to value
            int _pixelToRange( const QStyleOptionSlider& option, const QPoint& position )
            {
                const auto grooveRect = style()->subControlRect(QStyle::CC_Slider, &option, QStyle::SC_SliderGroove, this);
                const auto handleRect = style()->subControlRect(QStyle::CC_Slider, &option, QStyle::SC_SliderHandle, this);

                switch( orientation() )
                {
                    case Qt::Horizontal:
                    {
                        int sliderLength = handleRect.width();
                        int sliderMin = grooveRect.x();
                        int sliderMax = grooveRect.right() - sliderLength + 1;
                        return QStyle::sliderValueFromPosition( minimum(), maximum(), position.x() - sliderMin, sliderMax - sliderMin, option.upsideDown );
                    }

                    case Qt::Vertical:
                    {
                        int sliderLength = handleRect.height();
                        int sliderMin = grooveRect.y();
                        int sliderMax = grooveRect.bottom() - sliderLength + 1;
                        return QStyle::sliderValueFromPosition( minimum(), maximum(), position.y() - sliderMin, sliderMax - sliderMin, option.upsideDown );
                    }

                    default: return 0;

                }

            }

            private:

            //* true if seek is enabled
            bool seekEnabled_ = true;


        };

    }
}

//_________________________________________________________________
CurrentTrackWidget::CurrentTrackWidget( QWidget* parent ):
    QWidget( parent ),
    Counter( "CurrentTrackWidget" )
{

    Debug::Throw( "CurrentTrackWidget::CurrentTrackWidget" );

    // layout
    auto layout = new QHBoxLayout;
    setLayout( layout );
    layout->addWidget( elapsedTimeLabel_ = new QLabel( this ) );
    layout->addWidget( slider_ = new Local::Slider( this ) );
    layout->addWidget( remainingTimeLabel_ = new QLabel( this ) );

    // orientation
    slider_->setOrientation( Qt::Horizontal );
    setOrientation( static_cast<QToolBar*>( parent )->orientation() );

    auto fontMetrics( elapsedTimeLabel_->fontMetrics() );
    QSize minimumSize( fontMetrics.width( "00:00" ), fontMetrics.height() );

    elapsedTimeLabel_->setMinimumSize( minimumSize );
    elapsedTimeLabel_->setAlignment( Qt::AlignCenter );
    elapsedTimeLabel_->hide();

    remainingTimeLabel_->setMinimumSize( minimumSize );
    remainingTimeLabel_->setAlignment( Qt::AlignCenter );
    remainingTimeLabel_->hide();

    connect( slider_, SIGNAL(sliderMoved(int)), SLOT(_processSliderMoved()) );
    connect( slider_, SIGNAL(sliderReleased()), SLOT(_processSliderReleased()) );

}

//_________________________________________________________________
void CurrentTrackWidget::setState( Mpd::State state )
{
    if( state == state_ ) return;
    state_ = state;

    Debug::Throw() << "CurrentTrackWidget::setState -"
        << " state: " << Base::toIntegralType( state_ )
        << " time: " << elapsedTime_  << ":" << totalTime_ << endl;

    switch( state_ )
    {
        case Mpd::State::Stopped:
        {
            totalTime_ = 0;
            elapsedTime_ = 0;
            _updateSlider();
            break;
        }

        default: break;
    }

    _updateLabels();

}

//_________________________________________________________________
void CurrentTrackWidget::setTotalTime( quint32 value )
{
    totalTime_ = value * precision_;
    slider_->setMaximum( totalTime_ );
    if( elapsedTime_ <= totalTime_ )
    {
        _updateSlider();
        _updateLabels();
    }
}

//_________________________________________________________________
void CurrentTrackWidget::seek( quint32 value )
{
    elapsedTime_ = value * precision_;
    if( elapsedTime_ <= totalTime_ )
    {
        _updateSlider();
        _updateLabels();
        if( state_ != Mpd::State::Stopped )
        {
            emit requestSeek( value );
            stopTimer();
        }
    }

}

//_________________________________________________________________
void CurrentTrackWidget::setElapsedTime( quint32 value )
{
    elapsedTime_ = value * precision_;
    if( elapsedTime_ <= totalTime_ )
    {
        _updateSlider();
        _updateLabels();
    }

    Debug::Throw() << "CurrentTrackWidget::setElapsedTime - " << elapsedTime_  << ":" << totalTime_ << endl;

}

//_________________________________________________________________
void CurrentTrackWidget::setSeekEnabled( bool value )
{ static_cast<Local::Slider*>( slider_ )->setSeekEnabled( value ); }

//_________________________________________________________________
void CurrentTrackWidget::startTimer()
{ if( !timer_.isActive() && totalTime_ > 0 && elapsedTime_ < totalTime_ ) timer_.start( 1000/precision_, this ); }

//_________________________________________________________________
void CurrentTrackWidget::stopTimer()
{ if( timer_.isActive() ) timer_.stop(); }

//_________________________________________________________________
void CurrentTrackWidget::setOrientation( const Qt::Orientation& orientation )
{
    Debug::Throw( "CurrentTrackWidget::setOrientation.\n" );
    if( orientation == slider_->orientation() ) return;

    auto boxLayout( static_cast<QBoxLayout*>( layout() ) );
    if( orientation == Qt::Vertical )
    {

        boxLayout->setDirection( QBoxLayout::TopToBottom );

        boxLayout->takeAt(1);
        boxLayout->insertWidget( 1, slider_, 1, Qt::AlignHCenter );

    } else {

        boxLayout->setDirection( QBoxLayout::LeftToRight );

        boxLayout->takeAt(1);
        boxLayout->insertWidget( 1, slider_, 1, Qt::AlignVCenter );

    }

    slider_->setOrientation( orientation );
    slider_->setInvertedAppearance( orientation == Qt::Vertical );
}

//_________________________________________________________________
void CurrentTrackWidget::timerEvent( QTimerEvent* event )
{
    if( event->timerId() == timer_.timerId() )
    {
        if( elapsedTime_ < totalTime_ )
        {
            ++elapsedTime_;
            _updateSlider();
            _updateLabels();
        }
    }

}

//_________________________________________________________________
void CurrentTrackWidget::_processSliderMoved()
{
    Debug::Throw( "CurrentTrackWidget::_processSliderMoved.\n" );
    elapsedTime_ = slider_->value();
}

//_________________________________________________________________
void CurrentTrackWidget::_processSliderReleased()
{
    Debug::Throw( "CurrentTrackWidget::_processSliderReleased.\n" );
    if( state_ != Mpd::State::Stopped )
    {
        emit requestSeek( static_cast<qreal>(slider_->value())/precision_ );
        stopTimer();
    }
}
//_________________________________________________________________
void CurrentTrackWidget::_updateLabels()
{
    if( totalTime_ == 0 || elapsedTime_ > totalTime_ )
    {

        elapsedTimeLabel_->hide();
        remainingTimeLabel_->hide();

    } else {

        elapsedTimeLabel_->setText( TrackInformation::durationString( elapsedTime_/precision_ ) );
        elapsedTimeLabel_->show();

        remainingTimeLabel_->setText( TrackInformation::durationString( (totalTime_ - elapsedTime_)/precision_ ) );
        remainingTimeLabel_->show();

    }

    return;
}

//_________________________________________________________________
void CurrentTrackWidget::_updateSlider()
{
    if( !slider_->isSliderDown() && elapsedTime_ <= totalTime_ )
    { slider_->setValue( elapsedTime_ ); }
}
