/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "DBusPlayerAdaptor.h"

#include "Application.h"
#include "CppUtil.h"
#include "CurrentTrackWidget.h"
#include "MainWindow.h"
#include "PlaybackControlToolBar.h"
#include "PlaylistWidget.h"
#include "VolumeControlMenu.h"

#include <QDBusMessage>
#include <QDBusConnection>

namespace DBus
{

    //________________________________________________________________________
    PlayerAdaptor::PlayerAdaptor( MainWindow *parent ):
        QDBusAbstractAdaptor(parent),
        Counter( "PlayerAdaptor" )
    {
        Debug::Throw( "PlayerAdaptor::PlayerAdaptor.\n" );
        connect( parent, SIGNAL(statusChanged(MainWindow::StatusFlags)), SLOT(_updateStatus(MainWindow::StatusFlags)) );
    }

    //________________________________________________________________________
    bool PlayerAdaptor::canGoNext() const
    { return parent()->playlistWidget().model().hasNextTrack(); }

    //________________________________________________________________________
    bool PlayerAdaptor::canGoPrevious() const
    { return parent()->playlistWidget().model().hasPreviousTrack(); }

    //________________________________________________________________________
    bool PlayerAdaptor::canPause() const
    { return parent()->playlistWidget().model().currentTrackIndex().isValid(); }

    //________________________________________________________________________
    bool PlayerAdaptor::canPlay() const
    { return parent()->playlistWidget().model().currentTrackIndex().isValid(); }

    //________________________________________________________________________
    bool PlayerAdaptor::canSeek() const
    {
        if( parent()->state() == Mpd::State::Stopped ) return false;

        const auto index = parent()->playlistWidget().model().currentTrackIndex();
        return index.isValid() ?
            !parent()->playlistWidget().model().get( index ).isAudioDeviceStream() :
            false;
    }

    //________________________________________________________________________
    QString PlayerAdaptor::loopStatus() const
    {
        switch( parent()->playbackControlToolBar().repeatMode() )
        {
            default:
            case Mpd::RepeatMode::RepeatNone: return QLatin1String( "None" );
            case Mpd::RepeatMode::RepeatOne: return QLatin1String( "Track" );
            case Mpd::RepeatMode::RepeatAll: return QLatin1String( "Playlist" );
        }
    }

    //________________________________________________________________________
    void PlayerAdaptor::setLoopStatus( const QString& value )
    {
        if( value == QLatin1String( "None" ) ) parent()->playbackControlToolBar().setRepeatMode( Mpd::RepeatMode::RepeatNone );
        else if( value == QLatin1String( "Track" ) ) parent()->playbackControlToolBar().setRepeatMode( Mpd::RepeatMode::RepeatOne );
        else if( value == QLatin1String( "Playlist" ) ) parent()->playbackControlToolBar().setRepeatMode( Mpd::RepeatMode::RepeatAll );
    }

    //________________________________________________________________________
    QVariantMap PlayerAdaptor::metadata() const
    {
        const auto index = parent()->playlistWidget().model().currentTrackIndex();
        if( !index.isValid() ) return QVariantMap();

        QVariantMap metadataMap;
        const auto track( parent()->playlistWidget().model().get( index ) );
        metadataMap.insert( "mpris:trackid", _trackId( track ) );
        if( !track.artist().isEmpty() )
        {
            metadataMap.insert("xesam:albumArtist", QStringList( track.artist() ) );
            metadataMap.insert("xesam:artist", QStringList( track.artist() ) );
        }

        if( !track.album().isEmpty() ) metadataMap.insert( "xesam:album", track.album() );
        if( !track.title().isEmpty() ) metadataMap.insert( "xesam:title", track.title() );
        if( track.duration() > 0 ) metadataMap.insert( "mpris:length", qlonglong(track.duration())*1000000 );

        if( track.year() > 0 ) metadataMap.insert( "xesam:contentCreated", QString("%04d").arg( track.year() ) );
        if( track.disc() > 0 ) metadataMap.insert( "xesam:discNumber", track.disc() );
        if( track.track() > 0 ) metadataMap.insert( "xesam:trackNumber", track.track() );

        if( !track.isAudioDeviceStream() )
        {
            const File musicDirectory( XmlOptions::get().raw( "MUSIC_DIR" ) );
            metadataMap.insert( "xesam:url", QString( "file://%1" ).arg( File( track.file() ).addPath( musicDirectory ).get() ) );
        }


        {
            CollectionItem item( track );
            item.setType( CollectionItem::Type::Album );
            const auto coverFile( CoverManager::fileName( item ) );
            if( coverFile.exists() )
            { metadataMap.insert("mpris:artUrl", QString( "file://%1" ).arg( coverFile.get() ) ); }
        }

        return metadataMap;

    }

    //________________________________________________________________________
    QString PlayerAdaptor::playbackStatus() const
    {
        switch( parent()->state() )
        {
            case Mpd::State::Playing: return QLatin1String( "Playing" );
            case Mpd::State::Paused: return QLatin1String( "Paused" );

            default:
            case Mpd::State::Stopped: return QLatin1String( "Stopped" );
        }
    }

    //________________________________________________________________________
    qlonglong PlayerAdaptor::position() const
    { return 1000000 * parent()->currentTrackWidget().elapsedTime(); }

    //________________________________________________________________________
    bool PlayerAdaptor::shuffle() const
    { return parent()->playbackControlToolBar().isRandomPlayback(); }

    //________________________________________________________________________
    void PlayerAdaptor::setShuffle( bool value )
    { parent()->playbackControlToolBar().setRandomPlayback( value ); }

    //________________________________________________________________________
    double PlayerAdaptor::volume() const
    { return double( parent()->volumeControlMenu().volume() )/100; }

    //________________________________________________________________________
    void PlayerAdaptor::setVolume(double value)
    { parent()->volumeControlMenu().setVolume( value * 100 ); }

    //________________________________________________________________________
    void PlayerAdaptor::Previous()
    { parent()->playbackControlToolBar().previousTrackAction().trigger(); }

    //________________________________________________________________________
    void PlayerAdaptor::Next()
    { parent()->playbackControlToolBar().nextTrackAction().trigger(); }

    //________________________________________________________________________
    void PlayerAdaptor::Play()
    {
        if( parent()->state() != Mpd::State::Playing )
        { parent()->playbackControlToolBar().startPlaybackAction().trigger(); }
    }

    //________________________________________________________________________
    void PlayerAdaptor::Pause()
    {
        if( parent()->state() == Mpd::State::Playing )
        { parent()->playbackControlToolBar().startPlaybackAction().trigger(); }
    }

    //________________________________________________________________________
    void PlayerAdaptor::PlayPause()
    { parent()->playbackControlToolBar().startPlaybackAction().trigger(); }

    //________________________________________________________________________
    void PlayerAdaptor::Seek(qlonglong offset)
    {
        parent()->currentTrackWidget().seek( qMax<qint32>( 0, qMin<qint32>(
            parent()->currentTrackWidget().elapsedTime() + offset/1000000,
            parent()->currentTrackWidget().totalTime() ) ) );
    }

    //________________________________________________________________________
    void PlayerAdaptor::SetPosition(const QDBusObjectPath &trackId, qlonglong position)
    {
        // get current track
        const auto index = parent()->playlistWidget().model().currentTrackIndex();
        if( !index.isValid() ) return;
        const auto track( parent()->playlistWidget().model().get( index ) );

        if( trackId.path() == _trackId( track ) )
        {
            parent()->currentTrackWidget().seek( qMax<qint32>( 0, qMin<qint32>(
                position/1000000,
                parent()->currentTrackWidget().totalTime() ) ) );
        }

    }

    //________________________________________________________________________
    void PlayerAdaptor::Stop()
    { parent()->playbackControlToolBar().stopPlaybackAction().trigger(); }

    //________________________________________________________________________
    void  PlayerAdaptor::_updateStatus( MainWindow::StatusFlags statusFlags )
    {
        QVariantMap map;
        if( statusFlags & MainWindow::Volume ) map.insert( "Volume", volume() );
        if( statusFlags & MainWindow::CurrentTrack ) map.insert( "Metadata", metadata() );
        if( statusFlags & MainWindow::PreviousTrack ) map.insert( "CanGoPrevious", canGoPrevious() );
        if( statusFlags & MainWindow::PreviousTrack ) map.insert( "CanGoNext", canGoNext() );

        if( statusFlags & MainWindow::PlaybackState )
        {
            map.unite( Base::makeT<QVariantMap>( {
                { "PlaybackStatus", playbackStatus() },
                { "Position", position() },
                { "CanPlay", canPlay() },
                { "CanPause", canPause() },
                { "CanSeek", canSeek() } } ) );
        }

        if( statusFlags & MainWindow::PlaybackMode )
        {
            map.unite( Base::makeT<QVariantMap>( {
                { "LoopStatus", loopStatus() },
                { "Shuffle", shuffle() } } ) );
        }

        _signalPropertyChanged( map );

    }

    //________________________________________________________________________
    QString PlayerAdaptor::_trackId( const TrackInformation& trackInformation ) const
    { return QString("/org/mpris/MediaPlayer2/Track/%1").arg( QString::number( trackInformation.trackId() ) ); }

    //________________________________________________________________________
    void PlayerAdaptor::_signalPropertyChanged(const QString &property, const QVariant &value)
    { _signalPropertyChanged( Base::makeT<QVariantMap>( {{property, value}} ) ); }

    //________________________________________________________________________
    void PlayerAdaptor::_signalPropertyChanged(const QVariantMap& map)
    {
        if( map.isEmpty() ) return;

        // create signal
        QDBusMessage signal = QDBusMessage::createSignal(
            "/org/mpris/MediaPlayer2",
            "org.freedesktop.DBus.Properties",
            "PropertiesChanged");

        // assign arguments
        signal.setArguments(Base::makeT<QVariantList>({
            "org.mpris.MediaPlayer2.Player",
            map,
            QStringList() }));

        // send
        QDBusConnection::sessionBus().send(signal);

    }

}
