/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AudioDeviceSpeed.h"
#include "ConfigurationDialog.h"
#include "Color.h"
#include "CppUtil.h"
#include "DefaultFolders.h"
#include "IconSize.h"
#include "PlaylistWriter.h"
#include "TrackInformation.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QHostAddress>

//_____________________________________________________
//* Default options installer
void installDefaultOptions()
{

    XmlOptions::get().setAutoDefault( true );

    // resource file
    XmlOptions::get().set( "RC_FILE", Option( File("musicboxrc").addPath( Util::config() ), Option::Flag::None ) );

    // mpd configuration file
    XmlOptions::get().set( "MPD_CONFIG_FILE", Option( File( "musicbox/mpd/mpd.conf" ).addPath(Util::config()), Option::Flag::None ) );
    XmlOptions::get().set( "MPD_DATA_DIR", Option( File( "musicbox/mpd" ).addPath(Util::data()), Option::Flag::None ) );
    XmlOptions::get().set( "MPD_CACHE_DIR", Option( File( "musicbox/mpd" ).addPath(Util::cache()), Option::Flag::None ) );
    XmlOptions::get().set<bool>( "MPD_CLOSE_DEAMON_ON_EXIT", true );

    // logging
    XmlOptions::get().set<bool>( "LOG_ENABLED", false );

    #if defined(Q_OS_WIN)
    // force use TCP socket on windows
    XmlOptions::get().set<bool>( "MPD_USE_TCP_SOCKET",  true );
    #else
    // there is no pulse audio on windows
    XmlOptions::get().set<bool>( "MPD_USE_PULSE_AUDIO", true );

    // on linux we can use both local and tcp. We default to local
    XmlOptions::get().set<bool>( "MPD_USE_TCP_SOCKET",  false );
    #endif

    XmlOptions::get().set<int>( "MPD_SOCKET_PORT", 6600 );
    XmlOptions::get().set( "MPD_SOCKET_HOST", QHostAddress( QHostAddress::LocalHost ).toString() );

    // control toolbar
    XmlOptions::get().set( "CONTROL_TOOLBAR", Option::create<bool>( true, Option::None ) );
    XmlOptions::get().set( "CONTROL_TOOLBAR_LOCATION", Option::create<int>( Qt::TopToolBarArea, Option::None ) );
    XmlOptions::get().set<int>( "CONTROL_TOOLBAR_ICON_SIZE", IconSize::Large );
    XmlOptions::get().set<int>( "CONTROL_TOOLBAR_TEXT_POSITION", Qt::ToolButtonIconOnly );

    // side panel
    XmlOptions::get().set<bool>( "SHOW_SIDE_PANEL_WIDGET", true );
    XmlOptions::get().set<int>( "SIDE_PANEL_TOOLBAR_ICON_SIZE", IconSize::Small );
    XmlOptions::get().set<int>( "SIDE_PANEL_TOOLBAR_TEXT_POSITION", Qt::ToolButtonFollowStyle );

    XmlOptions::get().set<int>( "SIDE_PANEL_CURRENT_WIDGET", 0 );

    // fading
    XmlOptions::get().set<bool>( "USE_FADE_ANIMATION", true );
    XmlOptions::get().set<int>( "FADE_ANIMATION_DURATION", 500 );

    // audio device
    XmlOptions::get().set<QString>( "AUDIO_DEVICE", "/dev/cdrom" );
    XmlOptions::get().set<int>( "AUDIO_DEVICE_SPEED", Base::toIntegralType( AudioDeviceSpeed::SpeedMode::Default ) );

    // playlist
    XmlOptions::get().set<int>( "PLAYLIST_ICON_SIZE", IconSize::Large );

    // collection
    XmlOptions::get().set( "MUSIC_DIR", DefaultFolders::get().file( DefaultFolders::Type::Music ) );
    XmlOptions::get().set<bool>( "UPDATE_COLLECTION_AT_STARTUP", true );
    XmlOptions::get().set<bool>( "WATCH_COLLECTION_CHANGES", true );

    // covers
    XmlOptions::get().set<bool>( "ENABLE_COVER_MANAGER", true );
    XmlOptions::get().set( "ALBUM_ART_CACHE_DIR", Option( File( "musicbox/AlbumArt" ).addPath(Util::cache()), Option::Flag::None ) );
    XmlOptions::get().set<bool>( "SHOW_ARTIST_ICON", true );

    // cddb
    XmlOptions::get().set<bool>( "ENABLE_CDDB", true );
    XmlOptions::get().set( "CDDB_CACHE_DIR", Option( File( "musicbox/Cddb" ).addPath(Util::cache()), Option::Flag::None ) );
    XmlOptions::get().set( "CDDB_HOST", "freedb.freedb.org" );
    XmlOptions::get().set<int>( "CDDB_PORT", 8880 );

    // lyrics
    XmlOptions::get().set<bool>( "ENABLE_LYRICS_MANAGER", true );
    XmlOptions::get().set( "LYRICS_CACHE_DIR", Option( File( "musicbox/Lyrics" ).addPath(Util::cache()), Option::Flag::None ) );
    XmlOptions::get().set<bool>( "LYRICS_EDITOR_WRAP", true );

    XmlOptions::get().keep( "LYRICS_PROVIDER_FILENAME" );
    XmlOptions::get().add( "LYRICS_PROVIDER_FILENAME", Option( ":/lyricsProviderList.xml" , Option::Flag::None ) );

    // audio streams
    XmlOptions::get().set( "AUDIO_STREAM_CACHE_DIR", Option( File( "musicbox/AudioStreams" ).addPath(Util::cache()), Option::Flag::None ) );

    // default playlist format
    XmlOptions::get().set<int>( "DEFAULT_PLAYLIST_FORMAT", Base::toIntegralType( PlaylistWriter::Format::M3u ) );

    // system tray
    XmlOptions::get().set<bool>( "USE_SYSTEMTRAY", true );
    XmlOptions::get().set<bool>( "USE_MONOCHROME_SYSTEMTRAY_ICON", true );
    XmlOptions::get().set<int>( "MAIN_WINDOW_VISIBILITY", Base::toIntegralType( ConfigurationDialog::WindowVisibilityMode::Restore ) );

    // notifications
    XmlOptions::get().set<bool>( "USE_NOTIFICATIONS", true );

    // dbus
    #if WITH_DBUS
    XmlOptions::get().set<bool>( "USE_DBUS_INTERFACE", true );
    #endif

    // collapsed state
    XmlOptions::get().set<bool>( "COLLAPSED", false );
    XmlOptions::get().set<bool>( "COLLAPSE_ON_DOUBLE_CLICK", true );

    // tooltips
    XmlOptions::get().set<bool>( "SHOW_TOOLTIPS", true );

    // splitters
    XmlOptions::get().set<int>( "SIDE_PANEL_WIDTH", 200 );
    XmlOptions::get().set<unsigned int>( "SIDE_PANEL_MASK", 0x3f );

    // banner
    XmlOptions::get().set<bool>( "SHOW_BANNER", true );
    XmlOptions::get().set<bool>( "BANNER_USE_SYSTEM_COLORS", true );
    XmlOptions::get().set<Base::Color>( "BANNER_FOREGROUND_COLOR", Base::Color( Qt::white ) );
    XmlOptions::get().set<Base::Color>( "BANNER_BACKGROUND_COLOR", Base::Color( Qt::black ) );

    // misc
    XmlOptions::get().set<int>( "TRACK_INFORMATION_FILTER", TrackInformation::Default );
    XmlOptions::get().set<bool>( "SHOW_STOP_BUTTON", true );
    XmlOptions::get().set<bool>( "RESUME_PLAYBACK_ON_START", false );
    XmlOptions::get().setAutoDefault( false );

}
