/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "DeleteTracksDialog.h"

#include "Debug.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "TreeView.h"
#include "XmlOptions.h"

#include <QHeaderView>

#include <algorithm>

//_________________________________________________________
DeleteTracksDialog::SimpleModel::SimpleModel( QObject* parent ):
    ListModel( parent ),
    Counter( "DeleteTracksDialog::SimpleModel::SimpleModel" )
{}

//_________________________________________________________
QVariant DeleteTracksDialog::SimpleModel::data( const QModelIndex& index, int role ) const
{

    // check index
    if( !contains( index ) ) return QVariant();

    // return text associated to file and column
    if( role == Qt::DisplayRole )
    {
        const auto& item( get(index) );
        QString out = item.file();
        if( !item.title().isEmpty() )
        {

            if( !item.artist().isEmpty() ) out += tr( " (%1 by %2)" ).arg( item.title(), item.artist() );
            else out += tr( " (%1)" ).arg( item.title() );

        } else if( !item.artist().isEmpty() ) {

            out += tr( " (by %1)" ).arg( item.artist() );

        }

        return out;

    }

    return QVariant();
}

//_________________________________________________________
DeleteTracksDialog::DeleteTracksDialog( QWidget* parent ):
QuestionDialog( parent )
{
    Debug::Throw( "DeleteTracksDialog::DeleteTracksDialog.\n" );

    setWindowTitle( tr( "Delete Tracks" ) );
    setOptionName( "DELETE_TRACKS_DIALOG" );

    // create list
    auto list = new TreeView( this );
    list->setModel( &model_ );
    list->header()->hide();
    list->setSelectionMode( QAbstractItemView::NoSelection );
    list->setFindEnabled( false );
    setDetails( list );
    showDetails();

    okButton().setIcon( IconEngine::get( IconNames::Delete ) );
    okButton().setText( tr( "Delete" ) );
}

//_________________________________________________________
bool DeleteTracksDialog::requestsCompleted() const
{
    Debug::Throw( "DeleteTracksDialog::requestsCompleted.\n" );

    for( auto&& iter = items_.begin(); iter != items_.end(); ++iter )
    {
        switch( iter->type() )
        {

            default: continue;
            case CollectionItem::Type::Artist:
            {
                const auto ftor = TrackInformation::SameArtistFTor( *iter );
                if( std::none_of( iter, items_.end(),
                    [&ftor]( const CollectionItem& item )
                    { return item.type() == CollectionItem::Type::Track && ftor( item ); } ) )
                { return false; }
                break;
            }

            case CollectionItem::Type::Album:
            {
                const auto ftor = TrackInformation::SameAlbumFTor( *iter );
                if( std::none_of( iter, items_.end(),
                    [&ftor]( const CollectionItem& item )
                    { return item.type() == CollectionItem::Type::Track && ftor( item ); } ) )
                { return false; }
                break;
            }
        }

    }

    return true;
}

//_________________________________________________________
void DeleteTracksDialog::reset()
{
    Debug::Throw( "DeleteTracksDialog::reset.\n" );
    items_.clear();
    model_.clear();
    _updateText();
    _updateButtons();
}

//_________________________________________________________
void DeleteTracksDialog::addItems( const CollectionItem::List& items )
{

    Debug::Throw( "DeleteTracksDialog::addItems.\n" );

    // append
    items_.append( items );

    // loop over items and add to log widget
    CollectionItem::List trackItems;
    File musicDirectory( XmlOptions::get().raw( "MUSIC_DIR" ) );
    for( auto item:items )
    {
        if( item.type() != CollectionItem::Type::Track ) continue;
        auto file( item.file() );
        if( !file.isAbsolute() ) file.addPath( musicDirectory );
        item.setFile( file );
        trackItems.append( item );
    }

    model_.add( trackItems );

    _updateText();
    _updateButtons();

}

//_________________________________________________________
void DeleteTracksDialog::deleteTracks()
{
    Debug::Throw( "DeleteTracksDialog::deleteTracks.\n" );

    // get music directory
    File musicDirectory( XmlOptions::get().raw( "MUSIC_DIR" ) );
    for( const auto& item:items_ )
    {
        if( item.type() != CollectionItem::Type::Track ) continue;
        auto file( item.file() );
        if( !file.isAbsolute() ) file.addPath( musicDirectory );
        file.remove();
    }

}

//_________________________________________________________
void DeleteTracksDialog::_updateButtons()
{
    Debug::Throw( "DeleteTracksDialog::_updateButtons.\n" );
    okButton().setEnabled(
        std::any_of( items_.begin(), items_.end(), CollectionItem::SameTypeFTor( CollectionItem::Type::Track ) ) &&
        requestsCompleted() );
}

//_________________________________________________________
void DeleteTracksDialog::_updateText()
{
    Debug::Throw( "DeleteTracksDialog::_updateText.\n" );

    auto tracks = std::count_if( items_.begin(), items_.end(), CollectionItem::SameTypeFTor( CollectionItem::Type::Track ) );
    if( tracks == 0 ) setText( QString() );
    else if( tracks == 1 ) setText( tr(
        "Do you really want to delete this track? "
        "It will be removed from local collection "
        "and from file system." ) );
    else setText( tr(
        "Do you really want to delete %1 tracks? "
        "They will be removed from local collection "
        "and from file system." ).arg( tracks ) );
}
