/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "LyricsManager.h"

#include "LyricsProvider.h"
#include "LyricsProviderWikia.h"
#include "XmlError.h"
#include "XmlOptions.h"

#include <QDir>
#include <QDomDocument>

//____________________________________________________
LyricsManager::LyricsManager( QObject* parent ):
    QObject( parent ),
    Counter( "LyricsManager" ),
    networkAccessManager_( new QNetworkAccessManager( this ) )
{

    Debug::Throw( "LyricsManager::LyricsManager.\n" );

    // make sure lyrics path exists
    QDir().mkpath( XmlOptions::get().raw( "LYRICS_CACHE_DIR" ) );

    // install providers
    for( const auto& option:XmlOptions::get().specialOptions( "LYRICS_PROVIDER_FILENAME" ) )
    {

        Debug::Throw() << "LyricsManager::LyricsManager - reading from " << option.raw() << endl;

        // try open file
        QFile file( option.raw() );
        if( !file.open( QIODevice::ReadOnly ) ) continue;

        // read
        QDomDocument document;
        document.setContent( &file );

        const auto nodeList( document.elementsByTagName( "provider" ) );
        for( int i = 0; i < nodeList.count(); ++i )
        {

            // get element
            const auto element( nodeList.at( i ).toElement() );

            // check parent
            if( element.parentNode().nodeName() != QLatin1String( "lyricproviders") ) continue;

            // get provider name from attribute
            QString providerName;
            const auto attributes( element.attributes() );
            for( int i=0; i<attributes.count(); i++ )
            {
                const auto attribute( attributes.item( i ).toAttr() );
                if( attribute.isNull() ) continue;
                const auto name( attribute.name() );
                const auto value( attribute.value() );

                if( name == QLatin1String( "name" ) )
                {
                    providerName = value;
                    break;
                }

            }

            // check if not already in list
            if( providerName.isEmpty() || std::any_of( lyricsProviderList_.begin(), lyricsProviderList_.end(), BaseLyricsProvider::SameNameFTor( providerName ) ) )
            { continue; }

            // add new provider
            if( providerName == QLatin1String( "lyrics.wikia.com" ) )
            {

                _addLyricsProvider( new LyricsProviderWikia( this ) );

            } else {

                _addLyricsProvider( new LyricsProvider( this, element ) );

            }

        }

    }

}

//_______________________________________________________________________________
File LyricsManager::fileName( const TrackInformation& trackInformation )
{ return File( "Lyrics_{" + QString::number( trackInformation.hashKey() ) + "}.txt" ).addPath( File( XmlOptions::get().raw( "LYRICS_CACHE_DIR" ) ) ); }

//____________________________________________________
void LyricsManager::get( const TrackInformation& trackInformation )
{
    Debug::Throw() << "LyricsManager::get - track: " << trackInformation << endl;
    if( trackInformation.artist().isEmpty() || trackInformation.title().isEmpty() )
    { return; }

    // check if being fetched already
    if( trackInformationHash_.contains( trackInformation.hashKey() ) )
    { return; }

    // try fetch from disk
    if( _fromDisk( trackInformation ) ) return;

    // check enable state and list of providers
    if( !enabled_ || lyricsProviderList_.isEmpty() )
    {
        emit lyricsNotFound( trackInformation );
        return;
    }

    // insert hash key
    trackInformationHash_.insert( trackInformation.hashKey(), 0 );

    // try find lyrics using first provider in list
    lyricsProviderList_.front()->get( trackInformation );

}

//____________________________________________________
void LyricsManager::_lyricsFound( const TrackInformation& trackInformation, const LyricsInformation& lyricsInformation )
{

    Debug::Throw( "LyricsManager::_lyricsFound.\n" );

    // check lyricsInformation
    if( lyricsInformation.lyrics().isEmpty() ) return _lyricsNotFound( trackInformation );

    QDomDocument document;

    // append node
    document.appendChild( lyricsInformation.domElement( document ) );

    // save to disc
    QFile file( fileName( trackInformation ) );
    if( file.open( QIODevice::WriteOnly ) )
    { file.write( document.toByteArray() ); }

    // remove from hash
    trackInformationHash_.remove( trackInformation.hashKey() );

    // emit signal
    emit lyricsFound( trackInformation, lyricsInformation );

    return;

}

//____________________________________________________
void LyricsManager::_lyricsNotFound( const TrackInformation& trackInformation )
{
    Debug::Throw( "LyricsManager::_lyricsNotFound.\n" );

    // find item in hash
    auto iter = trackInformationHash_.find( trackInformation.hashKey() );
    if( iter == trackInformationHash_.end() ) emit lyricsNotFound( trackInformation );
    else {

        if( (++iter.value()) >= lyricsProviderList_.size() )
        {

            // all providers have been found. Remove from hash and emit signal
            trackInformationHash_.remove( trackInformation.hashKey() );
            emit lyricsNotFound( trackInformation );

        } else {

            // try get lyrics with next provider
            lyricsProviderList_.at(iter.value())->get( trackInformation );

        }
    }

}

//____________________________________________________
void LyricsManager::_addLyricsProvider( BaseLyricsProvider* provider )
{

    Debug::Throw() << "LyricsManager::_addLyricsProvider - adding " << provider->name() << endl;

    // assign access manager
    provider->setNetworkAccessManager( networkAccessManager_ );

    // append to list
    lyricsProviderList_.append( provider );

    // setup connections
    connect( provider, SIGNAL(lyricsFound(TrackInformation,LyricsInformation)), this, SLOT(_lyricsFound(TrackInformation,LyricsInformation)) );
    connect( provider, SIGNAL(lyricsNotFound(TrackInformation)), this, SLOT(_lyricsNotFound(TrackInformation)) );
}

//____________________________________________________
bool LyricsManager::_fromDisk( const TrackInformation& trackInformation )
{
    Debug::Throw( "LyricsManager::_fromDisk.\n" );

    // try open  file
    QFile qfile( fileName( trackInformation ) );
    if( !qfile.open( QIODevice::ReadOnly ) ) return false;

    // read content
    XmlError error;
    QDomDocument document;
    if( !document.setContent( qfile.readAll(), error ) ) return false;

    LyricsInformation lyricsInformation( document.documentElement() );
    if( lyricsInformation.lyrics().isEmpty() ) return false;
    else {

        emit lyricsFound( trackInformation, lyricsInformation );
        return true;

    }

}
