/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "MpdController.h"

#include "CppUtil.h"
#include "MpdBaseSocket.h"
#include "ObjectDeleter.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QDir>
#include <QElapsedTimer>
#include <QProcess>

#include <memory>

namespace Mpd
{

    //______________________________________________
    Controller::Controller( QObject* parent ):
        QObject( parent ),
        Counter( "Mpd::Controller" )
    {}

    //______________________________________________
    Controller::~Controller() = default;

    //______________________________________________
    bool Controller::checkServerConfiguration() const
    {

        Debug::Throw( "Mpd::Controller::checkServerConfiguration.\n" );

        // create server configuration
        serverConfiguration_.createDirectories();

        // write configuration
        const File configurationFile( XmlOptions::get().raw( "MPD_CONFIG_FILE" ) );
        QDir().mkpath( configurationFile.path() );
        return serverConfiguration_.write( configurationFile );

    }

    //______________________________________________
    bool Controller::serverIsRunning() const
    {
        Debug::Throw( "Mpd::Controller::serverIsRunning.\n" );
        std::unique_ptr<BaseSocket, Base::ObjectDeleter> socket( BaseSocket::create( serverConfiguration_ ) );
        return serverReady_ = socket->connect( serverConfiguration_ );
    }

    //______________________________________________
    bool Controller::startServer()
    {
        Debug::Throw( "Mpd::Controller::startServer.\n" );

        // run command
        const QString mpdCommand( XmlOptions::get().raw( "MPD_COMMAND" ) );
        const QString configurationFile( XmlOptions::get().raw( "MPD_CONFIG_FILE" ) );
        return QProcess::startDetached( mpdCommand, QStringList( configurationFile ) );

    }

    //______________________________________________
    bool Controller::stopServer()
    {
        Debug::Throw( "Mpd::Controller::stopServer.\n" );
        const QString mpdCommand( XmlOptions::get().raw( "MPD_COMMAND" ) );
        const QString configurationFile( XmlOptions::get().raw( "MPD_CONFIG_FILE" ) );
        return QProcess::startDetached( mpdCommand, Base::makeT<QStringList>({ configurationFile, "--kill" }) );
    }

    //_______________________________________________________________________
    bool Controller::waitForStarted( int msecs )
    {
        Debug::Throw( "Controller::waitForStarted.\n" );

        // do nothing if already started
        if( serverReady_ || serverIsRunning() ) return true;

        if( !serverStartedTimer_.isActive() )
        { serverStartedTimer_.start( 100, this ); }

        QElapsedTimer timer;
        timer.start();

        while( ( msecs < 0 || timer.elapsed() < msecs ) && !serverReady_ )
        { qApp->processEvents(); }

        if( serverStartedTimer_.isActive() ) serverStartedTimer_.stop();

        return serverReady_;

    }

    //_______________________________________________________________________
    bool Controller::waitForStopped( int msecs )
    {
        Debug::Throw( "Controller::waitForStopped.\n" );

        // do nothing if already stopped
        if( !( serverReady_ || serverIsRunning() ) ) return true;

        if( !serverStoppedTimer_.isActive() )
        { serverStoppedTimer_.start( 100, this ); }

        QElapsedTimer timer;
        timer.start();

        while( ( msecs < 0 || timer.elapsed() < msecs ) && serverReady_ )
        { qApp->processEvents(); }

        if( serverStoppedTimer_.isActive() ) serverStoppedTimer_.stop();

        return !serverReady_;

    }

    //_______________________________________________________________________
    void Controller::timerEvent( QTimerEvent* event )
    {

        if( event->timerId() == serverStartedTimer_.timerId() )
        {

            if( serverIsRunning() ) serverStartedTimer_.stop();

        } else if( event->timerId() == serverStoppedTimer_.timerId() ) {

            if( !serverIsRunning() ) serverStoppedTimer_.stop();

        } else return QObject::timerEvent( event );

    }

}
