/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "MpdInformationDialog.h"

#include "CppUtil.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "TrackInformation.h"

#include <algorithm>

namespace Mpd
{
    //____________________________________________________________
    InformationDialog::InformationDialog( QWidget* parent ):
        CustomDialog( parent, CloseButton|Separator )
    {

        Debug::Throw( "InformationDialog::InformationDialog.\n" );
        setWindowTitle( tr( "Server Information" ) );
        setOptionName( "MPD_STATS_DIALOG" );


        // icon
        auto hLayout = new QHBoxLayout;
        hLayout->setSpacing(10);
        hLayout->setMargin(0);
        mainLayout().addLayout( hLayout );

        {
            auto vLayout = new QVBoxLayout;
            vLayout->setMargin(0);
            vLayout->setSpacing(0);
            hLayout->addLayout( vLayout, 0 );

            auto label = new QLabel( this );
            label->setPixmap( IconEngine::get( IconNames::DialogInformation ).pixmap( iconSize() ) );
            vLayout->addWidget( label, 0 );
            vLayout->addStretch( 1 );
        }

        {
            auto vLayout = new QVBoxLayout;
            vLayout->setMargin(0);
            vLayout->setSpacing(0);
            hLayout->addLayout( vLayout, 0 );

            label_ = new QLabel( this );
            label_->setWordWrap( true );
            vLayout->addWidget( label_, 1 );
            vLayout->addStretch( 1 );
       }

    }

    //____________________________________________________________
    void InformationDialog::clear()
    {
        Debug::Throw( "InformationDialog::clear.\n" );

        const auto strings = Base::makeT<QStringList>( {
            mpdVersion_,
            upTime_,
            playTime_,
            artists_,
            albums_,
            tracks_,
            duration_,
            lastUpdate_
        } );

        // hide all items
        for( auto string:strings ) string.clear();

        // clear handlers and tags
        handlers_.clear();
        tags_.clear();

        // clear text
        label_->clear();

    }

    //____________________________________________________________
    void InformationDialog::update( const QString& reply )
    {
        Debug::Throw( "InformationDialog::update.\n" );

        auto parse = []( const QString& input, const QString& key )
        { return input.startsWith(key) ? std::make_pair( true, input.mid( key.length() ) ) : std::make_pair( false, QString() ); };

        const auto replyList( reply.split( '\n', QString::SkipEmptyParts ) );
        std::pair<bool, QString> result;

        QRegExp regExp( "(\\S+)://" );
        for( const auto& replyString:replyList )
        {
            if( ( result = parse( replyString, "uptime: " ) ).first ) upTime_ = TrackInformation::durationString( result.second.toUInt() );
            else if( ( result = parse( replyString, "playtime: " ) ).first ) playTime_ = TrackInformation::durationString( result.second.toUInt() );
            else if( ( result = parse( replyString, "artists: " ) ).first ) artists_ = result.second;
            else if( ( result = parse( replyString, "albums: " ) ).first ) albums_ = result.second;
            else if( ( result = parse( replyString, "songs: " ) ).first ) tracks_ = result.second;
            else if( ( result = parse( replyString, "db_playtime: " ) ).first ) duration_ = TrackInformation::durationString( result.second.toUInt() );
            else if( ( result = parse( replyString, "db_update: " ) ).first ) lastUpdate_ = TimeStamp( result.second.toInt() ).toString();
            else if( ( result = parse( replyString, "tagtype: " ) ).first ) tags_.append( result.second );
            else if( ( result = parse( replyString, "handler: " ) ).first && result.second.indexOf( regExp ) >= 0 ) handlers_.append( regExp.cap(1) );
        }

        std::sort( handlers_.begin(), handlers_.end() );
        std::sort( tags_.begin(), tags_.end() );

        // update text
        label_->setText(
            QLatin1String("<p><table>")+
            tr("<tr><td colspan=\"2\"><b>Server Information</b></td></tr>"
            "<tr><td align=\"right\">Version:&nbsp;</td><td>%1</td></tr>"
            "<tr><td align=\"right\">Uptime:&nbsp;</td><td>%2</td></tr>"
            "<tr><td align=\"right\">Playing:&nbsp;</td><td>%3</td></tr>"
            "<tr><td align=\"right\">Handlers:&nbsp;</td><td>%4</td></tr>"
            "<tr><td align=\"right\">Tags:&nbsp;</td><td>%5</td></tr>")
            .arg( mpdVersion_, upTime_, playTime_, handlers_.join( ", " ), tags_.join( ", " ) ) +
            QLatin1String("<tr/>")+
            tr("<tr><td colspan=\"2\"><b>Database Statistics</b></td></tr>"
            "<tr><td align=\"right\">Artists:&nbsp;</td><td>%1</td></tr>"
            "<tr><td align=\"right\">Albums:&nbsp;</td><td>%2</td></tr>"
            "<tr><td align=\"right\">Songs:&nbsp;</td><td>%3</td></tr>"
            "<tr><td align=\"right\">Duration:&nbsp;</td><td>%4</td></tr>"
            "<tr><td align=\"right\">Updated:&nbsp;</td><td>%5</td></tr>")
            .arg( artists_, albums_, tracks_, duration_, lastUpdate_ )+
            QLatin1String("</table></p>") );

    }

}
