/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "MpdServerConfiguration.h"

#include "CppUtil.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QDir>
#include <QSet>

namespace Mpd
{

    //_____________________________________________________________
    ServerConfiguration ServerConfiguration::fromOptions()
    {

        Debug::Throw( "Mpd::ServerConfiguration::fromOptions.\n" );
        ServerConfiguration out;
        out.musicDirectory_ = File( XmlOptions::get().raw( "MUSIC_DIR" ) );
        out.playlistDirectory_ = File( "playlists" ).addPath( File( XmlOptions::get().raw( "MPD_DATA_DIR" ) ) );
        out.stickerFile_ = File( "sticker.sql" ).addPath( File( XmlOptions::get().raw( "MPD_DATA_DIR" ) ) );
        out.useTcpSocket_ = XmlOptions::get().get<bool>( "MPD_USE_TCP_SOCKET" );
        if( out.useTcpSocket_ )
        {
            Debug::Throw( "Mpd::ServerConfiguration::fromOptions - using tcp socket.\n" );
            out.host_ = File( XmlOptions::get().raw( "MPD_SOCKET_HOST" ) );
            out.port_ = XmlOptions::get().get<int>( "MPD_SOCKET_PORT" );

        } else {

            Debug::Throw( "Mpd::ServerConfiguration::fromOptions - using local socket.\n" );
            out.host_ = File( "socket" ).addPath( File( XmlOptions::get().raw( "MPD_DATA_DIR" ) ) );

        }

        out.dbFile_ = File( "mpd.db" ).addPath( File( XmlOptions::get().raw( "MPD_CACHE_DIR" ) ) );
        out.pidFile_ = File( "pid" ).addPath( File( XmlOptions::get().raw( "MPD_CACHE_DIR" ) ) );
        out.stateFile_ = File( "state" ).addPath( File( XmlOptions::get().raw( "MPD_CACHE_DIR" ) ) );
        out.logFile_ = File( "mpd.log" ).addPath( File( XmlOptions::get().raw( "MPD_CACHE_DIR" ) ) );

        #if !defined(Q_OS_WIN)
        out.usePulseAudio_ = XmlOptions::get().get<bool>( "MPD_USE_PULSE_AUDIO" );
        #endif

        return out;
    }

    //_____________________________________________________________
    void ServerConfiguration::createDirectories() const
    {

        using FileSet = QSet<File>;
        FileSet directories = Base::makeT<FileSet>(
        {
            playlistDirectory_,
            stickerFile_.path(),
            dbFile_.path(),
            pidFile_.path(),
            stateFile_.path(),
            File( host_ ).path()
        } );

        for( const auto& path:directories )
        { QDir().mkpath( path ); }

        // remove previous log file
        logFile_.remove();

    }

    //_____________________________________________________________
    bool ServerConfiguration::write( const File& destination ) const
    {

        Debug::Throw() << "Mpd::ServerConfiguration::write - destination: " << destination << endl;

        QFile out( destination );
        if( !out.open( QIODevice::WriteOnly ) )
        {
            Debug::Throw(0) << "Mpd::ServerConfiguration::write - cannot write to file " << destination << endl;
            return false;
        }

        QTextStream stream( &out );
        auto addEntries = []( QTextStream& out, std::initializer_list< std::pair<QString, QString> >&& entries )
        {
            for( const auto& pair:std::move( entries ) )
            { out << pair.first << " \"" << pair.second << "\"" << endl; }
        };

        addEntries( stream, {
            { "user", Util::user() },
            { "music_directory", musicDirectory_ },
            { "playlist_directory", playlistDirectory_ },
            { "sticker_file", stickerFile_ },
            { "db_file", dbFile_ },
            { "pid_file", pidFile_ },
            { "state_file", stateFile_ },
            { "log_file", logFile_ },
            { "bind_to_address", host_ }
        } );

        if( useTcpSocket_ )
        {
            addEntries( stream, {{ "port", QString::number( port_ ) }});
        };

        stream << "metadata_to_use \"artist,album,title,track,date,disc\"" << endl;

        #if !defined(Q_OS_WIN)
        // for window, we do not define the audio output and rely on mpd to figure it out itself
        if( usePulseAudio_ )
        {

            stream << "audio_output { " << endl;
            stream << "   type \"pulse\"" << endl;
            stream << "   name \"MPD Pulse Audio Stream\"" << endl;
            stream << "}" << endl;

        } else {

            stream << "audio_output { " << endl;
            stream << "    type \"alsa\"" << endl;
            stream << "    name \"Output\"" << endl;
            stream << "}" << endl;

        }
        #endif

        stream << "mixer_type \"software\"" << endl;
        stream << "audio_buffer_size \"8192\"" << endl;
        stream << "filesystem_charset \"UTF-8\"" << endl;
        stream << "id3v1_encoding \"UTF-8\"" << endl;

        out.close();
        return true;
    }

}
