
/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "PlaybackControlToolBar.h"

#include "BaseMainWindow.h"
#include "CppUtil.h"
#include "CurrentTrackWidget.h"
#include "CustomToolButton.h"
#include "Debug.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "IconSize.h"
#include "IconSizeMenu.h"
#include "MainWindow.h"
#include "Singleton.h"
#include "ToolBarMenu.h"
#include "ToolButtonStyleMenu.h"
#include "VolumeControlMenu.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QWheelEvent>

namespace Local
{

    // volume button
    class VolumeControlButton: public QToolButton
    {

        Q_OBJECT

        public:

        // constructor
        VolumeControlButton( QWidget* parent = nullptr ):
           QToolButton( parent )
        { setProperty( "_kde_toolButton_noMenuArrow", true ); }

        protected:

        // wheel event
        void wheelEvent( QWheelEvent* event ) override
        {
            if( !menu() ) return;
            auto menu( static_cast<VolumeControlMenu*>( this->menu() ) );
            if( !menu->isMuted() )
            { QApplication::sendEvent( &menu->volumeSlider(), new QWheelEvent( *event ) ); }
        }

    };

}

//_________________________________________________________
PlaybackControlToolBar::PlaybackControlToolBar( QWidget* parent ):
    CustomToolBar( tr( "Playback Controls" ), parent, "CONTROL_TOOLBAR" )
{
    Debug::Throw( "CustomToolBar::CustomToolBar.\n" );

    setTransparent( true );
    setIconSize( IconSize( IconSize::Large ).get() );
    setToolButtonStyle( Qt::ToolButtonIconOnly );

    _installActions();
    setState( Mpd::State::Stopped );

    // fading connections
    connect( volumeControlMenu_, SIGNAL(fadeCompleted()), SLOT(_fadeCompleted()) );
    connect( volumeControlMenu_, SIGNAL(fadeAborted()), SLOT(_fadeCompleted()) );

    updateVolumeIcon();
    updateVolumeToolTip();

    // configuration
    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );

    // orientaition
    connect( this, SIGNAL(orientationChanged(Qt::Orientation)), currentTrackWidget_, SLOT(setOrientation(Qt::Orientation)) );

}

//_________________________________________________________
Mpd::RepeatMode PlaybackControlToolBar::repeatMode() const
{
    if( repeatOneAction_->isChecked() ) return Mpd::RepeatMode::RepeatOne;
    else if( repeatAllAction_->isChecked() ) return Mpd::RepeatMode::RepeatAll;
    else return Mpd::RepeatMode::RepeatNone;
}

//_________________________________________________________
bool PlaybackControlToolBar::fadingEnabled() const
{ return volumeControlMenu_->fadingEnabled(); }

//_________________________________________________________
bool PlaybackControlToolBar::isFading() const
{ return volumeControlMenu_->isFading(); }

//_________________________________________________________
void PlaybackControlToolBar::setState( Mpd::State state )
{
    Debug::Throw() << "CustomToolBar::setState - state: " << Base::toIntegralType( state ) << endl;
    state_ = state;
    switch( state_ )
    {
        case Mpd::State::Paused:
        case Mpd::State::Stopped:
        startPlaybackAction_->setIcon( IconEngine::get( IconNames::StartPlayback ) );
        startPlaybackAction_->setText( tr( "Play" ) );
        break;

        case Mpd::State::Playing:
        startPlaybackAction_->setIcon( IconEngine::get( IconNames::PausePlayback ) );
        startPlaybackAction_->setText( tr( "Pause" ) );
        break;
    }

}

//_________________________________________________________
void PlaybackControlToolBar::setRepeatMode( Mpd::RepeatMode repeatMode )
{
    Debug::Throw( "PlaybackControlToolBar::setRepeatMode.\n" );
    switch( repeatMode )
    {
        default:
        case Mpd::RepeatMode::RepeatNone:
        repeatNoneAction_->setChecked( true );
        break;

        case Mpd::RepeatMode::RepeatOne:
        repeatOneAction_->setChecked( true );
        break;

        case Mpd::RepeatMode::RepeatAll:
        repeatAllAction_->setChecked( true );
        break;
    }
}

//_________________________________________________________
void PlaybackControlToolBar::abortFade()
{
    if( fadingEnabled() && isFading() )
    volumeControlMenu_->abortFade();
}


//______________________________________________________________________
void PlaybackControlToolBar::contextMenuEvent( QContextMenuEvent* event )
{
    Debug::Throw( "PlaybackControlToolBar::contextMenuEvent.\n" );

    MainWindow* mainwindow( qobject_cast<MainWindow*>( window() ) );
    if( !mainwindow ) return;

    std::unique_ptr<ToolBarMenu> menu( static_cast<ToolBarMenu*>(mainwindow->createPopupMenu() ) );

    menu->toolButtonStyleAction().setVisible( true );
    menu->toolButtonStyleMenu().select( (Qt::ToolButtonStyle) XmlOptions::get().get<int>( "CONTROL_TOOLBAR_TEXT_POSITION" ) );

    menu->iconSizeAction().setVisible(true);
    menu->iconSizeMenu().select( (IconSize::Size) XmlOptions::get().get<int>( "CONTROL_TOOLBAR_ICON_SIZE" ) );

    connect( &menu->toolButtonStyleMenu(), SIGNAL(styleSelected(int)), SLOT(_updateToolButtonStyle(int)) );
    connect( &menu->iconSizeMenu(), SIGNAL(iconSizeSelected(IconSize::Size)), SLOT(_updateToolButtonIconSize(IconSize::Size)) );

    // move and show menu
    menu->adjustSize();
    menu->exec( event->globalPos() );

}

//_________________________________________________________
void PlaybackControlToolBar::updateVolumeIcon()
{
    Debug::Throw( "PlaybackControlToolBar::updateVolumeIcon.\n" );
    if( volumeControlMenu_->isMuted() ) volumeControlButton_->setIcon( IconEngine::get( IconNames::VolumeMuted ) );
    else
    {
        const int volume = volumeControlMenu_->volumeSlider().value();
        if( volume == 0 ) volumeControlButton_->setIcon( IconEngine::get( IconNames::VolumeMuted ) );
        else if( volume < 33 ) volumeControlButton_->setIcon( IconEngine::get( IconNames::VolumeLow ) );
        else if( volume < 66 ) volumeControlButton_->setIcon( IconEngine::get( IconNames::VolumeMedium ) );
        else volumeControlButton_->setIcon( IconEngine::get( IconNames::VolumeHigh ) );
    }
}

//_________________________________________________________
void PlaybackControlToolBar::updateVolumeToolTip()
{
    Debug::Throw( "PlaybackControlToolBar::toggleMute.\n" );
    const int volume = volumeControlMenu_->volumeSlider().value();
    auto tooltip = tr( "Volume: %1%" ).arg( volume );
    if( volumeControlMenu_->isMuted() ) tooltip += tr( " (muted)" );
    volumeControlButton_->setToolTip( tooltip );
}

//_________________________________________________________
void PlaybackControlToolBar::_updateConfiguration()
{
    Debug::Throw( "CustomToolBar::_updateConfiguration.\n" );
    stopPlaybackAction_->setVisible( XmlOptions::get().get<bool>( "SHOW_STOP_BUTTON" ) );

    // icon size
    const IconSize iconSize( static_cast<IconSize::Size>( XmlOptions::get().get<int>( "CONTROL_TOOLBAR_ICON_SIZE" ) ) );

    // text label for toolbars
    const int toolButtonTextPosition( XmlOptions::get().get<int>( "CONTROL_TOOLBAR_TEXT_POSITION" ) );
    const auto style = static_cast<Qt::ToolButtonStyle>( toolButtonTextPosition >= 0 ?
        toolButtonTextPosition:
        this->style()->styleHint( QStyle::SH_ToolButtonStyle ) );

    // assign to toolbar
    setIconSize( iconSize.get() );
    setToolButtonStyle( style );

    volumeControlButton_->setIconSize( iconSize.get() );
    volumeControlButton_->setToolButtonStyle( style );

}

//____________________________________________________________
void PlaybackControlToolBar::_updateToolButtonStyle( int style )
{

    Debug::Throw( "PlaybackControlToolBar::_updateToolButtonStyle.\n" );
    XmlOptions::get().set<int>( "CONTROL_TOOLBAR_TEXT_POSITION", static_cast<int>(style) );
    _updateConfiguration();

}

//____________________________________________________________
void PlaybackControlToolBar::_updateToolButtonIconSize( IconSize::Size size )
{

    Debug::Throw( "PlaybackControlToolBar::_updateToolButtonIconSize.\n" );
    XmlOptions::get().set<int>( "CONTROL_TOOLBAR_ICON_SIZE", size );
    _updateConfiguration();

}

//_________________________________________________________
void PlaybackControlToolBar::_stopPlayback()
{

    Debug::Throw( "CustomToolBar::_stopPlayback.\n" );

    setState( Mpd::State::Stopped );
    if( fadingEnabled() )
    {

        if( !isFading() ) volumeControlMenu_->startFade();
        else volumeControlMenu_->abortFade();

    } else {

        emit requestStopPlayback();
    }

}

//_________________________________________________________
void PlaybackControlToolBar::_toggleStartPlayback()
{

    switch( state_ )
    {

        case Mpd::State::Playing:
        {
            setState( Mpd::State::Paused );
            emit requestPausePlayback();
            break;
        }

        case Mpd::State::Stopped:
        {
            setState( Mpd::State::Playing );
            emit requestStartPlayback();
            break;
        }

        case Mpd::State::Paused:
        {
            setState( Mpd::State::Playing );
            emit requestResumePlayback();
            break;
        }

    }

}

//_________________________________________________________
void PlaybackControlToolBar::_repeatModeActionTriggered( QAction* action )
{
    Debug::Throw( "PlaybackControlToolBar::_repeatModeActionTriggered.\n" );
    if( !locked_ )
    {
        if( action == repeatNoneAction_ ) emit requestRepeatNone();
        else if( action == repeatOneAction_ ) emit requestRepeatOne();
        else if( action == repeatAllAction_ ) emit requestRepeatAll();
    }
}

//_________________________________________________________
void PlaybackControlToolBar::_toggleRandomPlayback( bool value )
{
    Debug::Throw( "PlaybackControlToolBar::_toggleRandomPlayback.\n" );
    if( !locked_ ) emit requestRandomPlayback( value );
}

//_________________________________________________________
void PlaybackControlToolBar::_fadeCompleted()
{
    Debug::Throw( "PlaybackControlToolBar::_fadeCompleted.\n" );

    // do stop
    setState( Mpd::State::Stopped );
    emit requestStopPlayback();
}

//_________________________________________________________
void PlaybackControlToolBar::_installActions()
{
    Debug::Throw( "CustomToolBar::_installActions.\n" );
    addAction( previousTrackAction_ = new QAction( IconEngine::get( IconNames::PreviousTrack ), tr( "Previous Track" ), this ) );
    connect( previousTrackAction_, SIGNAL(triggered()), SIGNAL(requestPreviousTrack()) );
    connect( previousTrackAction_, SIGNAL(triggered()), SLOT(abortFade()) );

    addAction( startPlaybackAction_ = new QAction( IconEngine::get( IconNames::StartPlayback ), tr( "Play" ), this ) );
    connect( startPlaybackAction_, SIGNAL(triggered()), SLOT(abortFade()) );
    connect( startPlaybackAction_, SIGNAL(triggered()), SLOT(_toggleStartPlayback()) );
    startPlaybackAction_->setShortcut( Qt::Key_Space );

    addAction( stopPlaybackAction_ = new QAction( IconEngine::get( IconNames::StopPlayback ), tr( "Stop" ), this ) );
    connect( stopPlaybackAction_, SIGNAL(triggered()), SLOT(_stopPlayback()) );

    addAction( nextTrackAction_ = new QAction( IconEngine::get( IconNames::NextTrack ), tr( "Next Track" ), this ) );
    connect( nextTrackAction_, SIGNAL(triggered()), SIGNAL(requestNextTrack()) );
    connect( nextTrackAction_, SIGNAL(triggered()), SLOT(abortFade()) );

    // repeat modes
    auto actionGroup = new QActionGroup( this );
    actionGroup->setExclusive( true );
    connect( actionGroup, SIGNAL(triggered(QAction*)), SLOT(_repeatModeActionTriggered(QAction*)) );

    repeatNoneAction_ = new QAction( tr( "No Repeat" ), this );
    repeatNoneAction_->setCheckable( true );
    repeatNoneAction_->setChecked( true );
    actionGroup->addAction( repeatNoneAction_ );

    repeatOneAction_ = new QAction( tr( "Repeat One" ), this );
    repeatOneAction_->setCheckable( true );
    repeatOneAction_->setChecked( false );
    actionGroup->addAction( repeatOneAction_ );

    repeatAllAction_ = new QAction( tr( "Repeat All" ), this );
    repeatAllAction_->setCheckable( true );
    repeatAllAction_->setChecked( false );
    actionGroup->addAction( repeatAllAction_ );

    // shuffle
    randomPlaybackAction_  = new QAction( tr( "Random" ), this );
    randomPlaybackAction_->setCheckable( true );
    randomPlaybackAction_->setChecked( false );

    connect( randomPlaybackAction_, SIGNAL(triggered(bool)), SLOT(_toggleRandomPlayback(bool)) );

    // current track widget
    addWidget( currentTrackWidget_ = new CurrentTrackWidget( this ) );

    // volume control
    addWidget( volumeControlButton_ = new Local::VolumeControlButton( this ) );
    volumeControlButton_->setIcon( IconEngine::get( IconNames::VolumeHigh ) );
    volumeControlButton_->setText( tr( "Volume" ) );
    volumeControlButton_->setMenu( volumeControlMenu_ = new VolumeControlMenu );
    volumeControlButton_->setPopupMode( QToolButton::InstantPopup );

    connect( volumeControlMenu_, SIGNAL(muteStateChanged(bool)), SLOT(updateVolumeIcon()) );
    connect( volumeControlMenu_, SIGNAL(volumeChanged(int)), SLOT(updateVolumeIcon()) );

    connect( volumeControlMenu_, SIGNAL(muteStateChanged(bool)), SLOT(updateVolumeToolTip()) );
    connect( volumeControlMenu_, SIGNAL(volumeChanged(int)), SLOT(updateVolumeToolTip()) );
}

#include "PlaybackControlToolBar.moc"
