#ifndef PlaybackControlToolBar_h
#define PlaybackControlToolBar_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CppUtil.h"
#include "CustomToolBar.h"
#include "IconSize.h"
#include "MpdDefines.h"

#include <QAction>
#include <QList>
#include <QToolButton>

class VolumeControlMenu;
class CurrentTrackWidget;

class PlaybackControlToolBar: public CustomToolBar
{

    Q_OBJECT;

    public:

    //* constructor
    explicit PlaybackControlToolBar( QWidget* = nullptr );

    //*@name accessors
    //@{

    Mpd::State state() const
    { return state_; }

    //* repeat mode
    Mpd::RepeatMode repeatMode() const;

    //* true if random playback is checked
    bool isRandomPlayback() const
    { return randomPlaybackAction_->isChecked(); }

    //* current track widget
    CurrentTrackWidget& currentTrackWidget() const
    { return *currentTrackWidget_; }

    //* volume control
    VolumeControlMenu& volumeControlMenu() const
    { return *volumeControlMenu_; }

    //* true if fading is enabled
    bool fadingEnabled() const;

    //* true if fading
    bool isFading() const;

    //@}

    //*@name actions
    //@{

    QAction& startPlaybackAction() const
    { return *startPlaybackAction_; }

    QAction& stopPlaybackAction() const
    { return *stopPlaybackAction_; }

    QAction& previousTrackAction() const
    { return *previousTrackAction_; }

    QAction& nextTrackAction() const
    { return *nextTrackAction_; }

    QList<QAction*> repeatModeActions() const
    {
        return Base::makeT<QList<QAction*> >(
        {
            repeatNoneAction_,
            repeatOneAction_,
            repeatAllAction_
        } );
    }

    //* random playback
    QAction& randomPlaybackAction() const
    { return *randomPlaybackAction_; }

    //@}

    //*@name modifiers
    //@{

    //* state
    void setState( Mpd::State );

    //* repeat mode
    void setRepeatMode( Mpd::RepeatMode );

    //* set random playback
    void setRandomPlayback( bool value )
    { randomPlaybackAction_->setChecked( value ); }

    //* lock
    void lock()
    { locked_ = true; }

    //* unlock
    void unlock()
    { locked_ = false; }

    //* request start playback
    void emitRequestStartPlayback()
    { emit requestStartPlayback(); }

    //* request start playback
    void emitRequestPausePlayback()
    { emit requestPausePlayback(); }

    //* request start playback
    void emitRequestResumePlayback()
    { emit requestResumePlayback(); }

    //* request start playback
    void emitRequestStopPlayback()
    { emit requestStopPlayback(); }

    //* request previous song
    void emitRequestPreviousTrack()
    { emit requestPreviousTrack(); }

    //* request next song
    void emitRequestNextTrack()
    { emit requestNextTrack(); }

    //@}

    Q_SIGNALS:

    //* request start playback
    void requestStartPlayback();

    //* request start playback
    void requestPausePlayback();

    //* request start playback
    void requestResumePlayback();

    //* request start playback
    void requestStopPlayback();

    //* request previous song
    void requestPreviousTrack();

    //* request next song
    void requestNextTrack();

    //* no reapeat
    void requestRepeatNone();

    //* repeat one
    void requestRepeatOne();

    //* repeat all
    void requestRepeatAll();

    //* shuffle
    void requestRandomPlayback( bool );

    public Q_SLOTS:

    //* update volume icon
    void updateVolumeIcon();

    //* volume tooltip
    void updateVolumeToolTip();

    //* abort fading
    void abortFade();

    protected:

    //* context menu
    void contextMenuEvent( QContextMenuEvent* ) override;

    private Q_SLOTS:

    //* update configuration
    void _updateConfiguration();

    //* toolbar text position
    void _updateToolButtonStyle( int );

    //* toolbar text position
    void _updateToolButtonIconSize( IconSize::Size );

    //* stop playback
    void _stopPlayback();

    //* start/pause playback
    void _toggleStartPlayback();

    //* toggle repeat none
    void _repeatModeActionTriggered( QAction* );

    //* toggle random playback
    void _toggleRandomPlayback( bool );

    //* fade completed
    void _fadeCompleted();

    private:

    //* create actions
    void _installActions();

    //* locked
    /** prevents round trips to mpd */
    bool locked_ = false;

    //* state
    Mpd::State state_ = Mpd::State::Stopped;

    //*@name actions
    //@{
    QAction* startPlaybackAction_ = nullptr;
    QAction* stopPlaybackAction_ = nullptr;
    QAction* previousTrackAction_ = nullptr;
    QAction* nextTrackAction_ = nullptr;

    QAction* repeatNoneAction_ = nullptr;
    QAction* repeatOneAction_ = nullptr;
    QAction* repeatAllAction_ = nullptr;

    QAction* randomPlaybackAction_ = nullptr;

    //@]

    //* current track widget
    CurrentTrackWidget* currentTrackWidget_ = nullptr;

    //* volume control
    QToolButton* volumeControlButton_ = nullptr;
    VolumeControlMenu* volumeControlMenu_ = nullptr;

};

#endif
