/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "PlaylistModel.h"

#include "CollectionItem.h"
#include "CppUtil.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "AudioStreamItem.h"
#include "StoredPlaylistItem.h"

#if WITH_CDIO_PARANOIA
#include "AudioDeviceTrack.h"
#endif

#include <QMimeData>
#include <QUrl>

//_______________________________________________________________________________________
PlaylistModel::PlaylistModel( QObject* parent ):
    TrackInformationModel( parent )
{ _setState( Mpd::State::Stopped ); }

//_______________________________________________________________________________________
Qt::ItemFlags PlaylistModel::flags( const QModelIndex& index ) const
{
    return index.isValid() ?
        Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsDragEnabled:
        Qt::ItemIsDropEnabled;
}

//__________________________________________________________________
QStringList PlaylistModel::mimeTypes() const
{
    #if WITH_CDIO_PARANOIA
    return Base::makeT<QStringList>({ TrackInformation::MimeType, CollectionItem::MimeType, StoredPlaylistItem::MimeType, AudioStreamItem::MimeType, AudioDeviceTrack::MimeType, "text/uri-list" });
    #else
    return Base::makeT<QStringList>({ TrackInformation::MimeType, CollectionItem::MimeType, StoredPlaylistItem::MimeType, AudioStreamItem::MimeType, "text/uri-list" });
    #endif
}

//______________________________________________________________________
QMimeData* PlaylistModel::mimeData(const QModelIndexList &indexes) const
{

    if( indexes.empty() ) return nullptr;

    // create mime data
    auto mime = new QMimeData;
    mime->setData( TrackInformation::MimeType, nullptr );
    return mime;

}

//__________________________________________________________________
void PlaylistModel::setState( Mpd::State state )
{
    if( state != state_ )
    {
        emit layoutAboutToBeChanged();
        _setState( state );
        emit layoutChanged();
    }
}

//__________________________________________________________________
bool PlaylistModel::dropMimeData(const QMimeData* data , Qt::DropAction action, int row, int, const QModelIndex&)
{

    // adjust row
    if( row < 0 ) row = rowCount();

    // Drag from Keyword model
    if( data->hasFormat( TrackInformation::MimeType ) ) emit requestMoveTracks( row );
    else if( data->hasFormat( CollectionItem::MimeType ) ) emit requestAddFromCollection( row );
    else if( data->hasFormat( StoredPlaylistItem::MimeType ) ) emit requestAddFromStoredPlaylist( row );
    else if( data->hasFormat( AudioStreamItem::MimeType ) ) emit requestAddFromAudioStream( row );
    #if WITH_CDIO_PARANOIA
    else if( data->hasFormat( AudioDeviceTrack::MimeType ) ) emit requestAddFromAudioDevice( row );
    #endif
    else if( data->hasUrls() ) {

        File::List files;
        for( const auto& url:data->urls() )
        {

            // check that local file
            #if QT_VERSION >= 0x040800
            if( !url.isLocalFile() ) continue;
            #endif

            // get file and check existence
            const File file( url.path() );
            if( !file.exists() ) continue;

            // in case of directory list content, recursively
            if( file.isDirectory() ) files.append( file.listFiles( static_cast<File::ListFlags>(File::Recursive|File::FollowLinks) ) );
            else files.append( file );

        }

        if( !files.empty() ) emit requestAddFiles( files, row );

    }

    return true;
}

//__________________________________________________________________
bool PlaylistModel::setCurrentTrackIndex( const QModelIndex& index )
{
    Debug::Throw() << "PlaylistModel::setCurrentTrackIndex - index: " << index.row() << endl;
    if( index != currentTrackIndex_ )
    {

        emit layoutAboutToBeChanged();
        currentTrackIndex_ = index;
        emit layoutChanged();
        return true;

    } else return false;
}

//__________________________________________________________________
bool PlaylistModel::selectPreviousTrack()
{
    Debug::Throw( "PlaylistModel::selectPreviousTrack.\n" );
    const auto newIndex =  _previousTrackIndex();
    if( contains( newIndex ) )
    {
        emit layoutAboutToBeChanged();
        currentTrackIndex_ = newIndex;
        emit layoutChanged();
        return true;

    } else return false;
}

//__________________________________________________________________
bool PlaylistModel::selectNextTrack()
{
    Debug::Throw( "PlaylistModel::selectNextTrack.\n" );
    const auto newIndex = _nextTrackIndex();
    if( contains( newIndex ) )
    {
        emit layoutAboutToBeChanged();
        currentTrackIndex_ = newIndex;
        emit layoutChanged();
        return true;

    } else return false;
}

//__________________________________________________________________
void PlaylistModel::_setState( Mpd::State state )
{
    state_ = state;
    switch( state )
    {
        case Mpd::State::Playing: stateIcon_ = IconEngine::get( IconNames::StartPlayback ); break;
        case Mpd::State::Paused: stateIcon_ = IconEngine::get( IconNames::PausePlayback ); break;
        case Mpd::State::Stopped: stateIcon_ = IconEngine::get( IconNames::StopPlayback ); break;
    }

}

//__________________________________________________________________
QModelIndex PlaylistModel::_previousTrackIndex() const
{
    if( currentTrackIndex_.isValid() )
    {

        const auto previousTrackIndex = index( currentTrackIndex_.row()-1, 0 );
        if( repeatMode_ == Mpd::RepeatMode::RepeatAll && !previousTrackIndex.isValid() ) return index( rowCount()-1, 0 );
        else return previousTrackIndex;

    } else return index( rowCount()-1, 0 );

}

//__________________________________________________________________
QModelIndex PlaylistModel::_nextTrackIndex() const
{
    if( currentTrackIndex_.isValid() )
    {

        const auto nextTrackIndex = index( currentTrackIndex_.row()+1, 0 );
        if( repeatMode_ == Mpd::RepeatMode::RepeatAll && !nextTrackIndex.isValid() ) return index( 0, 0 );
        else return nextTrackIndex;

    } else return index( 0, 0 );

}
