/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "PlaylistView.h"

#include "Debug.h"
#include "PlaylistDelegate.h"
#include "PlaylistModel.h"

#include <QHeaderView>

//_________________________________________________________
PlaylistView::PlaylistView( QWidget* parent ):
    TreeView( parent ),
    clickCounter_( this, 2 )
{
    Debug::Throw(  "PlaylistView::PlaylistView.\n" );

    setItemDelegate( new PlaylistDelegate( this ) );
    header()->hide();
    setSelectionMode( QAbstractItemView::ContiguousSelection );
    setDragEnabled( true );
    setAcceptDrops( true );
    setDropIndicatorShown( true );
    setMouseTracking( true );
    setFindEnabled( true );
    _setUseEmbeddedWidgets( true );

    setHorizontalScrollBarPolicy( Qt::ScrollBarAlwaysOff );

}

//_________________________________________________________
void PlaylistView::mousePressEvent( QMouseEvent* event )
{

    if( event->button() != Qt::LeftButton ) return TreeView::mousePressEvent( event );

    auto position( event->pos() );
    const auto index = indexAt( event->pos() );
    const auto visualRect( this->visualRect( index ) );
    position -= visualRect.topLeft();

    // cast item delegate
    const auto delegate = static_cast<const PlaylistDelegate*>( itemDelegate() );
    const auto inAlbum = delegate->isInAlbum( position, viewOptions(), index );
    if( inAlbum )
    {

        // process event
        TreeView::mousePressEvent( event );

        // update selection
        _updateSelection( index );

    } return TreeView::mousePressEvent( event );

}

//_________________________________________________________
void PlaylistView::mouseReleaseEvent( QMouseEvent* event )
{

    if( event->button() != Qt::LeftButton ) return TreeView::mouseReleaseEvent( event );

    clickCounter_.increment();

    auto position( event->pos() );
    auto index = indexAt( event->pos() );
    auto visualRect( this->visualRect( index ) );
    position -= visualRect.topLeft();

    // cast item delegate
    const auto delegate = static_cast<const PlaylistDelegate*>( itemDelegate() );
    const auto inAlbum = delegate->isInAlbum( position, viewOptions(), index );
    if( inAlbum )
    {

        // process event
        TreeView::mouseReleaseEvent( event );

        // update selection
        _updateSelection( index );

    }

    return TreeView::mouseReleaseEvent( event );

}

//_________________________________________________________
void PlaylistView::_updateSelection( const QModelIndex& index )
{

    Debug::Throw( "PlaylistView::_updateSelection.\n" );

    // cast model
    const auto& model = *static_cast<const PlaylistModel*>( this->model() );

    // store album
    const auto& album( model.get( index ).album() );
    const auto& channel( model.get( index ).channel() );
    for( QModelIndex copy = model.index( index.row()+1, 0 ); copy.isValid(); copy = model.index( copy.row()+1, 0 ) )
    {
        const auto track( model.get( copy ) );
        if( track.album() == album && track.channel() == channel )
        {

            selectionModel()->select( copy, QItemSelectionModel::Select|QItemSelectionModel::Rows );

        } else break;
    }

}
