/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "PlaylistWidget.h"

#include "BaseStatusBar.h"
#include "ContextMenu.h"
#include "CoverDisplayDialog.h"
#include "Debug.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "IconSize.h"
#include "PlaylistView.h"
#include "PlaylistFindWidget.h"
#include "ScrollBarMonitor.h"
#include "TrackInformationToolTipWidget.h"
#include "XmlOptions.h"

#include <QLayout>

#include <algorithm>

//________________________________________________________
PlaylistWidget::PlaylistWidget( QWidget* parent ):
    QWidget( parent ),
    Counter( "PlaylistWidget" )
{
    Debug::Throw( "PlaylistWidget::PlaylistWidget.\n" );

    _installActions();

    auto vLayout = new QVBoxLayout;
    vLayout->setMargin(0);
    vLayout->setSpacing(2);
    setLayout( vLayout );

    auto findWidget = new PlaylistFindWidget( this );
    vLayout->addWidget( findWidget );
    findWidget->hide();

    vLayout->addWidget( list_ = new PlaylistView( this ) );
    list_->setModel( &model_ );
    list_->setFindWidget( findWidget );

    scrollBarMonitor_ = new ScrollBarMonitor( list_ );

    // status bar
    vLayout->addWidget( statusBar_ = new BaseStatusBar( this ) );
    statusBar_->addLabel(1);
    statusBar_->addLabels(2, 0);
    statusBar_->addClock();
    statusBar_->label(0).setMinimumHeight(
        IconSize::Small
        + 2*statusBar_->label(0).margin() + 1 );

    // set icons for playback mode labels and hide
    statusBar_->label(1).setPixmap( IconEngine::get( IconNames::Repeat ).pixmap( IconSize( IconSize::Small ).get() ) );
    statusBar_->label(1).hide();

    statusBar_->label(2).setPixmap( IconEngine::get( IconNames::Random ).pixmap( IconSize( IconSize::Small ).get() ) );
    statusBar_->label(2).hide();

    // tooltip widget
    toolTipWidget_ = new TrackInformationToolTipWidget( this );

    connect( list_, SIGNAL(activated(QModelIndex)), &model_, SLOT(setCurrentTrackIndex(QModelIndex)) );
    connect( list_, SIGNAL(hovered(QModelIndex)), SLOT(_showToolTip(QModelIndex)) );
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(updateActions()) );

    {
        // context menu
        contextMenu_ = new ContextMenu( list_ );
        contextMenu_->addAction( addStreamToFavoritesAction_ );
        contextMenu_->addAction( removeTrackAction_ );
        contextMenu_->addAction( clearPlaylistAction_ );
        contextMenu_->addAction( savePlaylistAction_ );
        contextMenu_->addSeparator();
        contextMenu_->addAction( displayCoverAction_ );
        #if WITH_TAGLIB
        contextMenu_->addAction( editTagsAction_ );
        #endif

        contextMenu_->setHideDisabledActions( true );
        connect( contextMenu_, SIGNAL(aboutToShow()), SLOT(updateActions()) );
    }

}

//________________________________________________________
bool PlaylistWidget::setCurrentTrackIndex( int index )
{
    Debug::Throw( "PlaylistWidget::setCurrentTrackIndex.\n" );
    auto modelIndex( model_.index( index, 0 ) );
    if( model_.setCurrentTrackIndex( modelIndex ) )
    {

        list_->scrollTo( modelIndex );
        return true;

    } else return false;
}

//________________________________________________________
bool PlaylistWidget::selectPreviousTrack()
{
    Debug::Throw( "PlaylistWidget::selectPreviousTrack.\n" );
    if( model_.selectPreviousTrack() )
    {
        list_->scrollTo( model_.currentTrackIndex() );
        return true;
    } else return false;
}

//________________________________________________________
bool PlaylistWidget::selectNextTrack()
{
    Debug::Throw( "PlaylistWidget::selectNextTrack.\n" );
    if( model_.selectNextTrack() )
    {
        list_->scrollTo( model_.currentTrackIndex() );
        return true;
    } else return false;
}

//________________________________________________________
void PlaylistWidget::saveState()
{
    list_->setUpdatesEnabled(false);
    _saveCurrentIndex();
    _saveCurrentTrack();
    _saveSelectedTracks();
    scrollBarMonitor_->save();
}

//________________________________________________________
void PlaylistWidget::restoreState()
{
    scrollBarMonitor_->restore();
    _restoreSelectedTracks();
    _restoreCurrentTrack();
    _restoreCurrentIndex();
    list_->setUpdatesEnabled(true);
}

//____________________________________________________
void PlaylistWidget::updateTrack( const TrackInformation& oldTrack, const TrackInformation& newTrack )
{
    Debug::Throw( "PlaylistWidget::updateTrack.\n" );

    const auto& trackList( model_.get() );
    const auto iter = std::find_if(
        trackList.begin(), trackList.end(),
        TrackInformation::SameFileFTor( oldTrack ) );

    if( iter != trackList.end() )
    { model_.replace( model_.index( *iter ), newTrack ); }

}

//________________________________________________________
void PlaylistWidget::updateActions()
{
    Debug::Throw( "PlaylistWidget::updateActions.\n" );
    const auto selection( model_.get( list_->selectionModel()->selectedRows() ) );
    const bool hasSelection( !selection.empty() );
    removeTrackAction_->setEnabled( hasSelection );

    const bool empty( model_.empty() );
    clearPlaylistAction_->setEnabled( !empty );
    savePlaylistAction_->setEnabled( !empty );

    addStreamToFavoritesAction_->setEnabled( std::any_of( selection.begin(), selection.end(), TrackInformation::IsAudioStreamFTor() ) );

    displayCoverAction_->setEnabled( selection.size() == 1 && !model_.icon( selection.front() ).isNull() );

    #if WITH_TAGLIB
    editTagsAction_->setEnabled( std::any_of( selection.begin(), selection.end(), TrackInformation::IsEditableFTor() ) );
    #endif

}

//________________________________________________________
void PlaylistWidget::_showToolTip( const QModelIndex& index )
{
    Debug::Throw( "PlaylistWidget::_showToolTip.\n" );

    // check index and window
    if( !index.isValid() ) toolTipWidget_->hide();
    else {

        // get collection item and check
        const auto& item( model_.get( index ) );
        auto icon( model_.icon( item ) );
        if( icon.isNull() ) icon = IconEngine::get( IconNames::Track );

        // assign item
        static_cast<TrackInformationToolTipWidget*>(toolTipWidget_)->setTrackInformation( item, icon );

        // rect
        toolTipWidget_->setIndexRect(
            list_->visualRect( index ).
            translated( list_->viewport()->mapToGlobal( QPoint( 0, 0 ) ) ) );

        // show
        toolTipWidget_->showDelayed();

    }

}

//________________________________________________________
void PlaylistWidget::_displayCover()
{
    Debug::Throw( "PlaylistWidget::_displayCover.\n" );

    const auto selectedIndexes( list_->selectionModel()->selectedRows() );
    if( selectedIndexes.size() != 1 ) return;

    // get current item, convert to collection item and assign Album type
    CollectionItem item( model_.get( selectedIndexes.front() ) );
    item.setType( CollectionItem::Type::Album );
    const auto icon( model_.icon( item ) );
    if( icon.isNull() ) return;

    CoverDisplayDialog( this, item, icon ).exec();

}

//________________________________________________________
void PlaylistWidget::_addSelectionToFavorites()
{
    Debug::Throw( "PlaylistWidget::_addSelectionToFavorites.\n" );

    // get selection
    auto selection( model_.get( list_->selectionModel()->selectedRows() ) );

    // remove all non AudioStream tracks
    selection.erase( std::remove_if( selection.begin(), selection.end(), Base::Functor::UnaryFalse<TrackInformation, &TrackInformation::isAudioStream>() ), selection.end() );

    // check
    if( selection.isEmpty() ) return;

    // convert to AudioStreamItem
    AudioStreamItem::List items;
    std::transform( selection.begin(), selection.end(), std::back_inserter( items ),
        []( const TrackInformation& track )
        {
            AudioStreamItem item( AudioStreamItem::ProviderType::Favorites, AudioStreamItem::Type::Channel, track.file() );
            item.setChannel( track.channel() );
            return item;
        } );

    emit requestAddToFavorites( items );
}

//________________________________________________________
#if WITH_TAGLIB
void PlaylistWidget::_editTags()
{
    Debug::Throw( "PlaylistWidget::_editTags.\n" );

    // get selection
    const auto tracks( model_.get( list_->selectionModel()->selectedRows() ) );
    if( tracks.empty() ) return;

    File musicDirectory( XmlOptions::get().raw( "MUSIC_DIR" ) );
    musicDirectory.addTrailingSlash();

    File::List files;
    std::transform( tracks.begin(), tracks.end(), std::back_inserter( files ),
        [&musicDirectory]( const TrackInformation& track )
        {
            auto file( track.file() );
            if( !file.isAbsolute() ) file.addPath( musicDirectory );
            return file;
        });

    emit requestEditTags( files );
}
#endif

//________________________________________________________
void PlaylistWidget::_installActions()
{

    Debug::Throw( "PlaylistWidget::_installActions.\n" );
    addAction( removeTrackAction_ = new QAction( IconEngine::get( IconNames::Remove ), tr( "Remove From Playlist" ), this ) );
    removeTrackAction_->setShortcut( QKeySequence::Delete );
    removeTrackAction_->setShortcutContext( Qt::WidgetWithChildrenShortcut );
    removeTrackAction_->setEnabled( false );

    addAction( clearPlaylistAction_ = new QAction( IconEngine::get( IconNames::EditClear ), tr( "Clear Playlist" ), this ) );
    clearPlaylistAction_->setShortcut( Qt::CTRL + Qt::Key_K );

    addAction( savePlaylistAction_ = new QAction( IconEngine::get( IconNames::SaveAs ), tr( "Save Playlist..." ), this ) );
    savePlaylistAction_->setShortcut( QKeySequence::Save );
    savePlaylistAction_->setEnabled( false );

    addAction( addStreamToFavoritesAction_ = new QAction( IconEngine::get( IconNames::AddFavorite ), tr( "Add Stream to Favorites" ), this ) );
    connect( addStreamToFavoritesAction_, SIGNAL(triggered()), SLOT(_addSelectionToFavorites()) );
    addStreamToFavoritesAction_->setEnabled( false );

    addAction( displayCoverAction_ = new QAction( IconEngine::get( IconNames::DisplayCover ), tr( "Display cover" ), this ) );
    connect( displayCoverAction_, SIGNAL(triggered()), SLOT(_displayCover()) );
    displayCoverAction_->setEnabled( false );

    #if WITH_TAGLIB
    addAction( editTagsAction_ = new QAction( IconEngine::get( IconNames::Edit ), tr( "Edit Track Details..." ), this ) );
    editTagsAction_->setShortcut( Qt::Key_F2 );
    editTagsAction_->setShortcutContext( Qt::WidgetWithChildrenShortcut );
    connect( editTagsAction_, SIGNAL(triggered()), SLOT(_editTags()) );
    editTagsAction_->setEnabled( false );
    #endif
}

//________________________________________________________
void PlaylistWidget::_saveCurrentTrack()
{
    const auto index( model_.currentTrackIndex() );
    if( index.isValid() ) currentTrackId_ = model_.get( index ).trackId();
    else currentTrackId_ = -1;
}

//________________________________________________________
void PlaylistWidget::_restoreCurrentTrack()
{
    if( currentTrackId_ < 0 ) return;
    const auto tracks( model_.get() );
    const auto iter = std::find_if( tracks.begin(), tracks.end(), TrackInformation::SameTrackIdFTor( currentTrackId_ ) );
    if( iter != tracks.end() ) model_.setCurrentTrackIndex( model_.index( *iter ) );
}

//________________________________________________________
void PlaylistWidget::_saveSelectedTracks()
{
    const auto tracks( model_.get( list_->selectionModel()->selectedRows() ) );
    selectedTracks_.clear();
    std::transform( tracks.begin(), tracks.end(), std::back_inserter( selectedTracks_ ),
        []( const TrackInformation& track ) { return track.trackId(); } );
}

//________________________________________________________
void PlaylistWidget::_restoreSelectedTracks()
{
    if( selectedTracks_.empty() ) return;

    list_->selectionModel()->clear();

    const auto tracks( model_.get() );
    for( const auto& id:selectedTracks_ )
    {
        const auto iter = std::find_if( tracks.begin(), tracks.end(), TrackInformation::SameTrackIdFTor( id ) );
        if( iter != tracks.end() )
        { list_->selectionModel()->select( model_.index( *iter ), QItemSelectionModel::Select|QItemSelectionModel::Rows ); }
    }
}

//________________________________________________________
void PlaylistWidget::_saveCurrentIndex()
{
    const auto index( list_->selectionModel()->currentIndex() );
    if( index.isValid() ) currentIndexId_ = model_.get( index ).trackId();
    else currentIndexId_ = -1;
}

//________________________________________________________
void PlaylistWidget::_restoreCurrentIndex()
{
    if( currentIndexId_ < 0 ) return;
    const auto tracks( model_.get() );
    const auto iter = std::find_if( tracks.begin(), tracks.end(), TrackInformation::SameTrackIdFTor( currentIndexId_ ) );
    if( iter != tracks.end() ) list_->selectionModel()->setCurrentIndex( model_.index( *iter ), QItemSelectionModel::Current|QItemSelectionModel::Rows );
}

