#ifndef PlaylistWidget_h
#define PlaylistWidget_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AudioStreamItem.h"
#include "Counter.h"
#include "PlaylistModel.h"

#include <QAction>
#include <QWidget>

class BaseStatusBar;
class ContextMenu;
class ScrollBarMonitor;
class TreeView;
class ToolTipWidget;

class PlaylistWidget: public QWidget, private Base::Counter<PlaylistWidget>
{

    Q_OBJECT

    public:

    //* constructor
    explicit PlaylistWidget( QWidget* = nullptr );

    //*@name accessors
    //@{

    //* list
    TreeView& list() const
    { return *list_; }

    //* model
    const PlaylistModel& model() const
    { return model_; }

    //* statusbar
    BaseStatusBar& statusBar() const
    { return *statusBar_; }

    //* context menu
    ContextMenu& contextMenu() const
    { return *contextMenu_; }

    //@}

    //*@name actions
    //@{

    //* remove track action
    QAction& removeTrackAction() const
    { return *removeTrackAction_; }

    //* clear playlist action
    QAction& clearPlaylistAction() const
    { return *clearPlaylistAction_; }

    //* save playlist action
    QAction& savePlaylistAction() const
    { return *savePlaylistAction_; }

    //* tooltip widget
    ToolTipWidget& toolTipWidget() const
    { return *toolTipWidget_; }

    //@}

    //*@name modifiers
    //@{

    PlaylistModel& model()
    { return model_; }

    //* set current track index
    bool setCurrentTrackIndex( int );

    //* select previous track
    bool selectPreviousTrack();

    //* select next track
    bool selectNextTrack();

    //* save current playlist state
    void saveState();

    //* restore current playlist state
    void restoreState();

    //@}

    Q_SIGNALS:

    //* add selected audioStream to favorites
    void requestAddToFavorites( AudioStreamItem::List );

    #if WITH_TAGLIB
    //* request tag edition for specified files
    void requestEditTags( const File::List& );
    #endif

    public Q_SLOTS:

    //* update track
    /** track matching is based on file */
    void updateTrack( const TrackInformation& oldTrack, const TrackInformation& newTrack );

    //* update actions
    void updateActions();

    private Q_SLOTS:

    //* show tooltip
    void _showToolTip( const QModelIndex& );

    //* display cover
    void _displayCover();

    //* add selected audioStream to favorites
    void _addSelectionToFavorites();

    #if WITH_TAGLIB
    //* edit tags
    void _editTags();
    #endif

    private:

    //* install actions
    void _installActions();

    //* save current track id
    void _saveCurrentTrack();

    //* restore current track id
    void _restoreCurrentTrack();

    //* store selected tracks
    void _saveSelectedTracks();

    //* restore selected tracks
    void _restoreSelectedTracks();

    //* store model current index
    void _saveCurrentIndex();

    //* restore
    void _restoreCurrentIndex();

    //* treeview
    TreeView* list_ = nullptr;

    //* scrollbar position
    ScrollBarMonitor* scrollBarMonitor_ = nullptr;

    //* status bar
    BaseStatusBar* statusBar_ = nullptr;

    //* context menu
    ContextMenu* contextMenu_ = nullptr;

    //* model
    PlaylistModel model_;

    //* tooltip widget
    ToolTipWidget* toolTipWidget_ = nullptr;

    //* selected track (unique) ids
    using TrackIdList = QList<qint32>;
    TrackIdList selectedTracks_;

    //* current track (unique) id
    qint32 currentTrackId_ = -1;

    //* current index unique id
    qint32 currentIndexId_ = -1;

    //*@name actions
    //@{

    //* remove selected tracks
    QAction* removeTrackAction_ = nullptr;

    //* clear playlist
    QAction* clearPlaylistAction_ = nullptr;

    //* save playlist
    QAction* savePlaylistAction_ = nullptr;

    //* add to favorites
    QAction* addStreamToFavoritesAction_ = nullptr;

    //* display cover
    QAction* displayCoverAction_ = nullptr;

    #if WITH_TAGLIB
    //* edit tags
    QAction* editTagsAction_ = nullptr;
    #endif

    //@}

};

#endif
