/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "SidePanelToolBar.h"

#include "AudioDeviceWidget.h"
#include "CustomPixmap.h"
#include "CollectionWidget.h"
#include "CustomToolButton.h"
#include "Debug.h"
#include "FileSystemWidget.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "IconSizeMenu.h"
#include "LyricsWidget.h"
#include "AudioStreamWidget.h"
#include "MainWindow.h"
#include "SidePanelWidget.h"
#include "StoredPlaylistWidget.h"
#include "QtUtil.h"
#include "Singleton.h"
#include "ToolBarMenu.h"
#include "ToolButtonStyleMenu.h"

#include <QButtonGroup>
#include <memory>

//_______________________________________________________________
SidePanelToolBar::SidePanelToolBar( QWidget* parent ):
    CustomToolBar( tr( "Side bar" ), parent, "SIDE_PANEL_TOOLBAR" ),
    sidePanelWidget_( nullptr ),
    enabled_( true )
{
    Debug::Throw( "SidePanelToolBar:SidePanelToolBar.\n" );
    setTransparent( true );
    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
}

//_______________________________________________________________
void SidePanelToolBar::connect( SidePanelWidget& widget )
{

    Debug::Throw( "SidePanelToolBar::connect.\n" );
    sidePanelWidget_ = &widget;

    connect( &sidePanelWidget_->visibilityAction(), SIGNAL(toggled(bool)), SLOT(_sidePanelWidgetVisibilityChanged(bool)) );

    // button group
    buttonGroup_ = new QButtonGroup( this );
    connect( buttonGroup_, SIGNAL(buttonClicked(QAbstractButton*)), SLOT(_display(QAbstractButton*)) );
    buttonGroup_->setExclusive( false );

    // matching buttons
    QToolButton* button;

    // collection
    actions_.append( addWidget( button = _newToolButton( this,  IconEngine::get( IconNames::Collection ) ) ) );
    button->setText( tr( "Collection" ) );
    button->setToolTip( tr( "Collection" ) );

    buttonGroup_->addButton( button );
    buttons_.insert( button, &sidePanelWidget_->collectionWidget() );

    // stored playlists
    actions_.append( addWidget( button = _newToolButton( this, IconEngine::get( IconNames::Playlist ) ) ) );
    button->setText( tr( "Stored Playlists" ) );
    button->setToolTip( tr( "Stored Playlists" ) );

    buttonGroup_->addButton( button );
    buttons_.insert( button, &sidePanelWidget_->storedPlaylistWidget() );

    // lyrics
    actions_.append( addWidget( button = _newToolButton( this, IconEngine::get( IconNames::Lyrics ) ) ) );
    button->setText( tr( "Lyrics" ) );
    button->setToolTip( tr( "Lyrics" ) );

    buttonGroup_->addButton( button );
    buttons_.insert( button, &sidePanelWidget_->lyricsWidget() );

    // file system
    actions_.append( addWidget( button = _newToolButton( this, IconEngine::get( IconNames::FileSystem ) ) ) );
    button->setText( tr( "File System" ) );
    button->setToolTip( tr( "File System" ) );

    buttonGroup_->addButton( button );
    buttons_.insert( button, &sidePanelWidget_->fileSystemWidget() );

    // AudioStreams
    actions_.append( addWidget( button = _newToolButton( this, IconEngine::get( IconNames::AudioStreamList ) ) ) );
    button->setText( tr( "Audio Streams" ) );
    button->setToolTip( tr( "Audio Streams" ) );

    buttonGroup_->addButton( button );
    buttons_.insert( button, &sidePanelWidget_->audioStreamWidget() );

    // audio device
    #if WITH_CDIO_PARANOIA
    actions_.append( addWidget( button = _newToolButton( this, IconEngine::get( IconNames::AudioDevice ) ) ) );
    button->setText( tr( "Audio Device" ) );
    button->setToolTip( tr( "Audio Device" ) );

    buttonGroup_->addButton( button );
    buttons_.insert( button, &sidePanelWidget_->audioDeviceWidget() );
    #endif

    _updateConfiguration();

}

//____________________________________________________________
void SidePanelToolBar::_updateConfiguration()
{

    Debug::Throw( "SidePanelToolBar::_updateConfiguration.\n" );

    // icon size
    const IconSize iconSize( static_cast<IconSize::Size>( XmlOptions::get().get<int>( "SIDE_PANEL_TOOLBAR_ICON_SIZE" ) ) );

    // text label for toolbars
    const int toolButtonTextPosition( XmlOptions::get().get<int>( "SIDE_PANEL_TOOLBAR_TEXT_POSITION" ) );
    const auto style = static_cast<Qt::ToolButtonStyle>( toolButtonTextPosition >= 0 ?
        toolButtonTextPosition:
        this->style()->styleHint( QStyle::SH_ToolButtonStyle ) );

    // assign to toolbar
    setIconSize( iconSize.get() );
    setToolButtonStyle( style );

    // also update buttons independently
    for( auto&& iter = buttons_.begin(); iter != buttons_.end(); ++iter )
    {
        iter.key()->setToolButtonStyle( style );
        iter.key()->setIconSize( iconSize.get() );
    }

    // select proper button
    if( XmlOptions::get().contains( "SIDE_PANEL_CURRENT_WIDGET" ) )
    {
        const auto buttons( buttonGroup_->buttons() );
        const int currentWidget( XmlOptions::get().get<int>( "SIDE_PANEL_CURRENT_WIDGET" ) );
        if( buttons.size() > currentWidget && !buttons[currentWidget]->isChecked() ) buttons[currentWidget]->click();
    }

    _setMask( XmlOptions::get().get<unsigned int>( "SIDE_PANEL_MASK" ) );

    adjustSize();

}

//____________________________________________________________
void SidePanelToolBar::_updateToolButtonStyle( int style )
{

    Debug::Throw( "SidePanelToolBar::_updateToolButtonStyle.\n" );
    XmlOptions::get().set<int>( "SIDE_PANEL_TOOLBAR_TEXT_POSITION", (int)style );
    _updateConfiguration();

}

//____________________________________________________________
void SidePanelToolBar::_updateToolButtonIconSize( IconSize::Size size )
{

    Debug::Throw( "SidePanelToolBar::_updateToolButtonIconSize.\n" );
    XmlOptions::get().set<int>( "SIDE_PANEL_TOOLBAR_ICON_SIZE", size );
    _updateConfiguration();

}

//______________________________________________________________________
void SidePanelToolBar::_sidePanelWidgetVisibilityChanged( bool state )
{

    Debug::Throw() << "SidePanelToolBar::_sidePanelWidgetVisibilityChanged - state: " << state << endl;
    if( !state )
    {

        // make sure no button is checked
        for( auto&& iter = buttons_.begin(); iter != buttons_.end(); ++iter )
        { iter.key()->setChecked( false ); }

    } else {

        // make sure that one button is checked
        bool found( false );
        for( auto&& iter = buttons_.begin(); iter != buttons_.end() && !found; ++iter )
        {
            if( iter.value() == sidePanelWidget_->currentWidget() )
            {
                iter.key()->setChecked( true );
                found = true;
                break;
            }
        }

    }


}

//______________________________________________________________________
void SidePanelToolBar::_display( QAbstractButton* button )
{

    Debug::Throw( "SidePanelToolBar:_display.\n" );

    if( !enabled_ ) return;
    enabled_ = false;

    // retrieve widget in map
    bool state( button->isChecked() );
    QWidget* widget (0);
    for( auto&& iter = buttons_.begin(); iter != buttons_.end(); ++iter )
    {
        if( iter.key() == button ) widget = iter.value();
        else iter.key()->setChecked( false );
    }

    if( state )
    {

        /*
        warning: the order is important. The current widget must be changed first, otherwise two buttons
        will be checked in the end due to the _sidePanelWidgetVisibilityChanged slot.
        */
        sidePanelWidget_->setCurrentWidget( widget );
        sidePanelWidget_->visibilityAction().setChecked( true );

        // save
        const auto index( buttonGroup_->buttons().indexOf( button ) );
        if( index >= 0 ) XmlOptions::get().set<int>( "SIDE_PANEL_CURRENT_WIDGET", index );

    } else if( widget == sidePanelWidget_->currentWidget() ) {

        sidePanelWidget_->visibilityAction().setChecked( false );

    }

    enabled_ = true;
    return;

}

//______________________________________________________________________
void SidePanelToolBar::contextMenuEvent( QContextMenuEvent* event )
{
    Debug::Throw( "SidePanelToolBar::contextMenuEvent.\n" );

    MainWindow* mainwindow( qobject_cast<MainWindow*>( window() ) );
    if( !mainwindow ) return;
    std::unique_ptr<ToolBarMenu> menu( static_cast<ToolBarMenu*>(mainwindow->createPopupMenu() ) );

    menu->toolButtonStyleAction().setVisible( true );
    menu->toolButtonStyleMenu().select( (Qt::ToolButtonStyle) XmlOptions::get().get<int>( "SIDE_PANEL_TOOLBAR_TEXT_POSITION" ) );

    menu->iconSizeAction().setVisible(true);
    menu->iconSizeMenu().select( (IconSize::Size) XmlOptions::get().get<int>( "SIDE_PANEL_TOOLBAR_ICON_SIZE" ) );

    connect( &menu->toolButtonStyleMenu(), SIGNAL(styleSelected(int)), SLOT(_updateToolButtonStyle(int)) );
    connect( &menu->iconSizeMenu(), SIGNAL(iconSizeSelected(IconSize::Size)), SLOT(_updateToolButtonIconSize(IconSize::Size)) );

    // move and show menu
    menu->adjustSize();
    menu->exec( event->globalPos() );

}

//______________________________________________________________________
QToolButton* SidePanelToolBar::_newToolButton( QWidget* parent, QIcon icon ) const
{

    Debug::Throw( "SidePanelToolBar::_newToolButton.\n" );
    CustomToolButton* button = new CustomToolButton( parent );
    button->setIcon( icon );
    button->setCheckable( true );
    button->setSizePolicy( QSizePolicy::Expanding, QSizePolicy::Preferred );
    button->setUpdateFromOptions( false );
    button->setProperty( "_kde_toolButton_alignment", Qt::AlignLeft );

    return button;

}

//______________________________________________________________________
void SidePanelToolBar::_setMask( unsigned int mask )
{
    Debug::Throw() << "SidePanelToolBar::_setMask - mask: " << mask << endl;
    for( int i = 0; i < actions_.size(); ++i )
    { actions_[i]->setVisible( mask & (1<<i) ); }
}

//______________________________________________________________________
unsigned int SidePanelToolBar::_mask() const
{
    Debug::Throw( "SidePanelToolBar::_mask.\n" );
    unsigned int out(0);
    for( int i = 0; i < actions_.size(); ++i )
    { if( actions_[i]->isVisible() ) out |= (1<<i); }

    return out;
}
