/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TrackInformation.h"

#include "Debug.h"
#include "XmlDef.h"

#include <QRegExp>
#include <QStringList>
#include <QUrl>

#if QT_VERSION >= 0x050000
#include <QUrlQuery>
#endif

//_______________________________________________________________________
const QString TrackInformation::MimeType( "musicbox/track-information-list" );
const QString TrackInformation::UnknownArtist = QObject::tr( "Unknown Artist" );
const QString TrackInformation::UnknownAlbum = QObject::tr( "Unknown Album" );

//_______________________________________________________________________
TrackInformation::TrackInformation( const QString& reply ):
    Counter( "TrackInformation" )
{

    Debug::Throw( "TrackInformation::TrackInformation.\n" );

    auto parse = []( const QString& input, const QString& key )
    { return input.startsWith(key) ? std::make_pair( true, input.mid( key.length() ) ) : std::make_pair( false, QString() ); };

    const auto replyList( reply.split( '\n', QString::SkipEmptyParts ) );
    std::pair<bool, QString> result;
    for( const auto& replyString:replyList )
    {
        if( ( result = parse( replyString, "file: " ) ).first ) file_ = File( result.second );
        else if( ( result = parse( replyString, "Artist: " ) ).first ) artist_ = result.second;
        else if( ( result = parse( replyString, "Title: " ) ).first ) title_ = result.second;
        else if( ( result = parse( replyString, "Album: " ) ).first ) album_ = result.second;
        else if( ( result = parse( replyString, "Track: " ) ).first ) track_ = result.second.toInt();
        else if( ( result = parse( replyString, "Date: " ) ).first ) year_ = result.second.toInt();
        else if( ( result = parse( replyString, "Time: " ) ).first ) duration_ = result.second.toInt();
        else if( ( result = parse( replyString, "Pos: " ) ).first ) trackIndex_ = result.second.toInt();
        else if( ( result = parse( replyString, "Id: " ) ).first ) trackId_ = result.second.toInt();
        else if( ( result = parse( replyString, "Disc: " ) ).first ) disc_ = result.second.toInt();
    }

    // try parse from url
    if( isAudioDeviceStream() )
    {

        TrackInformation local( QUrl::fromEncoded( file_.get().toUtf8() ) );
        artist_ = local.artist_;
        album_ = local.album_;
        title_ = local.title_;
        track_ = local.track_;
        duration_ = local.duration_;

    } else if( isAudioStream() ) {

        if( artist_.isEmpty() )
        {
            QRegExp regExp( "^(?:[\\+\\*@]\\s+)?([^\\+\\*@].*)\\s+-\\s+(.*[^\\+\\*@])(?:\\s+[\\*\\+@])?$" );
            regExp.setMinimal( true );

            // try guess artist from title
            // QRegExp regExp( "\\s+-\\s+" );
            const auto position( title_.indexOf( regExp ) );
            if( position >= 0 )
            {
                artist_ = regExp.cap(1);
                title_ = regExp.cap(2);
            }
        }

    } else if( track_ < 0 ) {

        // try parse track from filename, if not set
        QRegExp regExp( "(?:^|\\(|\\.|-|_|\\s)(\\d+)(?:\\)|\\.|-|_|\\s)" );
        if( file_.localName().get().indexOf( regExp ) >= 0 )
        {
            Debug::Throw() << "TrackInformation::TrackInformation - title: " << file_.localName().get() << " capture: " << regExp.cap(0) << endl;
            track_ = regExp.cap(1).toInt();
        }

    }

}

//_______________________________________________________________________
TrackInformation::TrackInformation( const QUrl& url ):
    Counter( "TrackInformation" )
{

    Debug::Throw( "TrackInformation::TrackInformation (url).\n" );

    #if QT_VERSION >= 0x050000
    const QUrlQuery query( url );
    #else
    const QUrl& query( url );
    #endif

    if( query.hasQueryItem( "artist" ) ) artist_ =  query.queryItemValue( "artist" );
    if( query.hasQueryItem( "album" ) ) album_ = query.queryItemValue( "album" );
    if( query.hasQueryItem( "title" ) ) title_ =  query.queryItemValue( "title" );
    if( query.hasQueryItem( "track" ) ) track_ =  query.queryItemValue( "track" ).toInt();
    if( query.hasQueryItem( "duration" ) ) duration_ =  query.queryItemValue( "duration" ).toInt();
}

//_______________________________________________________________________
TrackInformation::TrackInformation( const QDomElement& element ):
    Counter( "TrackInformation" )
{

    Debug::Throw( "TrackInformation::TrackInformation (dom).\n" );

    // parse attributes
    const auto attributes( element.attributes() );
    for( int i=0; i<attributes.count(); i++ )
    {
        const auto attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;
        else if( attribute.name() == Xml::Artist ) artist_ = attribute.value();
        else if( attribute.name() == Xml::Album ) album_ = attribute.value();
        else if( attribute.name() == Xml::AudioStreamChannel ) channel_ = attribute.value();
        else if( attribute.name() == Xml::Title ) title_ = attribute.value();
        else if( attribute.name() == Xml::Disc ) disc_ = attribute.value().toInt();
        else if( attribute.name() == Xml::Year ) year_ = attribute.value().toInt();
        else if( attribute.name() == Xml::Track ) track_ = attribute.value().toInt();
        else if( attribute.name() == Xml::Duration ) duration_ = attribute.value().toInt();
    }

}

//_______________________________________________________________________
QString TrackInformation::shortArtist( const QString& artist )
{
    QRegExp regExp( "^\\s*(?:T|t)he\\s+" );
    return artist.indexOf( regExp ) == 0 ? artist.mid( regExp.matchedLength() ):artist;
}

//_______________________________________________________________________
QString TrackInformation::durationString( quint32 time, bool forceHours )
{
    quint32 days = time/(3600*24);
    quint32 hours = (time = (time - 3600*24*days) )/3600;
    quint32 minutes = ( time = (time - 3600*hours) )/60;
    quint32 seconds = time - 60*minutes;
    if( days > 0 ) return QString().sprintf( "%i:%02i:%02i:%02i", days, hours, minutes, seconds );
    else if( hours > 0 || forceHours ) return QString().sprintf( "%i:%02i:%02i", hours, minutes, seconds );
    else return QString().sprintf( "%i:%02i", minutes, seconds );
}

//_______________________________________________________________________
QUrl TrackInformation::url( const Base::QueryItemList& extraItems ) const
{
    Debug::Throw( "TrackInformation::url.\n" );

    QUrl url;

    // setup query items
    auto queryItems( extraItems );
    if( !artist_.isEmpty() ) queryItems.append( { "artist", artist_ });
    if( !album_.isEmpty() ) queryItems.append( { "album", album_ });
    if( !title_.isEmpty() ) queryItems.append( { "title", title_ });
    if( track_ > 0 ) queryItems.append( { "track", QString::number( track_ ) });
    if( duration_ > 0 ) queryItems.append( { "duration", QString::number( duration_ ) });

    QUrlUtil::addQueryItems( url, queryItems );
    return url;
}

//_______________________________________________________________________
QDomElement TrackInformation::domElement( QDomDocument& document ) const
{
    Debug::Throw( "TrackInformation::domElement.\n" );
    auto out( document.createElement( Xml::TrackInformation ) );

    if( !artist_.isEmpty() ) out.setAttribute( Xml::Artist, artist_ );
    if( !album_.isEmpty() ) out.setAttribute( Xml::Album, album_ );
    if( !channel_.isEmpty() ) out.setAttribute( Xml::AudioStreamChannel, channel_ );
    if( !title_.isEmpty() ) out.setAttribute( Xml::Title, title_ );
    if( disc_ > 0 ) out.setAttribute( Xml::Disc, QString::number( disc_ ) );
    if( year_ > 0 ) out.setAttribute( Xml::Year, QString::number( year_ ) );
    if( track_ > 0 ) out.setAttribute( Xml::Track, QString::number( track_ ) );
    if( duration_ > 0 ) out.setAttribute( Xml::Duration, QString::number( duration_ ) );
    return out;
}

//_______________________________________________________________________
bool TrackInformation::isAudioStream() const
{ return file_.get().startsWith( "http://" ) && !isAudioDeviceStream(); }

//_______________________________________________________________________
bool TrackInformation::isAudioDeviceStream() const
{ return file_.get().startsWith( "http://" ) && QUrl::fromEncoded( file_.get().toUtf8() ).path().startsWith( QLatin1String( "/musicbox-audiostream" ) ); }

//_______________________________________________________________________
QTextStream& operator << ( QTextStream& out, const TrackInformation& track )
{

    out << " artist: " << track.artist();
    if( !track.album().isEmpty() ) out << " album: " << track.album();
    if( !track.title().isEmpty() ) out << " title: " << track.title();
    if( track.year() > 0 ) out << " year: " << track.year();
    if( track.disc() > 0 ) out << " disc: " << track.disc();
    if( track.track() > 0 ) out << " track: " << track.track();

    return out;

};

//_______________________________________________________________________
TrackInformation::List TrackInformation::Helper::list( const QDomElement& parent )
{
    Debug::Throw( "TrackInformation::Helper::list.\n" );
    TrackInformation::List out;
    for( auto&& node = parent.firstChild(); !node.isNull(); node = node.nextSibling() )
    {
        auto element = node.toElement();
        if( element.isNull() ) continue;
        if( element.tagName() == Xml::TrackInformation )
        { out.append( TrackInformation( element ) ); }
    }

    return out;
}

//_______________________________________________________________________
QDomElement TrackInformation::Helper::domElement( const TrackInformation::List& trackInformationList, QDomDocument& parent )
{
    Debug::Throw( "TrackInformation::Helper::domElement.\n" );

    // create main element
    auto out( parent.createElement( Xml::TrackInformationList ) );
    for( const auto& trackInformation:trackInformationList )
    { out.appendChild( trackInformation.domElement( parent ) ); }

    return out;
}
