#ifndef TrackInformation_h
#define TrackInformation_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "Debug.h"
#include "File.h"
#include "Functors.h"
#include "QUrlUtil.h"

#include <QDomElement>
#include <QDomDocument>
#include <QList>
#include <QRegExp>
#include <QString>
#include <QTextStream>
#include <QUrl>

#include <functional>

class TrackInformation: private Base::Counter<TrackInformation>
{

    public:

    using List = QList<TrackInformation>;

    //* used when LogEntry keyword is not defined
    static const QString MimeType;

    static const QString UnknownArtist;
    static const QString UnknownAlbum;

    //* default constructor
    explicit TrackInformation(): Counter( "TrackInformation" )
    {}

    //* constructor from string
    explicit TrackInformation( const QString& );

    //* constructor from url
    explicit TrackInformation( const QUrl& );

    //* constructor from dom
    explicit TrackInformation( const QDomElement& );

    //* mask
    enum Tag
    {
        None = 0,
        Artist = 1<<0,
        Album = 1<<1,
        Title = 1<<2,
        Year = 1<<3,
        Track = 1<<4,
        Disc = 1<<5,
        Default = Artist|Album|Title,
        All = Artist|Album|Title|Year|Disc|Track
    };

    Q_DECLARE_FLAGS(Mask, Tag)

    //*@name accessors
    //@{

    //* file
    const File& file() const
    { return file_; }

    //* artist
    const QString& artist() const
    { return artist_; }

    //* short artist, used for sorting
    QString shortArtist() const
    { return shortArtist( artist_ ); }

    //* short artist, used for sorting
    static QString shortArtist( const QString& );

    //* album
    const QString& album() const
    { return album_; }

    //* channel (for audio streams)
    const QString& channel() const
    { return channel_; }

    //* album or channel
    const QString& albumOrChannel() const
    { return album_.isEmpty() ? channel_:album_; }

    //* disc
    qint32 disc() const
    { return disc_; }

    //* year
    qint32 year() const
    { return year_; }

    //* title
    const QString& title() const
    { return title_; }

    //* title of file
    QString titleOrFile() const
    { return title_.isEmpty() ? file_.localName().truncatedName():title_; }

    //* track number
    qint32 track() const
    { return track_; }

    //* duration
    qint32 duration() const
    { return duration_; }

    //* format time
    static QString durationString( quint32, bool forceHours = false );

    //* duration
    QString durationString( bool forceHours = false ) const
    { return durationString( duration_, forceHours ); }

    //* track index in playlist
    qint32 trackIndex() const
    { return trackIndex_; }

    //* track id in playlist
    qint32 trackId() const
    { return trackId_; }

    //* url from file
    /**
     created url is invalid: it has no path,
     however it contains queryitems that correspond to the track informations
    */
    QUrl url( const Base::QueryItemList& = Base::QueryItemList() ) const;

    //* dom element
    QDomElement domElement( QDomDocument& ) const;

    //* true if the track information corresponds to an audio stream
    bool isAudioStream() const;

    //* true if the track information corresponds to a stream
    bool isAudioDeviceStream() const;

    //* hash
    virtual uint hashKey() const
    { return qHash( (artist_+album_+title_).toLower() ); }

    //@}

    //*@name modifiers
    //@{

    //* reset
    void reset()
    { *this = TrackInformation(); }

    //* file
    void setFile( const File& value )
    { file_ = value; }

    //* artist
    void setArtist( const QString& value )
    { artist_ = value; }

    //* album
    void setAlbum( const QString& value )
    { album_ = value; }

    //* channel (for audio streams)
    void setChannel( const QString& value )
    { channel_ = value; }

    //* disc
    void setDisc( qint32 value )
    { disc_ = value; }

    //* year
    void setYear( qint32 value )
    { year_ = value; }

    //* title
    void setTitle( const QString& value )
    { title_ = value; }

    //* duration
    void setDuration( qint32 value )
    { duration_ = value; }

    //* track number in disc
    void setTrack( qint32 value )
    { track_ = value; }

    //* track id (internal mpd)
    void setTrackId( qint32 value )
    { trackId_ = value; }

    //* track index in playlist
    void setTrackIndex( qint32 value )
    { trackIndex_ = value; }

    //@}

    //* equal to ftor
    class WeakEqualFTor:
        public Base::Functor::BinaryEqual<TrackInformation, qint32, &TrackInformation::trackIndex>,
        public std::binary_function< TrackInformation, TrackInformation, bool>
    {};

    //* same file
    using SameFileFTor = Base::Functor::UnaryEqual<TrackInformation, const File&, &TrackInformation::file>;

    // same track number
    using SameTrackFTor = Base::Functor::UnaryEqual<TrackInformation, qint32, &TrackInformation::track>;

    //* same track id
    using SameTrackIdFTor = Base::Functor::UnaryEqual<TrackInformation, qint32, &TrackInformation::trackId>;

    //* same artist
    using SameArtistFTor = Base::Functor::UnaryEqual<TrackInformation, const QString&, &TrackInformation::artist>;

    //* same artist
    using SameArtistBinaryFTor = Base::Functor::BinaryEqual<TrackInformation, const QString&, &TrackInformation::artist>;

    //* is audio stream
    using IsAudioStreamFTor = Base::Functor::UnaryTrue<TrackInformation, &TrackInformation::isAudioStream>;

    //* is audio stream
    using IsAudioDeviceStreamFTor = Base::Functor::UnaryTrue<TrackInformation, &TrackInformation::isAudioDeviceStream>;

    //* is editable
    using IsEditableFTor = Base::Functor::UnaryFalse<TrackInformation, &TrackInformation::isAudioStream>;

    //* same album
    class SameAlbumFTor
    {

        public:

        //* constructor
        explicit SameAlbumFTor( const TrackInformation& track ):
            artist_( track.artist() ),
            album_( track.album() ),
            disc_( track.disc() )
        {}

        //* operator
        bool operator() ( const TrackInformation& track ) const
        { return track.artist() == artist_ && track.album() == album_ && track.disc() == disc_; }

        private:

        QString artist_;
        QString album_;
        qint32 disc_;

    };

    //* same album
    class SameAlbumBinaryFTor
    {

        public:

        bool operator() ( const TrackInformation& lhs, const TrackInformation& rhs ) const
        { return lhs.artist() == rhs.artist() && lhs.album() == rhs.album() && lhs.disc() == rhs.disc(); }

    };

    //* helper class to parse List from string
    class Helper
    {
        public:

        //* get list from string
        template<class T>
        static typename T::List list( const QString& );

        //* get list from dom element
        static TrackInformation::List list( const QDomElement& );

        //* get dom element from list
        static QDomElement domElement( const TrackInformation::List&, QDomDocument& );

    };

    private:

    File file_;
    QString artist_;
    QString album_;
    QString channel_;
    QString title_;
    qint32 disc_ = -1;
    qint32 year_ = -1;

    //* track number in disc
    qint32 track_ = -1;

    //* track duration
    qint32 duration_ = 0;

    //* track position in playlist
    qint32 trackIndex_ = -1;

    //* track id (internal mpd)
    qint32 trackId_ = -1;

    friend QTextStream& operator << (QTextStream&, const TrackInformation& );

};

//_______________________________________________________________________
template<class T>
typename T::List TrackInformation::Helper::list( const QString& reply )
{
    typename T::List tracks;
    QRegExp regExp( "\n(OK|file: )" );
    int previousPosition = 0;
    for( int position = 0; ( position = reply.indexOf( regExp, position ) ) >= 0; position += regExp.matchedLength() )
    {
        if( position == previousPosition ) continue;
        const auto trackReply = reply.mid( previousPosition, position - previousPosition );
        const T track( trackReply );

        if( !track.file().isEmpty() ) tracks.push_back( track );
        previousPosition = position;
    }

    return tracks;

}

Q_DECLARE_OPERATORS_FOR_FLAGS( TrackInformation::Mask )


#endif
