#ifndef VolumeControlMenu_h
#define VolumeControlMenu_h

/******************************************************************************
*
* Copyright (C) 2017 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"

#include <QLabel>
#include <QMenu>
#include <QPropertyAnimation>
#include <QSlider>
#include <QToolButton>
#include <QWheelEvent>

class VolumeControlMenu: public QMenu, private Base::Counter<VolumeControlMenu>
{

    Q_OBJECT

    Q_PROPERTY( int fadeVolume READ fadeVolume WRITE setFadeVolume );

    public:

    //* constructor
    explicit VolumeControlMenu( QWidget* = nullptr );

    //*@name accessors
    //@{

    //* slider
    QSlider& volumeSlider() const
    { return *volumeSlider_; }

    //* volume
    int volume() const
    { return volumeSlider_->value(); }

    //* true if muted
    bool isMuted() const
    { return muteButton_->isChecked(); }

    //* true if fading is enabled
    bool fadingEnabled() const
    { return useFadeAnimation_; }

    //* true if fading
    bool isFading() const;

    //* fade volume
    int fadeVolume() const
    { return fadeVolume_; }

    //@}

    //*@name modifiers
    //@{

    //* lock
    void lock()
    { locked_ = true; }

    //* unlock
    void unlock()
    { locked_ = false; }

    //* start fading
    void startFade();

    //* abort fading
    void abortFade();

    //* set fade volume
    void setFadeVolume( int );

    //@}

    Q_SIGNALS:

    //* emitted when volume is changed
    void volumeChanged( int );

    //* emitted when volume state is changed
    void muteStateChanged( bool );

    //* emitted when fade starts
    void fadeStarted();

    //* emitted when fading is finished
    void fadeCompleted();

    //* emitted when fade is aborted
    void fadeAborted();

    public Q_SLOTS:

    //* volume
    void setVolume( int );

    protected:

    //* wheel event
    void wheelEvent( QWheelEvent* ) override;

    private Q_SLOTS:

    //* update configuration
    void _updateConfiguration();

    //* volume changed from slider
    void _volumeChanged( int );

    //* mute state changed from button
    void _muteStateChanged( bool );

    //* fade completed
    void _fadeCompleted();

    private:

    //* locked
    /** prevents roundtrip to mpd */
    bool locked_ = false;

    //* volume label
    QLabel* volumeLabel_ = nullptr;

    //* slider
    QSlider* volumeSlider_ = nullptr;

    //* mute
    QToolButton* muteButton_ = nullptr;

    //* fade property
    QPropertyAnimation* fadeAnimation_ = nullptr;

    //* true if fade animation is to be used
    bool useFadeAnimation_ = false;

    //* duration
    int fadeAnimationDuration_ = 0;

    //* fade starting volume
    int fadeStartingVolume_ = 0;

    //* fade volume
    int fadeVolume_ = 0;

};

#endif
